#! /usr/bin/env python
"""
alignment_functions.py

Created by Lucas Swanson
Copyright (c) 2012 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

# import standard modules
import time

# import custom modules
from utils.error import MyError
from utils.general import (TimeSpent, AddChr, NormalizeChrID,
    FormatChromosomeName)
from utils.messages import ErrMsg, LogMsg, DebugMsg, TestMsg#, ExtremeDebugMsg
from parsers.psl import parse as ParsePSL
from parsers.psl import PSL
from parsers.sam import SAM
from common.coord_pair import CoordPairCls

# CONSTANTS
PSL_TNAME_COL = 13

def ParseAlignmentFile(
      aligns_path, filters, log_info, ParseFunction=ParsePSL,
      robust_mode=False): #{
  start = time.time()
  LogMsg(log_info, "Parsing alignment file: %s" % aligns_path)
  if robust_mode:
      DebugMsg(log_info, '** Using robust mode! **')
  aligns = None
  # parse the alignment file
  aligns = ParseFunction(
      aligns_path, filters, log_info=log_info, robust_mode=robust_mode)
  DebugMsg(log_info, "Found %i alignments" % len(aligns))
  LogMsg(log_info, "Time spent parsing alignment file: %s" % TimeSpent(start))
  if (None == aligns or 0 == len(aligns)): #{
    raise NoAlignmentsError("No alignments were found.")
  #} end if
  return aligns
#} end def

def CalcAlignOverlap(align1, align2): #{
  return CalcOverlap(align1.qstart, align1.qend, align2.qstart, align2.qend)
#} end def

def CalcOverlap(startA, endA, startB, endB): #{
  # order the coordinate pairs so that the smaller coordinate is first
  leftA  = min(startA, endA)
  rightA = max(startA, endA)
  leftB  = min(startB, endB)
  rightB = max(startB, endB)
  # determine the overlap (will be negative if there is a space)
  overlap_left  = max(leftA,  leftB)
  overlap_right = min(rightA, rightB)
  overlap = (overlap_right - overlap_left) + 1
  #ErrMsg("lA: %i, rA: %i, lB: %i, rB: %i, lO: %i, rO: %i, O: %i" %
  #  (leftA, rightA, leftB, rightB, overlap_left, overlap_right, overlap))
  overlap_fraction = 0.0
  # if the coordinates overlapped, calculate the fraction
  if (0 < overlap): #{
    spanA = (rightA - leftA) + 1
    spanB = (rightB - leftB) + 1
    min_span = min(spanA, spanB)
    #ErrMsg("spanA: %i, spanB: %i, min: %i" % (spanA, spanB, min_span))
    overlap_fraction = float(overlap) / float(min_span)
  #} end if
  #ErrMsg("Overlap fraction: %f" % overlap_fraction)
  return (overlap, overlap_fraction)
#} end def

def FixNonChromAlign(align): #{
  return FixAlign(align, chromosomal=False)
#} end def

def FixAlign(align, chromosomal=True): #{
  if (hasattr(align, "fixed") and align.fixed): #{
    return align
  #} end if
  if (not hasattr(align, "multi_mapped")): #{
    # assume the alignment was not multi_mapped
    align.multi_mapped = False
  #} end if
  # convert coordinates to integers
  align.qstart   = int(align.qstart)
  align.qend     = int(align.qend)
  align.tstart   = int(align.tstart)
  align.tend     = int(align.tend)
  align.score    = int(align.score)
  align.identity = float(align.identity)
  # assume that the alignment does not contain a query gap
  align.qgap = False
  # assume that the alignment is not spliced
  align.spliced = False
  # ensure that the chromosome name does not contain chr
  if (chromosomal): #{
    align.target = NormalizeChrID(align.target)
  #} end if
  # fix psl or sam alignment
  if (align.method in ["blat", PSL, SAM]): #{
    # convert numeric values to integers
    numeric_fields = ["match", "mismatch", "query_len", "target_len",
      "qnuminsert", "qbaseinsert", "tnuminsert", "tbaseinsert"]
    for field in numeric_fields: #{
      try: #{
        int_value = int(getattr(align, field))
        setattr(align, field, int_value)
      except ValueError, e:
        raise AlignmentError("%s must be a numeric value, not \"%s\"" %
          (field, getattr(align, field)))
      #} end try
    #} end for
    # reorder target co-ordinates of alignment to the "-" strand
    if ("-" == align.query_strand): #{
      #temp_tstart  = align.tstart
      #align.tstart = align.tend
      #align.tend   = temp_tstart
      (align.tend, align.tstart) = sorted((align.tstart, align.tend))
    #} end if
    # fix the psl string
    if (None != align.psl_str and chromosomal): #{
      align.psl_str = FormatChromosomeNameInPsl(align.psl_str, True, True)
      #psl_info = align.psl_str.split("\t")
      #target_name = psl_info[PSL_TNAME_COL]
      ## convert chrMT to chrM and ensure that the psl string has "chr"
      ##  in the chromosome name
      #if (chromosomal): #{
      #  target_name = NormalizeChrID(target_name, use_chr=True)
      ##} end if
      #psl_info[PSL_TNAME_COL] = target_name
      #align.psl_str = "\t".join(psl_info)
    #} end if
    # check whether the alignment contains a query gap
    if (0 < align.qnuminsert): #{
      #print "QGap: %i\n" % int(align.qnuminsert)
      #print "PSL: %s\n" % align.psl_str
      align.qgap = True
    #} end if
    # use target gap value to determine whether the alignment is spliced
    if (0 < align.tnuminsert): #{
      align.spliced = True
    #} end if
  # fix exonerate alignments
  elif (align.method in ["exonerate", "exon"]):
    # set the query strand
    if (align.target_strand): #{
      align.query_strand = "+"
    else:
      align.query_strand = "-"
      align.blocks.reverse()
    #} end if
    # use the number of blocks to determine whether the alignment is spliced
    if (1 < len(align.blocks)): #{
      align.spliced = True
    #} end if
  #} end if
  # setup "coords" data members after fixing coordinate order
  align.qcoords  = CoordPairCls(align.qstart, align.qend)
  align.tcoords  = CoordPairCls(align.tstart, align.tend)
  # alias for target blocks
  align.target_blocks = align.blocks
  # setup "coords" block data members
  align.qblock_coords = [CoordPairCls(block) for block in align.query_blocks]
  align.tblock_coords = [CoordPairCls(block) for block in align.target_blocks]
  align.fixed = True
  return align
#} end def

def FormatChromosomeNameInPsl(psl_str, use_chr, use_mt): #{
  psl_info = psl_str.split("\t")
  target_name = psl_info[PSL_TNAME_COL]
  target_name = FormatChromosomeName(target_name, use_chr, use_mt)
  psl_info[PSL_TNAME_COL] = target_name
  psl_str = "\t".join(psl_info)
  return psl_str
#} end def

def FixNonChromAlign(align): #{
  return FixAlign(align, chromosomal=False)
#} end def

def IntifyBlock(block): #{
  return map(int, block)
#} end def

def WriteBlockCoords(align, id, file, use_chr=False): #{
  target = NormalizeChrID(align.target, use_chr=use_chr)
  #if (use_chr): #{
    #if (align.target.startswith("chr")): #{
    #  target = align.target
    #else:
    #  target = "chr%s" % align.target
    #} end if
  #  target = AddChr(target)
  #else:
    #target = align.target.replace("chr", "", 1)
  #} end if
  for block in map(IntifyBlock, align.target_blocks): #{
    (block_start, block_end) = block
    block_left  = min(block_start, block_end)
    block_right = max(block_start, block_end)
    coord_str = "%s %i %i %i" % (target, block_left, block_right, id)
    TestMsg("Writing coords: %s" % coord_str)
    file.WriteLine(coord_str)
  #} end for
#} end def

def AlignListString(align_list, use_chr=True): #{
  align_strs = list()
  align_strs.append("---")
  for i in range(len(align_list)): #{
    align_strs.append("%i) %s" %
        (i,AlignString(align_list[i], use_chr=use_chr)))
  #} end for
  align_strs.append("---")
  return "\n".join(align_strs)
#} end def

def AlignString(align, use_chr=True): #{
  if (use_chr): #{
    target = AddChr(align.target)
  else:
    target = align.target
  #} end if
  return ("S:%i ID:%.3f BA:%i QL:%i Q:%s QS:%i QE:%i T:%s TS:%i TE:%i" %
    (int(align.score), float(align.identity),
     int(align.num_bases_aligned), int(align.query_len),
     align.query,  align.qstart, align.qend,
     target, align.tstart, align.tend))
#} end def

def ShortAlignString(align, use_chr=True): #{
  if (use_chr): #{
    target = AddChr(align.target)
  else:
    target = align.target
  #} end if
  return ("S:%i ID:%.1f Q:%s QS:%i QE:%i QL:%i T:%s:%i" %
    (int(align.score), float(align.identity), align.query, int(align.qstart),
     int(align.qend), int(align.query_len), target, align.tstart))
#} end def

class AlignBlockCls: #{
  def __init__(self, block_coords_tuple): #{
    if (block_coords_tuple[0] <= block_coords_tuple[1]): #{
      self.strand = "+"
    else:
      self.strand = "-"
    #} end if
    (self.start, self.end) = sorted(map(int, block_coords_tuple))
    self.span = (self.end-self.start)+1
  #} end def

  def __str__(self): #{
    return "%i-%i" % (self.start, self.end)
  #} end def

  def Gap(self, other): #{
    return min((other.start-self.end),(self.start-other.end))-1
  #} end def
#} end class

#### EXCEPTION CLASSES ####
class AlignmentError(MyError): #{
  pass
#} end class

class NoAlignmentsError(MyError): #{
  pass
#} end class
