#! /usr/bin/env python
"""
gap_filter.py

Created by Lucas Swanson
Copyright (c) 2012 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

# import standard modules
import re

# import custom modules
from utils.error import MyError
from utils.general import SetupMainClass, NormalizeChrID, NonStandardChr
from utils.messages import LogMsg, DebugMsg, ExtremeDebugMsg
from utils.files_paths import FileBoxCls
from utils.subprocesses import RunCommandFromString
from common.coord_pair import CoordPairCls
from alignment_processing.gene_overlap import (ChooseBestTranscripts,
    GenesOverlapped, NearbyGenes)
from alignment_processing.gap_candidate import GapCandidateCls
from alignment_processing.alignment_functions import (IntifyBlock,
    ShortAlignString, AlignString, FormatChromosomeNameInPsl)

# CONSTANTS
GAP_REALIGNER_FAIL = 0
GAP_REALIGNER_MIN_SUCCESS = 1
GAP_REALIGNER_MAX_SUCCESS = 100
GAP_REALIGNER_ONLY_NON_BASIC = 110
MAX_INTERNAL_DUP_OFFSET = 5

class GapFilterCls: #{
  def __init__(self, ctg_info, gap_out_path, chr_conventions,
      options, log_info=None): #{
    SetupMainClass(self, options, log_info=log_info)
    if (None == ctg_info): #{
      self.ctg_seq = ""
    else:
      self.ctg_seq = ctg_info.sequence
    #} end if
    self.gap_out_path = gap_out_path
    self.chr_conventions = chr_conventions
    self.ResetFilter()
  #} end def

  def ResetFilter(self): #{
    ExtremeDebugMsg(self, "Resetting gap filter!!")
    self.gapped_event_found = "N"
    self.seq_too_long = False
    # assume no gapped alignments will be found for the current contig
    self.num_gapped_aligns = 0
    self.max_score = 0
    self.multi_mapped = False
    # gaps_to_check is a list of CoordPairCls objects representing gaps
    # to check for duplications
    self.gaps_to_check = list()
    self.gaps_with_dups = list()
  #} end def

  # TODO: Break up this too-long function.
  def FindGappedAlignment(self, curr_align): #{
    if (self.seq_too_long): #{
      return False
    #} end if
    #LogMsg(self, "Running gap filter") # DEBUG
    if (None == curr_align.psl_str): #{
      psl_str = curr_align.psl()
    else:
      psl_str = curr_align.psl_str
    #} end if
    # ensure that the chromosome name is formatted correctly
    psl_str = FormatChromosomeNameInPsl(psl_str,
        self.chr_conventions["use_chr"], self.chr_conventions["use_mt"])
    if (None == psl_str or "" == psl_str): #{
      msg = "Cannot use gap_realigner on alignment without psl string"
      raise GapFilterError(msg)
    #} end if
    # if the alignment should be passed to the gap realigner
    #ExtremeDebugMsg(self, "Calling gap_realigner with sequence: %s" %
    #    self.ctg_seq)
    if (not self.FilterCheck(curr_align)): #{
      return False
    #} end if
    self.CheckBlocks(curr_align)
    if (not self.gaps_to_check): #{
      return False
    #} end if
    # do some simple internal duplication checking
    self.internal_duplication_checking(curr_align)
    if (self.gaps_with_dups): #{
      self.create_gap_candidates(curr_align)
      return True
    #} end if
    #DebugMsg(self, "--alignment passed gap filter") # DEBUG
    # run next gap-checking step
    gap_realign_cmd = self.get_gap_realign_command(curr_align, psl_str)
    if (not gap_realign_cmd): #{
      return False
    #} end if
    #LogMsg(self, "Checking gapped alignment...") # DEBUG
    realigner_return = RunCommandFromString(gap_realign_cmd,
      dpt=self.options.dpt)
    #DebugMsg(self, "Gap realigner returned: %i" % realigner_return) # DEBUG
    if (GAP_REALIGNER_MIN_SUCCESS <= realigner_return and
        GAP_REALIGNER_MAX_SUCCESS >= realigner_return): #{
      if (GAP_REALIGNER_MAX_SUCCESS == realigner_return): #{
        self.more_than_99 = True
      #} end if
      self.num_gapped_aligns += realigner_return
      DebugMsg(self, "%i gapped alignment(s) found for %s" %
        (realigner_return, curr_align.query))
      self.gapped_event_found = "Y"
      return True
    elif (GAP_REALIGNER_ONLY_NON_BASIC == realigner_return):
      ExtremeDebugMsg(self, "Warning: gap realigner found only "
        "non-basic events.")
    elif (0 > realigner_return):
      raise GapFilterError("Gap realigner command terminated by "
        "signal %i:\n%s" % (-realigner_return, gap_realign_cmd))
    elif (GAP_REALIGNER_FAIL != realigner_return):
      raise GapFilterError("Error running gap realigner: %i:\n%s" %
        (realigner_return, gap_realign_cmd))
    #} end if
    return False
  #} end def

  def FilterCheck(self, align): #{
    ExtremeDebugMsg(self, ShortAlignString(align))
    # check that the contig span is wide enough
    if (not self.CheckContigSpan(align)): #{
      return False
    # check to see whether there are enough unaligned bases for a
    # sufficiently sized gap to be possible
    elif (not self.CheckUnalignedBases(align)): #{
      return False
    # check against the max score seen for the current contig
    elif (not self.CheckScoreFraction(align)): #{
      return False
    # check to see whether the alignment has a high enough percent identity
    elif (not self.CheckPercentIdentity(align)): #{
      return False
    # do not check alignments to mitochondrial DNA if the discard
    # mitochondrial DNA flag is set
    elif (not self.CheckMitochondrial(align)): #{
      return False
    # check that the alignment target is a standard chromosome
    # (not random, hap, etc.)
    elif (not self.CheckChromosome(align)): #{
      return False
    # filter by blocks
    #return self.CheckBlocks(align)
    return True
  #} end def

  def CheckContigSpan(self, align): #{
    contig_span = abs(align.qstart - align.qend) + 1
    cspan_fract = float(contig_span) / float(align.query_len)
    if (self.options.min_gap_check_len_fract > cspan_fract): #{
      ExtremeDebugMsg(self, "NO GAP CHECK: too short contig span: "
        "%i (%.2f)" % (contig_span, cspan_fract))
      return False
    #} end if
    return True
  #} end def

  def CheckUnalignedBases(self, align): #{
    num_unaligned = align.query_len - align.match - align.mismatch
    if (self.options.min_gap_size > num_unaligned): #{
      ExtremeDebugMsg(self, "NO GAP CHECK: too few unaligned bases")
      return False
    #} end if
    return True
  #} end def

  def CheckScoreFraction(self, align): #{
    if (self.max_score < align.score): #{
      self.max_score = align.score
    else:
      try: #{
        score_fract = float(align.score) / float(self.max_score)
      except ZeroDivisionError:
        raise GapFilterError("Error calculating score fraction! score: %i, "
          "max: %i\n  %s" % (align.score, self.max_score, AlignString(align)))
      #} end try
      if (self.options.min_score_fract > score_fract): #{
        ExtremeDebugMsg(self, "NO GAP CHECK: score less than max")
        return False
      #} end if
    #} end if
    # do not check alignments with not enough of the contig involved
    # in the alignment (based on score:length ratio)
    score_fract = float(align.score) / float(align.query_len)
    if (self.options.min_gap_check_score_fract > score_fract): #{
      ExtremeDebugMsg(self, "NO GAP CHECK: too little contig aligned")
      return False
    #} end if
    return True
  #} end def

  def CheckPercentIdentity(self, align): #{
    if (self.options.min_gap_check_pid >= align.identity): #{
      ExtremeDebugMsg(self, "NO GAP CHECK: pid too low")
      return False
    #} end if
    return True
  #} end def

  def CheckMitochondrial(self, align): #{
    # TODO: Remove unnecessary parentheses around "'M' == ...".
    if (self.options.no_mito and
        #("chrM" == align.target or "M" == align.target)): #{}
        ("M" == NormalizeChrID(align.target))): #{
      ExtremeDebugMsg(self, "NO GAP CHECK: mitochondrial")
      return False
    #} end if
    return True
  #} end def

  def CheckChromosome(self, align): #{
    #if (None == re.search(r"^(chr)?(\d+|[XY]|MT?)$", align.target)): #{}
    if (NonStandardChr(align.target)): #{
      ExtremeDebugMsg(self, "NO GAP CHECK: non-standard chromosome (%s)" %
        align.target)
      return False
    #} end if
    return True
  #} end def

  def CheckBlocks(self, align): #{
    # assume there are no gaps worth looking at
    #self.good_gap_found = False
    # get the blocks, as integers, in the correct order
    blocks = map(IntifyBlock, align.query_blocks)
    if ("-" == align.query_strand): #{
      blocks.reverse()
    #} end if
    ExtremeDebugMsg(self, "MIN: %i\n" % self.options.min_gap_size +
      "Num Blocks: %i Strand: %s " % (len(blocks), align.query_strand) +
      "Blocks: %s" % ",".join([BlockString(block) for block in blocks]))
    #} end if
    #gap_start = 1
    gap = CoordPairCls(1)
    for block in blocks: #{
      if ("-" == align.query_strand): #{
        block.reverse()
      #} end if
      #gap_end = block[0] - 1
      gap.SetMax(block[0] - 1)
      #ExtremeDebugMsg(self, "Gap: %i-%i Block: %i-%i" %
      #  (gap_start, gap_end, block[0], block[1]))
      ExtremeDebugMsg(self, "Gap: %s Block: %i-%i" % (gap, block[0], block[1]))
      # check the block order! the block should start after the previous block
      # and should start before it ends
      #if (block[0] < gap_start or block[1] < block[0]): #{}
      if (block[0] < gap.min or block[1] < block[0]): #{
        ExtremeDebugMsg(self,
          "gap_start: %i, block_start: %i, block_end: %i" %
          (gap.min, block[0], block[1]))
        raise GapFilterError("incorrect block ordering: %s" %
          ",".join([BlockString(block) for block in blocks]))
      #} end if
      #if (not self.CheckGap(gap_start, gap_end, align.query_len)): #{}
      if (not self.CheckGap(gap, align.query_len)): #{
        return False
      #} end if
      #gap_start = block[1] + 1
      gap.SetMin(block[1] + 1)
    #} end for
    # look for gap at the very end of the contig
    gap.SetMax(align.query_len)
    #ExtremeDebugMsg(self, "Last Gap: %i-%i" % (gap_start, align.query_len))
    ExtremeDebugMsg(self, "Last Gap: %s" % (gap))
    #if (not self.CheckGap(gap_start, align.query_len, align.query_len)): #{}
    if (not self.CheckGap(gap, align.query_len)): #{
      return False
    #} end if
    #if (not self.good_gap_found): #{
    if (not self.gaps_to_check): #{
      ExtremeDebugMsg(self, "NO GAP CHECK: no good gap found")
    #} end if
    #return self.good_gap_found
  #} end def

  #def CheckGap(self, gap_start, gap_end, query_len): #{}
  def CheckGap(self, ref_gap, query_len): #{
    #gapsize = (gap_end - gap_start) + 1
    #if (gapsize >= self.options.min_gap_size): #{}
    if (ref_gap.Span() >= self.options.min_gap_size): #{
      gap = ref_gap.Copy()
      #gap_seq = self.ctg_seq[gap_start-1:gap_end].lower()
      gap.seq = self.ctg_seq[gap.min-1:gap.max].lower()
      ExtremeDebugMsg(self, "Gap Start: %i Gap End: %i\nSequence: %s" %
        (gap.min, gap.max, gap.seq))
        #(gap_start, gap_end, gap_seq))
      # make sure that the gap does not include a single 'n'
      pattern = r"(^n[^n]\|[^n]n[^n]\|[^n]n$)"
      if (None != re.search(pattern, gap.seq.lower())): #{
        ExtremeDebugMsg(self, "NO GAP CHECK: single 'n' in gap sequence")
        return False
      #} end if
      # make sure the gap is not just T's or A's at the beginning
      # or end of the contig
      #if ((1 < gap_start and gap_end < query_len) or
      if ((1 < gap.min and gap.max < query_len) or
          (None == re.search(r"^a+$", gap.seq) and
           None == re.search(r"^t+$", gap.seq))):
        #self.good_gap_found = True
        ExtremeDebugMsg(self, 'Appending gap ({0}) to gaps_to_check!'.format(
          gap))
        self.gaps_to_check.append(gap)
      else:
        ExtremeDebugMsg(self, "Poly-A tail")
    else:
      ExtremeDebugMsg(self, "Gap too small: %i, Min: %i" %
        (ref_gap.Span(), self.options.min_gap_size))
        #(gapsize, self.options.min_gap_size))
    #} end if
    return True
  #} end def

  def get_gap_realign_command(self, curr_align, psl_str): #{
    ChooseBestTranscripts(curr_align,
      self.options.transcript_selection_buffer)
    olap_genes = GenesOverlapped(curr_align,
      self.options.add_gene_annotation)
    nearby_genes = NearbyGenes(curr_align, self.options.add_gene_annotation)
    gap_realign_args = [
      self.options.gap_realigner,
      "\"%s\"" % psl_str,
      self.ctg_seq,
      "\"%s NEARBY:%s\"" % (olap_genes, nearby_genes),
      self.gap_out_path,
      self.options.gap_config,
      "%i"   % self.options.min_gap_size,
      "%.2f" % self.options.min_gap_pid,
      "%.2f" % self.options.min_gap_fract,
      "MM:%i" % self.multi_mapped,
    ]
    if (self.options.gap_debug): #{
      gap_realign_args.append("-d")
    #} end if
    gap_realign_cmd = " ".join(gap_realign_args)
    ExtremeDebugMsg(self, "Gap realigner Command:\n    %s" %
      gap_realign_cmd)
    if (self.options.arg_max < len(gap_realign_cmd)): #{
      ExtremeDebugMsg(self, "WARNING: Cannot use gap realigner on contig "
        "%s: sequence is too long" % curr_align.query)
      self.seq_too_long = True
      return ""
    #} end if
    return gap_realign_cmd
  #} end def

  def internal_duplication_checking(self, curr_align): #{
    ExtremeDebugMsg(self, "Checking {0} gaps...".format(
      len(self.gaps_to_check)))
    for gap in self.gaps_to_check: #{
      ExtremeDebugMsg(self, '  Gap: {0} Min: {1} Max: {2} Sequence: {3}'.format(
        gap, gap.min, gap.max, gap.seq))
      match_before, toffset_before = self.check_before(gap)
      match_after, toffset_after = self.check_after(gap)
      if (match_before > match_after): #{
        ExtremeDebugMsg(self, '    Using match before gap...')
        gap.match_len = match_before
        gap.toffset = toffset_before
        gap.match_is_before_gap = True
      else:
        ExtremeDebugMsg(self, '    Using match after gap...')
        gap.match_len = match_after
        gap.toffset = toffset_after
        gap.match_is_before_gap = False
      #} end if
      match_fraction = float(gap.match_len) / float(gap.Span())
      ExtremeDebugMsg(self, '    Fraction of gap matched: {0}'.format(
          match_fraction))
      if (gap.match_len > self.options.min_gap_size and
          match_fraction > self.options.min_gap_fract): #{
        ExtremeDebugMsg(self, '    Saving matched gap!')
        self.gaps_with_dups.append(gap)
        ExtremeDebugMsg(self, '    {0} gaps saved.'.format(
          len(self.gaps_with_dups)))
      #} end if
    #} end for
    # clear out the gaps_to_check list in preparation for the next alignment
    self.gaps_to_check = list()
  #} end def

  def check_before(self, gap): #{
    before_seq_start = max(0, gap.min-(gap.Span()+MAX_INTERNAL_DUP_OFFSET+1))
    ExtremeDebugMsg(self, '    Checking before...\n    "{0}"'.format(
      self.ctg_seq[before_seq_start:gap.min-1]))
    gap_seq_end = gap.Span()
    before_seq_end = gap.min-1
    matches = [0 for i in xrange(min(MAX_INTERNAL_DUP_OFFSET, gap.Span()))]
    # only want offset to affect target, might use second offset for gap?
    for offset in xrange(len(matches)): #{
      longest_possible = min(gap.min-(offset+1), gap.Span())
      ExtremeDebugMsg(
          self, '    Offset: {0} Longest possible match: {1}'.format(
            offset, longest_possible))
      if (1 > longest_possible): #{
        break
      #} end if
      for match_len in range(longest_possible): #{
        matches[offset] = match_len
        gap_seq_start = gap_seq_end-(match_len+1)
        gap_seq = gap.seq[gap_seq_start:gap_seq_end].lower()
        before_seq_start = before_seq_end-(match_len+1)
        before_seq = self.ctg_seq[before_seq_start:before_seq_end].lower()
        #ExtremeDebugMsg(self, '      {0}\n      {1}'.format(
        #    gap_seq, before_seq))
        if (gap_seq != before_seq): #{
          #ExtremeDebugMsg(self, '      no match!')
          break
        #} end if
      #} end for
      ExtremeDebugMsg(self, '      match length: {0}'.format(matches[offset]))
      matches.append(matches[offset])
      # update for offset
      before_seq_end -= 1
    #} end for
    best_match = max(matches)
    return best_match, matches.index(best_match)
  #} end def

  def check_after(self, gap): #{
    after_seq_end = min(len(self.ctg_seq),
                        gap.max+(gap.Span()+MAX_INTERNAL_DUP_OFFSET))
    ExtremeDebugMsg(self, '    Checking after...\n    "{0}"'.format(
      self.ctg_seq[gap.max:after_seq_end]))
    gap_seq_start = 0
    after_seq_start = gap.max
    matches = [0 for i in xrange(min(MAX_INTERNAL_DUP_OFFSET, gap.Span()))]
    # only want offset to affect target, might use second offset for gap?
    for offset in xrange(len(matches)): #{
      longest_possible = min(len(self.ctg_seq)-(gap.max+offset), gap.Span())
      ExtremeDebugMsg(
          self, '    Offset: {0} Longest possible match: {1}'.format(
            offset, longest_possible))
      if (1 > longest_possible): #{
        break
      #} end if
      for match_len in range(longest_possible): #{
        matches[offset] = match_len
        gap_seq_end = gap_seq_start+match_len+1
        gap_seq = gap.seq[gap_seq_start:gap_seq_end].lower()
        after_seq_end = after_seq_start+match_len+1
        after_seq = self.ctg_seq[after_seq_start:after_seq_end].lower()
        #ExtremeDebugMsg(self, '      {0}\n      {1}'.format(
        #    gap_seq, after_seq))
        if (gap_seq != after_seq): #{
          #ExtremeDebugMsg(self, '      no match!')
          break
        #} end if
      #} end for
      ExtremeDebugMsg(self, '      match length: {0}'.format(matches[offset]))
      matches.append(matches[offset])
      # update for offset
      after_seq_start += 1
    #} end for
    best_match = max(matches)
    return best_match, matches.index(best_match)
  #} end def

  def create_gap_candidates(self, curr_align): #{
    ExtremeDebugMsg(self, 'Creating Gap Candidates...')
    ExtremeDebugMsg(
        self, 'There are {0} gaps representing potentially duplicated '
        'sequence; each will create a separate candidate.'.format(
            len(self.gaps_with_dups)))
    # Open gap candidate output file, in append mode
    gap_candidate_file = FileBoxCls(self.gap_out_path, 'a')
    for gap_with_dup in self.gaps_with_dups: #{
      gap_candidate = GapCandidateCls(curr_align, gap_with_dup,
                                      self.options.add_gene_annotation,
                                      self.ctg_seq)
      # check the multi-mapping status of the alignments
      gap_candidate.CheckMultiMapping()
      if self.options.add_gene_annotation: #{
        # choose the best guesses for transcript(s) the contig represents
        ChooseBestTranscripts(gap_candidate.align1,
          self.options.transcript_selection_buffer)
      #} end if
      # skip alignments involving non-standard chromosomes
      if (NonStandardChr(gap_candidate.align1.target)): #{
        DebugMsg(self, "Skipping non-standard chromosome for contig %s: %s" %
          (contig.id, gap_candidate.align1.target))
        continue
      #} end if
      gap_candidate_string = gap_candidate.Details()
      ExtremeDebugMsg(self, '  {0}'.format(gap_candidate_string))
      # Write gap candidate details to output file
      gap_candidate_file.WriteLine(gap_candidate_string)
      self.num_gapped_aligns += 1
    #} end for
    gap_candidate_file.close()
    # clear out the gaps_with_dups list in preparation for the next alignment
    self.gaps_with_dups = list()
  #} end def
#} end class

def BlockString(block): #{
  return "%i-%i" % tuple(sorted([block[0],block[1]]))
#} end def

#### EXCEPTION CLASSES ####
class GapFilterError(MyError): #{
  """Exception raised for errors encountered by the GapFilterCls class"""
  pass
#} end class
