#! /usr/bin/env python
"""
breakpoint_pair.py

Created by Lucas Swanson
Copyright (c) 2012 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

# import standard modules
import re
from utils.error import MyError

# import custom modules
from utils.general import StrJoin
from utils.messages import ErrMsg, LogMsg, DebugMsg, ExtremeDebugMsg
from common.breakpoint import SortBreakpoints, BreakpointPairsAreEqual
from parsers.tokenizer import TokenizerCls

# CONSTANTS

class BreakpointPairCls: #{
  def __init__(self, buffer, other_type="other"): #{
    #print "Creating bp pair object"
    self.buffer = buffer
    self.other_type = other_type
    #self.pair = None
    self.pairs = list()
    self.extreme_coords = (dict(),dict())
    self.barnacle_preds = list()
    self.barnacle_transcribed = 0
    self.other_preds = list()
    self.other_qualities = dict()
    self.other_transcribed = 0
    self.interchromosomal = False
  #} end def

  def AddBarnaclePrediction(self, b_member): #{
    new_pair = SortBreakpoints(b_member.breakpointA, b_member.breakpointB)
    #print "PAIR: ", self.pair
    #if (None == self.pair): #{
    #  #print "Setting pair to %s-%s" % (new_pair[0].ToString(),
    #  #  new_pair[1].ToString())
    #  self.pair = new_pair
    #elif (not BreakpointPairsAreEqual(self.pair, new_pair, self.buffer)): #{
    #  raise BreakpointPairError("Barnacle breakpoint does not match: %s, %s" %
    #    (BreakpointPairString(self.pair), BreakpointPairString(new_pair)))
    #} end if
    self.AddPair(new_pair, "Barnacle")
    self.barnacle_preds.append(b_member.IDString())
    new_transcribed = 0
    for region_id in ["A", "B"]: #{
      if (b_member.OverlapsGene(region_id)): #{
        new_transcribed += 1
      #} end if
    #} end for
    self.barnacle_transcribed = max(self.barnacle_transcribed, new_transcribed)
  #} end def

  def AddOtherPrediction(self, pred, type="Other"): #{
    new_pair = SortBreakpoints(pred.breakpointA, pred.breakpointB)
    #if (None == self.pair): #{
    #  self.pair = new_pair
    #elif (not BreakpointPairsAreEqual(self.pair, new_pair, self.buffer)): #{
    #  raise BreakpointPairError("Barnacle breakpoint does not match: %s, %s" %
    #    (BreakpointPairString(self.pair), BreakpointPairString(new_pair)))
    #} end if
    self.AddPair(new_pair, type)
    self.other_preds.append(pred.id)
    # use maximum quality of all "other" predictions
    if (hasattr(pred, "quality")): #{
      self.other_qualities[pred.id] = pred.quality
      #if (None == self.other_quality or
      #    self.other_quality < pred.quality): #{
      #  self.other_quality = pred.quality
      #} end if
    #} end if
  #} end def

  def UpdateOtherQuality(self, pred_id, new_quality): #{
    if (pred_id not in self.other_qualities): #{
      raise BreakpointPairError("No prediction with id %i in breakpoint "
        "collection: %s" % (pred_id, StrJoin(", ",
        sorted(self.other_qualities))))
    #} end if
    self.other_qualities[pred_id] = new_quality
  #} end def

  def SetBreakpointQuality(self): #{
    # use maximum quality of all "other" predictions
    if (0 < len(self.other_qualities)): #{
      self.other_quality = max(self.other_qualities.values())
    #} end if
  #} end def

  def AddPair(self, new_pair, type): #{
    if (0 < len(self.pairs)): #{
      match_found = False
      for pair in self.pairs: #{
        if (BreakpointPairsAreEqual(pair, new_pair, self.buffer)): #{
          match_found = True
          break
        #} end if
      #} end for
      if (not match_found): #{
        raise BreakpointPairError("%s breakpoint %s does not match any of %s" %
          (type, BreakpointPairString(new_pair), ", ".join((
          BreakpointPairString(pair) for pair in self.pairs))))
      #} end if
    #} end if
    self.pairs.append(new_pair)
    self.UpdateExtremes(0, new_pair[0])
    self.UpdateExtremes(1, new_pair[1])
    if (new_pair[0].chr != new_pair[1].chr): #{
      self.interchromosomal = True
    #} end if
  #} end def

  def UpdateExtremes(self, index, breakpoint): #{
    if ("low" not in self.extreme_coords[index] or
        self.extreme_coords[index]["low"] > breakpoint.coord): #{
      self.extreme_coords[index]["low"] = breakpoint.coord
    #} end if
    if ("high" not in self.extreme_coords[index] or
        self.extreme_coords[index]["high"] < breakpoint.coord): #{
      self.extreme_coords[index]["high"] = breakpoint.coord
    #} end if
  #} end def

  def Common(self): #{
    return (0 < len(self.barnacle_preds) and 0 < len(self.other_preds))
  #} end def

  def BarnacleOnly(self): #{
    return (0 < len(self.barnacle_preds) and 0 == len(self.other_preds))
  #} end def

  def OtherOnly(self): #{
    return (0 == len(self.barnacle_preds) and 0 < len(self.other_preds))
  #} end def

  def SetComparisonResult(self): #{
    if (self.Common()): #{
      self.ltype = "common"
    elif (self.BarnacleOnly()):
      self.ltype = "barnacle"
    elif (self.OtherOnly()):
      self.ltype = self.other_type
    else:
      raise BreakpointPairError("Unresolved breakpoint comparison! %s" %
        self.ToString())
    #} end if
  #} end def

  def ToString(self): #{
    data = list()
    #if (None == self.pair): #{}
    if (0 == len(self.pairs)): #{
      pair_data = "Pair data not set!"
    else:
      #pair_data = "%s-%s" % (self.pair[0].ToString(), self.pair[1].ToString())
      breakpoint_strs = ["",""]
      #ErrMsg("  Creating breakpoint strings for breakpoint pair:")
      for index in range(2): #{
        breakpoint_strs[index] = "chr%s:%i-%i(%s)" % (
          self.pairs[0][index].chr, self.extreme_coords[index]["low"],
          self.extreme_coords[index]["high"], self.pairs[0][index].dir)
        #ErrMsg("    %i: %s" % (index, breakpoint_strs[index]))
      #} end for
      pair_data = " ".join(breakpoint_strs)
    #} end if
    data.append(pair_data)
    #if (hasattr(self, "primary")): #{
    #  data.append("%i" % self.primary)
    #} end if
    flag_list = list()
    for flag in ["primary", "repeats"]: #{
      if (hasattr(self, flag) and None != getattr(self, flag)): #{
        flag_list.append("%i" % getattr(self, flag))
      else:
        flag_list.append("x")
      #} end if
    #} end for
    common_flags = StrJoin(",", flag_list)
    data.append(common_flags)
    if (0 == len(self.barnacle_preds)): #{
      barnacle_str = "None"
      barnacle_flags = "x"
    else:
      barnacle_str = ",".join(sorted(self.barnacle_preds))
      flag_list = list()
      for flag in ["wide_spread", "local_event", "barnacle_transcribed",
          "multi_maps", "full_repeat"]: #{
        if (hasattr(self, flag) and None != getattr(self, flag)): #{
          flag_list.append("%i" % getattr(self, flag))
        else:
          flag_list.append("x")
        #} end if
      #} end for
      barnacle_flags = StrJoin(",", flag_list)
    #} end if
    data.extend((barnacle_str, barnacle_flags))
    #barnacle_str += " %i" % self.barnacle_transcribed
    #if (hasattr(self, "wide_spread") and self.wide_spread): #{
    #  barnacle_str += " wide"
    #else:
    #  barnacle_str += " tight"
    #} end if
    if (0 == len(self.other_preds)): #{
      other_str = "None"
      other_flags = "x"
    else:
      other_str = StrJoin(",", sorted(self.other_preds))
      flag_list = list()
      for flag in ["other_quality", "other_transcribed", "same_gene"]: #{
        if (hasattr(self, flag) and None != getattr(self, flag)): #{
          flag_list.append("%i" % getattr(self, flag))
        else:
          flag_list.append("x")
        #} end if
      #} end for
      other_flags = StrJoin(",", flag_list)
    #} end if
    #other_str += " %i" % self.other_transcribed
    #if (None == self.other_quality): #{
    #  qual_str = "None"
    #else:
    #  qual_str = "%i" % (self.other_quality)
    #} end if
    data.extend((other_str, other_flags))
    #repeat_list = list()
    #for rpt_attr in ["repeats", "full_repeat"]: #{
    #  if (hasattr(self, rpt_attr) and None != getattr(self, rpt_attr)): #{
    #    repeat_list.append("%i" % getattr(self, rpt_attr))
    #  else:
    #    repeat_list.append("x")
    #  #} end if
    #} end for
    #rpt_str = StrJoin(",", repeat_list)
    #data.append(rpt_str)
    #if (hasattr(self, "repeats")): #{
    #  data.append("%i" % self.repeats)
    #} end if
    #return "\t".join([pair_data, barnacle_str, other_str, qual_str])
    return StrJoin("\t", data)
  #} end def

  def ParseCompareLine(self, line): #{
    tokenizer = TokenizerCls(line)
    # parse locus and breakpoint ids
    full_id = tokenizer.next()
    id_match = re.search(r"^(?P<locus>[0-9]+)\.(?P<breakpoint>[0-9]+)$",
        full_id)
    (self.locus_id, self.id) = map(int, id_match.groups)
    # parse breakpoints
    self.breakpoint_A = tokenizer.next()
    self.breakpoint_B = tokenizer.next()
    # parse common flags
    common_flags = tokenizer.next().split(",")
    for flag, value in zip(("primary", "repeats"), common_flags): #{
      setattr(self, flag, bool(int(value)))
    #} end for
    # parse Barnacle predictions
    pred_list_str = tokenizer.next()
    if ("none" == pred_list_str.lower()): #{
      self.barnacle_preds = list()
    else:
      self.barnacle_preds = pred_list_str.split(",")
    #} end if
    # parse Barnacle flags
    if (0 == len(self.barnacle_preds)): #{
      tokenizer.next()
    else:
      barnacle_flags = tokenizer.next().split(",")
      for flag, value in zip(("wide_spread", "local_event", "multi_maps",
          "full_repeat"), barnacle_flags[:2]+barnacle_flags[3:]): #{
        setattr(self, flag, bool(int(value)))
      #} end for
      self.barnacle_transcribed = int(barnacle_flags[2])
    #} end if
    # parse other predictions
    pred_list_str = tokenizer.next()
    if ("none" == pred_list_str.lower()): #{
      self.other_preds = list()
    else:
      self.other_preds = pred_list_str.split(",")
    #} end if
    if (0 == len(self.other_preds)): #{
      tokenizer.next()
    else:
      other_flags = tokenizer.next().split(",")
      for flag, value in zip(("other_transcribed", "same_gene"),
          other_flags[1:]): #{
        setattr(self, flag, bool(int(value)))
      #} end for
      self.other_quality = int(other_flags[0])
    #} end if
  #} end def
#} end class

#### EXCEPTION CLASSES ####
class BreakpointPairError(MyError): #{
  pass
#} end class
