#! /usr/bin/env python

# import standard modules
from optparse import OptionParser, OptionGroup
from glob import glob, iglob
import os, re, sys, time, traceback

# import custom modules
from version import VERSION
from utils.log import GetLogPath, CloseLogFile
from utils.error import MyError
from utils.general import (SetupMainClass, TimeSpent, WriteCommand, StrJoin)
from utils.messages import ErrMsg, LogMsg, DebugMsg, ExtremeDebugMsg
from utils.files_paths import (CheckFilePath, CheckDirPath, EnsureAbsPath,
  FileBoxCls, CheckNewFilePath)
from parsers.candidate_group_parser import CandidateGroupParserCls
from parsers.tokenizer import TokenizerCls
from parsers.fasta import FastaFileCls

# CONSTANTS
ES_SUCCESS = 0
ES_RUN_ERR = 1
ES_OPT_ERR = 2
ES_PATH_ERR = 3
ES_EXCEPTION = 4
MSG_SUCCESS = "SUCCESS"
MSG_FAIL = "FAIL"
EXPLANATIONS = ["predicted", "no_gene", "misc_pass", "r2c", "pid",
  "too_many_groups", "p2g", "ctg_olap", "ctg_rep", "rna", "topology",
  "mitochondrial", "misc_fail", "readthrough", "no_aligns",
  "missing_aligns", "bad_aligns", "other"]

class MainClassCls: #{
  def __init__(self, options): #{
    SetupMainClass(self, options)
    # defuse predictions, keyed by contig
    self.dpreds = dict()
    self.unexplained_dpreds = set()
    self.unassembled = set()
    self.filtered_assembly = set()
    # contigs, keyed by defuse prediction cluster id
    self.contigs = dict()
    self.unexplained_contigs = set()
    self.explained_contigs = dict((type, set()) for type in EXPLANATIONS)
    self.ctg_counts = dict()
    self.dpred_counts = dict((type, 0) for type in EXPLANATIONS)
    # contigs, keyed by Barnacle candidate identification job number
    self.job2ctg = dict()
  #} end def

  def __del__(self): #{
    CloseLogFile(self)
  #} end def

  def Run(self): #{
    LogMsg(self, "Checking Defuse specific predictions...")
    start = time.time()
    # load the Defuse-to-contig alignment results
    self.LoadDefuseToContigAlignments()
    # check for contigs with fusion predictions
    self.CheckForPredictions()
    # check for contigs with filter-passing candidates
    self.CheckPassingCandidates()
    # check for contigs with filter-failing candidates
    self.CheckFailingCandidates()
    # check for non-candidate contigs
    self.CheckNonCandidates()
    # handle any remaining unexplained contigs
    self.HandleUnexplained()
    # write output
    self.WriteOutput()
    # write counts
    self.WriteCounts()
    LogMsg(self, "Total time spent checking Defuse specific predictions: %s" %
      TimeSpent(start))
  #} end def

  def LoadDefuseToContigAlignments(self): #{
    LogMsg(self, "Loading Defuse to contig alignment results...")
    start = time.time()
    infile = FileBoxCls(self.options.def2ctg_align_notes_path, "r",
      "cannot read Defuse-to-contig alignment results summary file")
    self.dpred_counts["total"] = 0
    for line in infile: #{
      # skip the header
      if (line.lower().startswith("cluster_id")): #{
        continue
      #} end if
      self.dpred_counts["total"] += 1
      tokenizer = TokenizerCls(line)
      cluster_id = int(tokenizer.next())
      DebugMsg(self, "  Loading Defuse prediction %i" % cluster_id)
      partners = tokenizer.next()
      coverage = int(tokenizer.next())
      align_type = tokenizer.next()
      if ("no_align" == align_type): #{
        self.unassembled.add(cluster_id)
        ExtremeDebugMsg(self, "    Unassembled")
        continue
      #} end if
      if ("_realign" in align_type): #{
        self.filtered_assembly.add(cluster_id)
        ExtremeDebugMsg(self, "    Assembly-filtered")
        continue
      #} end if
      contigs = tokenizer.next().split(",")
      ExtremeDebugMsg(self, "    Contigs: %s" % ", ".join(contigs))
      for contig in contigs: #{
        if contig not in self.dpreds: #{
          self.dpreds[contig] = list()
        #} end if
        self.dpreds[contig].append(cluster_id)
      #} end for
      self.contigs[cluster_id] = contigs
      self.unexplained_contigs.update(contigs)
      self.unexplained_dpreds.add(cluster_id)
    #} end for
    #DebugMsg(self, "\n".join(["  %s: %s" % (contig,
    #  StrJoin(", ", sorted(self.dpreds[contig]))) for contig in
    #  sorted(self.dpreds)]))
    #DebugMsg(self, "Unexplained: %s" %
    #  ", ".join(sorted(self.unexplained_contigs)))
    self.ctg_counts["total"] = len(self.unexplained_contigs)
    LogMsg(self, "  %i Defuse predictions to explain (%i contigs)" %
      (self.dpred_counts["total"], self.ctg_counts["total"]))
    LogMsg(self, "Time spent loading Defuse to contig alignment results: %s" %
      TimeSpent(start))
  #} end def

  def CheckForPredictions(self): #{
    self.ctg_counts["pred"] = 0
    if (0 == len(self.unexplained_contigs)): #{
      return
    #} end if
    LogMsg(self, "Checking Barnacle predictions for missed commons...")
    start = time.time()
    infile = CandidateGroupParserCls(self.options.barnacle_fus_path)
    predicted_ctgs = set()
    for group in infile: #{
      for member in group.members: #{
        ctg_id = member.contig_info.id
        if (ctg_id in self.unexplained_contigs): #{
          predicted_ctgs.add(ctg_id)
        #} end if
      #} end for
    #} end for
    for ctg_id in predicted_ctgs: #{
      self.ctg_counts["pred"] += 1
      self.explained_contigs['predicted'].add(ctg_id)
      self.unexplained_contigs.remove(ctg_id)
    #} end for
    LogMsg(self, "  Found %i contigs, %i remain unexplained" %
      (self.ctg_counts["pred"], len(self.unexplained_contigs)))
    LogMsg(self, "Time spent checking Barnacle predictions: %s" %
      TimeSpent(start))
  #} end def

  def CheckPassingCandidates(self): #{
    self.ctg_counts["pass"] = 0
    if (0 == len(self.unexplained_contigs)): #{
      return
    #} end if
    LogMsg(self, "Checking filter-passing candidates...")
    start = time.time()
    infile = CandidateGroupParserCls(self.options.barnacle_pass_path)
    outfile = FileBoxCls(self.options.pass_outpath, "w",
      "cannot create filter-passing candidates output file")
    # passing_contigs[ctg_id] = exp
    passing_contigs = dict()
    for group in infile: #{
      #DebugMsg(self, "  Checking group %i" % group.id)
      for member in group.members: #{
        ctg_id = member.contig_info.id
        #ExtremeDebugMsg(self, "    Checking member %s, contig: %s" %
        #  (member.candidate_id, ctg_id))
        if (ctg_id in self.unexplained_contigs): #{
          if (NoGene(member)): #{
            passing_contigs[ctg_id] = "no_gene"
          elif (ctg_id not in passing_contigs):
            passing_contigs[ctg_id] = "misc_pass"
          #} end if
          outfile.WriteLine(member.DataString())
        #} end if
      #} end for
    #} end for
    self.ctg_counts["pass"] = len(passing_contigs)
    for (ctg_id, exp) in passing_contigs.iteritems(): #{
      self.explained_contigs[exp].add(ctg_id)
      self.unexplained_contigs.remove(ctg_id)
    #} end for
    outfile.close()
    LogMsg(self, "  Found %i contigs, %i remain unexplained" %
      (self.ctg_counts["pass"], len(self.unexplained_contigs)))
    LogMsg(self, "Time spent checking filter-passing candidates: %s" %
      TimeSpent(start))
  #} end def

  def CheckFailingCandidates(self): #{
    self.ctg_counts["fail"] = 0
    if (0 == len(self.unexplained_contigs)): #{
      return
    #} end if
    LogMsg(self, "Checking filter-failing candidates...")
    start = time.time()
    infile = CandidateGroupParserCls(self.options.barnacle_fail_path)
    outfile = FileBoxCls(self.options.fail_outpath, "w",
      "cannot create filter-failing candidates output file")
    # failing_contigs[ctg_id] = fail_reasons
    failing_contigs = dict()
    for group in infile: #{
      target_members = list()
      for member in group.members: #{
        ctg_id = member.contig_info.id
        if (ctg_id in self.unexplained_contigs): #{
          if (ctg_id not in failing_contigs): #{
            failing_contigs[ctg_id] = set()
          #} end if
          failing_contigs[ctg_id].update(member.fail_reasons)
          target_members.append(member)
        #} end if
      #} end for
      # write failing group line and members with target contigs
      if (0 < len(target_members)): #{
        outfile.WriteLine(group.DataString())
        for member in target_members: #{
          outfile.WriteLine(member.DataString())
          ctg_id = member.contig_info.id
          # use group fail reasons if member had none
          if (0 == len(failing_contigs[ctg_id])): #{
            failing_contigs[ctg_id].update(group.fail_reasons)
          #} end if
        #} end for
      #} end if
    #} end for
    self.ctg_counts["fail"] = len(failing_contigs)
    for (ctg_id, fail_reasons) in failing_contigs.iteritems(): #{
      exp = GetFailingCandidateExplanation(fail_reasons)
      self.explained_contigs[exp].add(ctg_id)
      self.unexplained_contigs.remove(ctg_id)
    #} end for
    outfile.close()
    LogMsg(self, "  Found %i contigs, %i remain unexplained" %
      (self.ctg_counts["fail"], len(self.unexplained_contigs)))
    LogMsg(self, "Time spent checking filter-failing candidates: %s" %
      TimeSpent(start))
  #} end def

  def CheckNonCandidates(self): #{
    self.ctg_counts["nonc"] = 0
    if (0 == len(self.unexplained_contigs)): #{
      return
    #} end if
    LogMsg(self, "Checking non-candidates...")
    start = time.time()
    # load the job-to-contig map
    self.LoadJobToContigMap()
    # process the log file for every job involving unexplained contigs
    LogMsg(self, "  Processing job logs...")
    for job in sorted(self.job2ctg): #{
      self.ProcessJobLog(job, self.job2ctg[job])
    #} end for
    LogMsg(self, "  Found %i non-candidate contigs, %i remain unexplained" %
      (self.ctg_counts["nonc"], len(self.unexplained_contigs)))
    LogMsg(self, "Time spent checking non-candidates: %s" %
      TimeSpent(start))
  #} end def

  def LoadJobToContigMap(self): #{
    if (not os.path.exists(self.options.job2ctg_path) or
        self.options.regen_job2ctg): #{
      self.CreateJobToContigMapFile()
      return
    #} end if
    LogMsg(self, "  Loading job-to-contig map...")
    start = time.time()
    mapfile = FileBoxCls(self.options.job2ctg_path, "r",
      "cannot open job-to-contig map file")
    for line in mapfile: #{
      # skip the header
      if ("FILE_NUMBER CONTIG" == line): #{
        continue
      #} end if
      (job, ctg) = line.split(" ",1)
      if (job not in self.job2ctg): #{
        self.job2ctg[job] = list()
      #} end if
      self.job2ctg[job].append(ctg)
    #} end for
    mapfile.Close()
    LogMsg(self, "  Time spent loading job-to-contig map: %s" %
      TimeSpent(start))
  #} end def

  def CreateJobToContigMapFile(self): #{
    LogMsg(self, "  Creating job-to-contig map...")
    start = time.time()
    mapfile = FileBoxCls(self.options.job2ctg_path, "w",
      "cannot create job-to-contig map file")
    job_template = os.path.join(self.options.ta_cluster_dir, "seq.*.fa")
    count = 25
    for jobpath in iglob(job_template): #{
      if (25 <= count): #{
        DebugMsg(self, "  Checking file: %s" % jobpath)
        count = 0
      #} end if
      count += 1
      jobmatch = re.search(r".*seq\.(?P<num>\d+)\.fa", jobpath)
      job = jobmatch.group("num")
      jobfile = FastaFileCls(jobpath)
      for seq_obj in jobfile: #{
        if (seq_obj.id in self.unexplained_contigs): #{
          if (job not in self.job2ctg): #{
            self.job2ctg[job] = list()
          #} end if
          self.job2ctg[job].append(seq_obj.id)
          mapfile.WriteLine("%s %s" % (job, seq_obj.id))
        #} end if
      #} end for
      jobfile.Close()
    #} end for
    mapfile.Close()
    LogMsg(self, "  Time spent creating job-to-contig map: %s" %
      TimeSpent(start))
  #} end def

  def ProcessJobLog(self, jobnum, ctg_list): #{
    ExtremeDebugMsg(self, "    Processing job %s, contigs: %s" %
        (jobnum, ", ".join(ctg_list)))
    log_template = os.path.join(self.options.cid_cluster_dir,
      "%s-cid.%s.sh.e*" % (self.options.lib, jobnum))
    logpath = sorted(glob(log_template))[-1]
    logfile = FileBoxCls(logpath, "r", "cannot open cid log file")
    curr_ctg = None
    for line in logfile: #{
      # check whether line marks the start of processing a new contig
      match = re.search(r"^\d+\) (?P<ctg>k\d+[:_-]\d+)$", line)
      if (None != match): #{
        curr_ctg = match.group("ctg")
        continue
      #} end if
      # check whether the line marks a contig with no alignments
      match = re.search(r"(?P<ctg>k\d+[:_-]\d+) not present in align", line)
      if (None != match): #{
        if (match.group("ctg") in ctg_list and
            match.group("ctg") in self.unexplained_contigs): #{
          self.explained_contigs["no_aligns"].add(match.group("ctg"))
          self.unexplained_contigs.remove(match.group("ctg"))
          ExtremeDebugMsg(self, "      Found contig with no alignments")
          self.ctg_counts["nonc"] += 1
        #} end if
        continue
      #} end if
      # if the current contig is not a target contig,
      #   the line does not matter
      if (None == curr_ctg or curr_ctg not in ctg_list): #{
        continue
      #} end if
      # check for readthrough-like alignments
      if (line.lower().startswith("perfect align") or
          line.lower().startswith("single align")): #{
        self.explained_contigs["readthrough"].add(curr_ctg)
        self.unexplained_contigs.remove(curr_ctg)
        ExtremeDebugMsg(self, "      Found contig with single alignment")
        self.ctg_counts["nonc"] += 1
        continue
      #} end if
      # check for contigs with only bad alignments
      if (line.lower().startswith("no partial align")): #{
        self.explained_contigs["bad_aligns"].add(curr_ctg)
        self.unexplained_contigs.remove(curr_ctg)
        ExtremeDebugMsg(self, "      Found contig with only bad alignments")
        self.ctg_counts["nonc"] += 1
        continue
      #} end if
      match = re.search(r"(Found )?1 (align(ment)? groups|best alignments)", line)
      if (None != match): #{
        self.explained_contigs["missing_aligns"].add(curr_ctg)
        self.unexplained_contigs.remove(curr_ctg)
        ExtremeDebugMsg(self, "      Found contig with missing alignments")
        self.ctg_counts["nonc"] += 1
        continue
      #} end if
      LogMsg(self, "Unprocessed line: %s" % line)
    #} end for
    logfile.Close()
  #} end def

  def HandleUnexplained(self): #{
    self.ctg_counts["other"] = 0
    for contig in sorted(self.unexplained_contigs): #{
      self.explained_contigs["other"].add(contig)
      self.unexplained_contigs.remove(contig)
      self.ctg_counts["other"] += 1
    #} end for
    LogMsg(self, "%i contigs remain unexplained" % self.ctg_counts["other"])
  #} end def

  def WriteOutput(self): #{
    LogMsg(self, "Writing output...")
    start = time.time()
    outfile = FileBoxCls(self.options.main_outpath, "w",
      "cannot create main output file")
    for dpred in sorted(self.unassembled): #{
      outfile.WriteLine("%i\tnot_assembled\tnone" % dpred)
    #} end for
    for dpred in sorted(self.filtered_assembly): #{
      outfile.WriteLine("%i\tassembly_filtered\tnone" % dpred)
    #} end for
    for type in EXPLANATIONS: #{
      for contig in sorted(self.explained_contigs[type]): #{
        for dpred in sorted(self.dpreds[contig]): #{
          outfile.WriteLine("%i\t%s\t%s" % (dpred, type, contig))
          if (dpred in self.unexplained_dpreds): #{
            self.dpred_counts[type] += 1
            self.unexplained_dpreds.remove(dpred)
          #} end if
        #} end for
      #} end for
    #} end for
    outfile.close()
    LogMsg(self, "Time spent writing output: %s" %
      TimeSpent(start))
  #} end def

  def WriteCounts(self): #{
    countsfile = FileBoxCls(self.options.counts_path, "w",
        "cannot create counts file")
    countsfile.WriteLine("\n".join([
      "Number of predictions: %i (%i contigs)" % (self.dpred_counts["total"],
        self.ctg_counts["total"]),
      "-"*20,
      "Number not assembled: %i" % len(self.unassembled),
      "Number assembled but filtered: %i" % len(self.filtered_assembly),
      "-"*20,
      "Number aligning to Barnacle fusion predictions: %i (%i contigs)" %
        (self.dpred_counts["predicted"], self.ctg_counts["pred"]),
      "-"*20,
      "(%i contigs have filter-passing Barnacle candidates)" %
        self.ctg_counts["pass"],
      "Defuse predictions with non-exonic Barnacle candidates: %i" %
        self.dpred_counts["no_gene"],
      "Defuse predictions with other filter-passing Barnacle candidates: %i" %
        self.dpred_counts["misc_pass"],
      "-"*20,
      "(%i contigs have filter-failing Barnacle candidates)" %
        self.ctg_counts["fail"],
      "Defuse predictions with low read-to-contig Barnacle candidates: %i" %
        self.dpred_counts["r2c"],
      "Defuse predictions with low percent identity Barnacle candidates: %i" %
        self.dpred_counts["pid"],
      "Defuse predictions with highly ambiguous Barnacle candidates: %i" %
        self.dpred_counts["too_many_groups"],
      "Defuse predictions with low read-to-genome Barnacle candidates: %i" %
        self.dpred_counts["p2g"],
      "Defuse predictions with high contig-overlap Barnacle candidates: %i" %
        self.dpred_counts["ctg_olap"],
      "Defuse predictions with low contig-representation Barnacle "
        "candidates: %i" % self.dpred_counts["ctg_rep"],
      "Defuse predictions with structural RNA Barnacle candidates: %i" %
        self.dpred_counts["rna"],
      "Defuse predictions with ignored topology Barnacle candidates: %i" %
        self.dpred_counts["topology"],
      "Defuse predictions with mitochondrial Barnacle candidates: %i" %
        self.dpred_counts["mitochondrial"],
      "Defuse predictions with other filter-failing Barnacle candidates: %i" %
        self.dpred_counts["misc_fail"],
      "-"*20,
      "(%i non-candidate contigs)" % self.ctg_counts["nonc"],
      "Defuse predictions with readthrough alignments: %i" %
        self.dpred_counts["readthrough"],
      "Defuse predictions with no Trans-ABySS alignments: %i" %
        self.dpred_counts["no_aligns"],
      "Defuse predictions with a missing alignment or bad grouping: %i" %
        self.dpred_counts["missing_aligns"],
      "Defuse predictions with no good Trans-ABySS alignments: %i" %
        self.dpred_counts["bad_aligns"],
      "-"*20,
      "Remaining uncategorized Defuse predictions: %i (%i contigs)" %
        (self.dpred_counts["other"], self.ctg_counts["other"]),
    ]))
  #} end def
#} end class

def NoGene(member): #{
  for (main, nearby) in zip(member.main_genes, member.nearby_genes): #{
    if (0 == len(main) and 0 == len(nearby)): #{
      return True
    #} end if
  #} end for
  return False
#} end def

def GetFailingCandidateExplanation(fail_reasons): #{
  if ("READS_TO_CONTIG" in fail_reasons): #{
    return "r2c"
  #} end if
  if ("PERCENT_IDENTITY" in fail_reasons): #{
    return "pid"
  #} end if
  if ("TOO_MANY_GROUPS" in fail_reasons): #{
    return "too_many_groups"
  #} end if
  if ("PAIR_SUPPORT" in fail_reasons): #{
    return "p2g"
  #} end if
  if ("CONTIG_OVERLAP" in fail_reasons): #{
    return "ctg_olap"
  #} end if
  if ("CONTIG_REPRESENTED" in fail_reasons): #{
    return "ctg_rep"
  #} end if
  if ("RNA_REPEAT" in fail_reasons): #{
    return "rna"
  #} end if
  if ("IGNORED_TOPOLOGY" in fail_reasons): #{
    return "topology"
  #} end if
  if ("MITOCHONDRIAL" in fail_reasons): #{
    return "mitochondrial"
  #} end if
  return "misc_fail"
#} end def

#### EXCEPTION CLASSES ####
class MainClassError(MyError): #{
  pass
#} end class

def SetupOptionsParser(): #{
  description_string = ("Determine why Defuse predictions are not present "
    "in Barnacle predictions")
  args = [ "LIB", "BARNACLE_RESULTS_DIR", ]
  usage_string = "%prog " + " ".join(args) + " [ OPTIONS ]"
  parser = OptionParser(description=description_string,
                        version="%prog " + VERSION,
                        usage=usage_string)
  parser.num_args = len(args)
  dir_group = OptionGroup(parser, "Input File Directory Options")
  dir_group.add_option("--filt-num",
    type="int",
    help="The number at the beginning of the Barnacle filtering "
      "directory name. [default: %default]")
  dir_group.add_option("--pred-num",
    type="int",
    help="The number at the beginning of the Barnacle predictions "
      "directory name. [default: %default]")
  dir_group.add_option("--comp-num",
    type="int",
    help="The number at the beginning of the prediction comparisons "
      "directory name. [default: %default]")
  parser.add_option_group(dir_group)
  misc_group = OptionGroup(parser, "Miscellaneous Options")
  misc_group.add_option("--regen-job2ctg",
      action="store_true",
      help="Force regeneration of the job-to-contig mappings file.")
  misc_group.add_option("-f", "--force",
      action="store_true",
      help="Force filtering to take place, even if the output "
           "directory already exists.")
  misc_group.add_option("-d", "--debug",
      action="store_true",
      help="Print debug information while the program runs.")
  misc_group.add_option("--extreme-debug",
      action="store_true", dest="extreme_debug",
      help="Print extremely in-depth debug information while "
        "the program runs. Not recommended for large jobs.")
  parser.add_option_group(misc_group)
  parser.set_defaults(filt_num=8,
                      pred_num=9,
                      comp_num=10,
                      regen_job2ctg=False,
                      force=False,
                      debug=False,
                      extreme_debug=False)
  return parser
#} end def

def CheckPaths(options): #{
  opts_good = True
  if (not opts_good): #{
    ErrMsg("bad option") #TODO
    opts_good = False
  #} end if
  path_errors = list()
  CheckDirPath(options.base_dir, "Barnacle results", path_errors)
  options.ta_cluster_dir = os.path.join(os.path.dirname(os.path.dirname(
    options.base_dir.rstrip("/"))), "merge", "cluster",
    "%s-contigs" % options.lib, "input")
  CheckDirPath(options.ta_cluster_dir, "Trans-ABySS contig-to-genome "
    "alignments cluster", path_errors)
  options.cid_cluster_dir = os.path.join(options.base_dir, "cluster_cid")
  CheckDirPath(options.cid_cluster_dir, "Barnacle candidate identification "
    "cluster", path_errors)
  options.filt_dir = os.path.join(options.base_dir, "%i_filtered" %
    options.filt_num)
  CheckDirPath(options.filt_dir, "Barnacle filtering", path_errors)
  options.pred_dir = os.path.join(options.base_dir, "%i_predicted_events" %
    options.pred_num)
  CheckDirPath(options.pred_dir, "Barnacle predictions", path_errors)
  options.comp_dir = os.path.join(options.base_dir, "%i_defuse_comparison" %
    options.comp_num)
  CheckDirPath(options.comp_dir, "comparison", path_errors)
  options.main_outpath = os.path.join(options.comp_dir, "defuse_check.txt")
  options.pass_outpath = os.path.join(options.comp_dir, "defuse_check.pass")
  options.fail_outpath = os.path.join(options.comp_dir, "defuse_check.fail")
  options.counts_path = os.path.join(options.comp_dir,
    "defuse_check.counts.txt")
  options.job2ctg_path = os.path.join(options.comp_dir, "job2ctg_map.txt")
  if (not options.force): #{
    CheckNewFilePath(options.main_outpath, "main output", path_errors)
    CheckNewFilePath(options.pass_outpath, "filter-passing candidates "
      "output", path_errors)
    CheckNewFilePath(options.fail_outpath, "filter-passing candidates "
      "output", path_errors)
    CheckNewFilePath(options.counts_path, "main output", path_errors)
  #} end if
  options.def2ctg_align_notes_path = os.path.join(options.comp_dir,
    "defuse_align_results.txt")
  CheckFilePath(options.def2ctg_align_notes_path, "Defuse sequence to "
    "assembled contigs alignment results summary", path_errors)
  options.barnacle_fus_path = os.path.join(options.pred_dir,
    "%s.barnacle.fus.all" % options.lib)
  CheckFilePath(options.barnacle_fus_path, "Barnacle fusion predictions",
    path_errors)
  options.barnacle_pass_path = os.path.join(options.filt_dir,
    "%s.barnacle.pass" % options.lib)
  CheckFilePath(options.barnacle_pass_path, "filter-passing Barnacle "
    "candidates", path_errors)
  options.barnacle_fail_path = os.path.join(options.filt_dir,
    "%s.barnacle.fail" % options.lib)
  CheckFilePath(options.barnacle_fail_path, "filter-failing Barnacle "
    "candidates", path_errors)
  if (opts_good and 0 == len(path_errors)): #{
    # get the log file name
    options.log_file_name = GetLogPath(options.lib,
      "check_defuse_specific", options.comp_dir)
  #} end if
  if (0 < len(path_errors)): #{
    ErrMsg("Errors in input arguments:\n  %s" % "\n  ".join(path_errors))
  #} end if
  # the paths are good if there are no path errors and no conflicting options
  return (opts_good and 0 == len(path_errors))
#} end def

def Main(): #{
  # get options and arguments
  parser = SetupOptionsParser()
  (options, args) = parser.parse_args()
  # if the right number of args was used
  if (parser.num_args == len(args)): #{
    options.lib      = args[0]
    options.base_dir = EnsureAbsPath(args[1])
    if (CheckPaths(options)): #{
      try: #{
        main_class_object = MainClassCls(options)
        WriteCommand(main_class_object, sys.argv)
        main_class_object.Run()
      except (MyError), e:
        ErrMsg("ERROR while RUNNING MAIN FUNCTION:\n  %s" % e) #TODO
        return ES_RUN_ERR
      #} end try
    else:
      return ES_PATH_ERR
    #} end if
  else:
    parser.error("you must specify a library name (LIB); and the path to a "
      "Barnacle data file for that library (BARNACLE_FILE).")
    return ES_OPT_ERR
  #} end if
  return ES_SUCCESS
#} end def

if __name__ == '__main__': #{
  try: #{
    exit_status = Main()
  except Exception, e:
    traceback.print_exc()
    exit_status = ES_EXCEPTION
  #} end try
  if (ES_SUCCESS == exit_status): #{
    print MSG_SUCCESS
  else:
    print MSG_FAIL
  #} end if
  sys.exit(exit_status)
#} end if
