#! /usr/bin/env python
"""
with_defuse.py

Created by Lucas Swanson
Copyright (c) 2012 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

# import standard modules
from optparse import OptionParser, OptionGroup
import os, re, sys, time, traceback

# import custom modules
from version import VERSION
from utils.log import GetLogPath, CloseLogFile
from utils.error import MyError
from utils.general import (SetupMainClass, TimeSpent, WriteCommand,
  CheckConfigCommands, GetCommand, RunOverlapCode, COORD_PATT, StrJoin,
  ChrToInt)
from utils.messages import ErrMsg, LogMsg, DebugMsg, ExtremeDebugMsg
from utils.multi_dict import (SetupMultiDict, CheckMultiDict, AccessMultiDict,
  RemoveFromMultiDict, TraverseMultiDict, MultiDictError, MultiDictSortKeyCls)
from utils.files_paths import (CheckFilePath, CheckDirPath, EnsureAbsPath,
  GetOutDir, FileBoxCls, CleanLine)
from utils.subprocesses import RunCommandFromList, STREAM_OUT
from common.breakpoint import (SortBreakpointTuple,
    BreakpointPairsAreEqual)
from common.grouped_candidate import ParseFullEventID
from common.coord_pair import CoordPairCls
from parsers.candidate_group_parser import CandidateGroupParserCls
from parsers.overlap import OverlapResultCls
from parsers.fasta import FastaFileCls
from alignment_processing.alignment_functions import (ParseAlignmentFile,
  FixNonChromAlign, FixAlign, AlignBlockCls, ShortAlignString,
  NoAlignmentsError)
from alignment_processing.contig_with_alignments import (
    GetAlignLengthAndFraction)
from defuse_prediction import (DEFUSE_COMPARISON_HEADER, ALIGN_RESULTS_HEADER,
  DefusePredictionCls, DefusePredictionError)
from breakpoint_pair import BreakpointPairCls

# CONSTANTS
ES_SUCCESS = 0
ES_RUN_ERR = 1
ES_OPT_ERR = 2
ES_PATH_ERR = 3
ES_EXCEPTION = 4
MSG_SUCCESS = "SUCCESS"
MSG_FAIL = "FAIL"
COUNTS_HEADER1 = "\t".join(["        ","total      ","lcommon    ",
  "gcommon    ","bcommon    ","total   ","lcommon ","gcommon ","bcommon "])
COUNTS_HEADER2 = "\t".join(["Category","predictions","predictions",
  "predictions","predictions","partners","partners","partners","partners"])
CONTINUOUS_THRESHOLD = 5
BP_ATTRS = [ "wide_spread", "local_event", "untranscribed", "full_repeat",
  "repeats", "multi_maps" ]
BP_ATTR_DESCRIPTIONS = {
  "wide_spread":      "wide spread",
  "local_event": "contig can be explained by local event",
  "untranscribed":    "not in mature transcripts",
  "full_repeat":      "falls fully within repeat",
  "repeats":          "overlaps or is near repeat",
  "multi_maps":       "contig multi-maps",
}
COMPARE_RESULTS = ["common", "barnacle", "defuse"]
QUALITY_DESCRIPTIONS = [ "mitochondrial", "readthrough", "low confidence",
  "medium confidence", "high confidence", "good coverage", ]

class DefuseComparisonCls: #{
  def __init__(self, options): #{
    SetupMainClass(self, options)
    if (self.options.align): #{
      CheckConfigCommands(self, "blat")
    #} end if
    # lcom: common by general location (big_buffer)
    # gcom: common by gene partner names
    # bcom: common by specific breakpoint (small_buffer)
    # qual0: total
    # qual1: not mitochondrial
    # qual2: not mitochondrial, not a readthrough-like event
    # qual3: not mitochondrial, not readthrough, and medium or better
    #         confidence (based on min value for 'probability' field)
    # qual4: not mitochondrial, not readthrough, and high confidence
    #         (based on min value for 'probability' field)
    # qual5: not mitochondrial, not readthrough, high confidence, and
    #         good coverage
    self.counts = {
      "defuse": {
        "total": {"qual0":0,    "qual1":0,    "qual2":0,
                  "qual3":0,    "qual4":0,    "qual5":0,},
        "lcom":  {"qual0":0,    "qual1":0,    "qual2":0,
                  "qual3":0,    "qual4":0,    "qual5":0,},
        "gcom":  {"qual0":0,    "qual1":0,    "qual2":0,
                  "qual3":0,    "qual4":0,    "qual5":0,},
        "bcom":  {"qual0":0,    "qual1":0,    "qual2":0,
                  "qual3":0,    "qual4":0,    "qual5":0,},
      },
      "barnacle": {
        "total":0, "lcom":0, "gcom":0, "bcom":0,
      }
    }
    self.partners = {
      "defuse": {
        "total": {"qual0":set(),"qual1":set(),"qual2":set(),
                  "qual3":set(),"qual4":set(),"qual5":set(),},
        "lcom":  {"qual0":set(),"qual1":set(),"qual2":set(),
                  "qual3":set(),"qual4":set(),"qual5":set(),},
        "gcom":  {"qual0":set(),"qual1":set(),"qual2":set(),
                  "qual3":set(),"qual4":set(),"qual5":set(),},
        "bcom":  {"qual0":set(),"qual1":set(),"qual2":set(),
                  "qual3":set(),"qual4":set(),"qual5":set(),},
      },
      "barnacle": {
        "total":set(),"lcom":set(),"gcom":set(),"bcom":set(),
      }
    }
    # barnacle_by_chrom[chromA][chromB][pos_keyA][pos_keyB] = list
    # chromosome "X" := "Xr" if align to "-" strand
    # pos_key = floor(pos / big_buffer)
    self.barnacle_by_chrom = dict()
    # barnacle_by_gene[geneA][geneB] = list
    # genes sorted by position
    self.barnacle_by_gene = dict()
    # barnacle contigs (used to determine primary breakpoints for
    #   multi-mapped predictions)
    self.multi_mapped_contigs = dict()
    # candidate_contig_map[full_candidate_id] = contig_id
    self.candidate_contig_map = dict()
    # defuse results to be aligned to contigs (qual2 or better)
    self.defuse_to_align = {'gen':dict(),'ctg1':dict(),'ctg2':dict()}
    #self.defuse_to_realign = dict()
    # common gene-partners with tool-specific breakpoints
    self.breakpoint_partners = set()
    # breakpoints[chromA][chromB][pos_keyA][pos_keyB] = breakpoint_pair
    self.loci = dict()
    self.local_loci = dict()
    self.excluded_gene_loci = list()
    self.bad_loci = dict()
    self.breakpoints = dict()
    self.excluded_breakpoints = list()
    #self.excluded_genes = set()
    self.excluded_genes = list()
    self.local_event_ctgs = set()
    self.overlap_streams = {'barnacle':list(),'defuse':list()}
    self.curr_overlaps = {'barnacle':list(),'defuse':list()}
    self.outfiles = dict()
    #self.ctg_parts = dict()
    self.members_in_repeats = set()
    # dpred_bps[cluster_id] = matching_bp
    self.dpred_bps = dict()
    self.dpreds = list()
    self.dpred_loci_types = dict()
  #} end def

  def __del__(self): #{
    CloseLogFile(self)
    if (hasattr(self, "outfiles")): #{
      for file_key in self.outfiles: #{
        if (isinstance(self.outfiles[file_key], list)): #{
          for file in self.outfiles[file_key]: #{
            file.close()
          #} end for
        else:
          self.outfiles[file_key].close()
        #} end if
      #} end for
    #} end if
  #} end def

  def RunComparison(self): #{
    LogMsg(self, "Comparing results...")
    start = time.time()
    # setup the output files
    self.SetupOutputFiles()
    # load the genes to be set aside
    self.LoadExcludedGenes()
    # load the contigs that appear in local events
    self.LoadLocalEventContigs()
    # align contigs to repeat sequences
    self.DoContigRepeatAnalysis()
    # load Barnacle results
    self.DisplayLoci(self.loci)
    self.LoadBarnacleResults()
    self.DisplayLoci(self.loci)
    # prepare the repeat overlap streams for Defuse events
    self.PrepareRepeatOverlapStreams()
    # process Defuse results
    self.ProcessDefuseResults()
    self.DisplayLoci(self.loci)
    # count Barnacle events marked as common
    self.CountBarnacleCommons()
    if (0 < len(self.defuse_to_align['gen']) or
        0 < len(self.defuse_to_align['ctg1'])): #{
      self.DoAlignmentAnalysis()
    #} end if
    # write count results
    self.WriteCounts()
    if (self.options.align): #{
      self.WriteAlignCounts()
    #} end if
    # write the breakpoint analysis
    #self.WriteBreakpointAnalysis("main", self.breakpoints, 'breakpoints')
    # write the breakpoint analysis for the excluded genes
    #self.WriteBreakpointAnalysis("excluded", self.excluded_breakpoints,
    #    "excluded_breakpoints")
    # write the locus analysis
    main_outfiles = dict((ltype, self.outfiles["loci_%s" % ltype]) for
      ltype in COMPARE_RESULTS)
    #self.WriteLocusAnalysis("main", self.loci, 'loci')
    self.WriteLocusAnalysis("main", self.loci, main_outfiles)
    # write the breakpoint analysis for the excluded genes
    for (index, loci_list) in enumerate(self.excluded_gene_loci): #{
      excluded_outfiles = dict((ltype,
        self.outfiles["excluded_loci_%s" % ltype][index]) for
        ltype in COMPARE_RESULTS)
      #self.WriteLocusAnalysis("excluded", self.excluded_gene_loci,
      #  "excluded_loci")
      self.WriteLocusAnalysis("excluded %i" % index, loci_list,
        excluded_outfiles)
    #} end for
    if (not self.options.allow_bad_commons): #{
      bad_outfiles = dict((ltype, self.outfiles["bad_loci_%s" % ltype]) for
        ltype in COMPARE_RESULTS)
      self.WriteLocusAnalysis("bad", self.bad_loci, bad_outfiles)
    #} end if
    LogMsg(self, "Time spent comparing results: %s" % TimeSpent(start))
  #} end def

  def SetupOutputFiles(self): #{
    LogMsg(self, "Preparing output files...")
    self.OpenOutputFile('barnacle', "only_barnacle.data",
      "unique Barnacle results")
    # only put qual3 or higher predictions in only Defuse file
    self.OpenOutputFile('defuse', "only_defuse.data", "unique Defuse results",
      header=DEFUSE_COMPARISON_HEADER)
    self.OpenOutputFile('counts', "comparison_counts.txt", "counts output",
      header="\n".join((COUNTS_HEADER1,COUNTS_HEADER2)))
    self.OpenOutputFile('defuse_seqs', "defuse_seqs.fa",
      "Defuse sequences output")
    self.OpenOutputFile('defuse_short_seqs', "defuse_short_seqs.fa",
      "Defuse short sequences output")
    self.OpenOutputFile('u_breakpoints', "unique_breakpoints.txt",
      "common gene-partners with tool-specific breakpoints")
    if (self.options.align): #{
      #self.OpenOutputFile('align', "defuse_to_contigs.psl",
      #  "Defuse-to-contig alignment results", mode="r")
      self.OpenOutputFile('align_results', "defuse_align_results.txt",
        "Processed results of Defuse-to-contig alignments",
        header=ALIGN_RESULTS_HEADER)
    #} end if
    if (self.options.realign): #{
      self.OpenOutputFile('unaligned_defuse_seqs', "unaligned_defuse_seqs.fa",
        "unaligned Defuse sequences output")
      #self.OpenOutputFile('realign', "defuse_to_unmerged.psl",
      #  "Defuse-to-unmerged contigs realignment results", mode="r")
      #self.OpenOutputFile('realign_results', "defuse_realign_results.txt",
      #  "Processed results of Defuse-to-unmerged contig realignments",
      #  header=ALIGN_RESULTS_HEADER)
    #} end if
    self.OpenOutputFile("breakpoints", "breakpoints.txt",
      "breakpoint analysis file")
    self.OpenOutputFile("mm_ctgs", "multi_mapping_contigs.txt",
      "multi-mapping contigs file")
    #self.OpenOutputFile("loci_common", "loci_common.txt",
    #  "common loci file")
    #self.OpenOutputFile("loci_barnacle", "loci_barnacle.txt",
    #  "barnacle loci file")
    #self.OpenOutputFile("loci_defuse", "loci_defuse.txt",
    #  "defuse loci file")
    for ltype in COMPARE_RESULTS: #{
      self.OpenOutputFile("loci_%s" % ltype, "loci_%s.txt" % ltype,
        "%s loci file" % ltype)
    #} end for
    if (None != self.options.exclude_genes_paths): #{
      for excl_path in self.options.exclude_genes_paths: #{
        output_filename = ("%s_breakpoints.txt" % os.path.splitext(
          os.path.basename(excl_path))[0])
        self.OpenOutputFile("excluded_breakpoints", output_filename,
          "breakpoints involving excluded genes", file_list=True)
        for ltype in COMPARE_RESULTS: #{
          self.OpenOutputFile("excluded_loci_%s" % ltype,
            "%s_loci_%s.txt" % (output_filename, ltype),
            "%s loci involving excluded genes file" % ltype, file_list=True)
        #} end for
      #} end for
    #} end if
    if (not self.options.allow_bad_commons): #{
      for ltype in COMPARE_RESULTS: #{
        self.OpenOutputFile("bad_loci_%s" % ltype, "bad_loci_%s.txt" % ltype,
          "%s bad loci file" % ltype)
      #} end for
    #} end if
  #} end def

  def OpenOutputFile(self, key, filename, desc, header=None, mode="w",
      file_list=False): #{
    output_path = os.path.join(self.options.output_dir, filename)
    #self.outfiles[key] = FileBoxCls(output_path, mode,
    #  "could not create %s file" % desc)
    #if (None != header): #{
    #  self.outfiles[key].WriteLine(header)
    #} end if
    newfile = FileBoxCls(output_path, mode,
      "could not create %s file" % desc)
    if (None != header): #{
      newfile.WriteLine(header)
    #} end if
    if (file_list): #{
      if (key not in self.outfiles): #{
        self.outfiles[key] = list()
      #} end if
      self.outfiles[key].append(newfile)
    else:
      self.outfiles[key] = newfile
    #} end if
  #} end def

  def LoadExcludedGenes(self): #{
    if (None == self.options.exclude_genes_paths or
        0 == len(self.options.exclude_genes_paths)): #{
      return
    #} end if
    LogMsg(self, "Loading genes to exclude...")
    start = time.time()
    for excl_path in self.options.exclude_genes_paths: #{
      genes_file = FileBoxCls(excl_path, "r",
        "could not open excluded genes file")
      gene_set = set()
      for line in genes_file: #{
        #self.excluded_genes.add(line.lower())
        gene_set.add(line.lower())
      #} end for
      genes_file.close()
      self.excluded_genes.append(gene_set)
      self.excluded_gene_loci.append(dict())
      #gene_list = sorted(self.excluded_genes)
      #gene_str_list = list()
      #for i in range((len(self.excluded_genes)/10)+1): #{
      #  s = i*10
      #  e = min(s+10,len(self.excluded_genes))
      #  gene_str_list.append("  %s" % ", ".join(gene_list[s:e]))
      #} end for
      #LogMsg(self, "  Loaded %i excluded genes" % len(self.excluded_genes))
      LogMsg(self, "  Loaded %i excluded genes" % len(gene_set))
    #} end for
    #ExtremeDebugMsg(self, "Excluded genes:\n%s" % "\n".join(gene_str_list))
    LogMsg(self, "Time spent loading genes to exclude: %s" % TimeSpent(start))
  #} end def

  def LoadLocalEventContigs(self): #{
    if (None == self.options.local_event_ctgs_path): #{
      return
    #} end if
    LogMsg(self, "Loading local event contigs...")
    start = time.time()
    local_event_ctgs_file = FileBoxCls(self.options.local_event_ctgs_path,
      "r", "could not open local event contigs file")
    for line in local_event_ctgs_file: #{
      self.local_event_ctgs.add(line)
    #} end for
    local_event_ctgs_file.close()
    LogMsg(self, "Time spent loading local event contigs: %s" %
        TimeSpent(start))
  #} end def

  # use coordinates
  def DoContigRepeatAnalysis(self): #{
    if (None == self.options.repeat_paths or
        0 == len(self.options.repeat_paths)): #{
      return
    #} end if
    LogMsg(self, "Doing contig repeat analysis...")
    start = time.time()
    coords_path = self.CreateContigCoordsFile()
    for repeat_path in self.options.repeat_paths: #{
      ExtremeDebugMsg(self, "    creating overlap stream for %s" %
          os.path.basename(repeat_path))
      self.overlap_streams['barnacle'].append(RunOverlapCode(repeat_path,
        coords_path, STREAM_OUT, full=True, dpt=self.options.dpt,
        log_info=self.log_info))
      self.curr_overlaps['barnacle'].append(None)
    #} end for
  #} end if

  def TEMP(self): #{
    for align in map(FixAlign, aligns): #{
      # we only care about contig parts that align completely within a repeat
      if (align.query_len != align.matches + align.mismatches): #{
        continue
      #} end if
      ExtremeDebugMsg(self, "  Found contig part aligning within repeat:\n"
          "    %s" % AlignString(align))
      member_id = align.query[:-1]
      self.members_in_repeats.add(member_id)
    #} end for
    LogMsg(self, "Time spent doing contig repeat analysis: %s" % TimeSpent(start))
  #} end def

  # uses alignment (but BLAT cannot handle repeat sequences as ref)
  def OldDoContigRepeatAnalysis(self): #{
    if (None == self.options.repeat_seqs_path): #{
      return
    #} end if
    LogMsg(self, "Doing contig repeat analysis...")
    start = time.time()
    self.LoadContigParts()
    query_path = self.CreateContigQueryFile()
    output_path = os.path.join(self.options.output_dir,
        "contigs_to_repeats.psl")
    self.RunBlat(self.options.repeat_seqs_path, query_path, output_path)
    filters = { 'identity': 90.0 }
    try: #{
      aligns = ParseAlignmentFile(output_path, filters, log_info=self.log_info)
    except NoAlignmentsError, e:
      LogMsg(self, "  Found no alignments between contig parts and repeats")
      return
    #} end if
    for align in map(FixAlign, aligns): #{
      # we only care about contig parts that align completely within a repeat
      if (align.query_len != align.matches + align.mismatches): #{
        continue
      #} end if
      ExtremeDebugMsg(self, "  Found contig part aligning within repeat:\n"
          "    %s" % AlignString(align))
      member_id = align.query[:-1]
      self.members_in_repeats.add(member_id)
    #} end for
    LogMsg(self, "Time spent doing contig repeat analysis: %s" % TimeSpent(start))
  #} end def

  def CreateContigCoordsFile(self): #{
    LogMsg(self, "  Creating Barnacle coordinates file...")
    start = time.time()
    coords_path = os.path.join(self.options.output_dir, "barnacle_coords.bed")
    coords_file = FileBoxCls(coords_path, "w",
      "cannot create Barnacle coordinates file")
    parser = CandidateGroupParserCls(self.options.barnacle_path)
    for event in parser: #{
      for member in event.members: #{
        #ctg_id = member.contig_info.id
        #if (ctg_id not in self.ctg_parts): #{
        #  self.ctg_parts[ctg_id] = list()
        #} end if
        #ctg_part = GetContigPart(member.align_info_A, member.align_info_B)
        #self.ctg_parts[ctg_id].append(("%sA" % member.IDString(), ctg_part))
        #ctg_part = GetContigPart(member.align_info_B, member.align_info_A)
        #self.ctg_parts[ctg_id].append(("%sB" % member.IDString(), ctg_part))
        for (r_key, align) in zip("AB", member.alignments): #{
          coords_file.WriteLine(StrJoin(" ", [align.chrom, align.gen_coords.min,
            align.gen_coords.max, "%s%s" % (member.IDString(), r_key)]))
        #} end for
      #} end for
    #} end for
    LogMsg(self, "Time spent creating Barnacle coordinates file: %s" %
        TimeSpent(start))
    return coords_path
  #} end def

  def LoadContigParts(self): #{
    parser = CandidateGroupParserCls(self.options.barnacle_path)
    for event in parser: #{
      for member in event.members: #{
        ctg_id = member.contig_info.id
        if (ctg_id not in self.ctg_parts): #{
          self.ctg_parts[ctg_id] = list()
        #} end if
        ctg_part = GetContigPart(member.align_info_A, member.align_info_B)
        self.ctg_parts[ctg_id].append(("%sA" % member.IDString(), ctg_part))
        ctg_part = GetContigPart(member.align_info_B, member.align_info_A)
        self.ctg_parts[ctg_id].append(("%sB" % member.IDString(), ctg_part))
      #} end for
    #} end for
  #} end def

  def CreateContigQueryFile(self): #{
    ctg_parts_path = os.path.join(self.options.output_dir,
        "event_contig_parts.fa")
    ctg_parts_file = FileBoxCls(ctg_parts_path, "w",
        "cannot create event contig parts sequences file.")
    full_ctgs_file = FastaFileCls(self.options.contigs_path,
        "cannot open contig sequences file")
    for seq_obj in full_ctgs_file: #{
      if (seq_obj.id not in self.ctg_parts): #{
        continue
      #} end if
      for (id, coords) in self.ctg_parts[seq_obj.id]: #{
        ctg_parts_file.WriteLine(">%s\n%s" % (id,
          seq_obj.sequence[coords.min-1:coords.max]))
      #} end for
    #} end for
    ctg_parts_file.Close()
    return ctg_parts_file.path
  #} end def

  def LoadBarnacleResults(self): #{
    LogMsg(self, "Loading Barnacle results...")
    start = time.time()
    parser = CandidateGroupParserCls(self.options.barnacle_path)
    local_event_counts = {"events": 0, "candidates": 0}
    for event in parser: #{
      DebugMsg(self, "Loading event %i" % (event.id))
      # calculate spread
      self.CalculateEventSpread(event)
      local_event_found = False
      for member in sorted(event.members, key=GetBreakpointCoordsTuple): #{
        DebugMsg(self, "  Loading member %s\n    BP Coords: %s" %
          (member.candidate_id, StrJoin(" -- ",
            GetBreakpointCoordsTuple(member))))
        #breakpoint_list = self.ChooseBreakpointListForBarnacle(member)
        loci_list = self.ChooseLociListForBarnacle(member)
        # track the Barnacle breakpoint
        #DebugMsg(self, "Getting breakpoint for member %s:\n"
        #  "  %s" % (member.IDString(), member.BreakpointsString()))
        member.id = member.IDString()
        locus = self.GetLocus(member, loci_list)
        locus.AddBarnaclePrediction(member)
        #matching_bp = self.GetBreakpoint(member, breakpoint_list,
        #  self.options.small_buffer)
        matching_bp = self.GetBreakpoint(member, locus.breakpoints,
          self.options.small_buffer)
        #DebugMsg(self, "  Adding Barnacle prediction to breakpoint:")
        matching_bp.AddBarnaclePrediction(member)
        ExtremeDebugMsg(self, "    Transcribed: %i" %
            matching_bp.barnacle_transcribed)
        #DebugMsg(self, "    %s" % matching_bp.ToString())
        ctg_id = member.contig_info.id
        self.CheckMultiMapping(ctg_id, member, matching_bp)
        if (self.options.small_buffer < max(event.spread_pair)): #{
          DebugMsg(self, "    Setting \"wide spread\" to true for member "
              "breakpoint. Max spread: %i" % max(event.spread_pair))
          matching_bp.wide_spread = True
        #} end if
        if (ctg_id in self.local_event_ctgs): #{
          local_event_found = True
          local_event_counts["candidates"] += 1
          matching_bp.local_event = True
          ExtremeDebugMsg(self, "    contig can be explained as "
              "local event.")
        #} end if
        # check whether candidate overlaps any repeats
        self.CheckBarnacleRepeats(member.repeats, matching_bp)
        # check whether the part of the candidate aligns fully within a repeat
        self.ProcessRepeats("barnacle", member)
        if (not member.comparison_repeats): #{
          matching_bp.full_repeat = False
        #} end if
        #if (member.IDString() not in self.members_in_repeats): #{
        #  matching_bp.full_repeat = False
        #} end if
      #} end for
      if (local_event_found): #{
        local_event_counts["events"] += 1
      #} end if
      self.OldBarnacleProcessing(event)
    #} end for
    if (0 < len(self.local_event_ctgs)): #{
      LogMsg(self, "Found %i local event contigs in %i events" %
          (local_event_counts["candidates"], local_event_counts["events"]))
    #} end if
    if (self.log_info['debug']): #{
      partners = list()
      for genesA in self.barnacle_by_gene: #{
        for genesB in self.barnacle_by_gene[genesA]: #{
          partners.append(";".join([genesA, genesB]))
        #} end for
      #} end for
      LogMsg(self, "Barnacle gene partners:\n%s" % " ".join(sorted(partners)))
    #} end if
    LogMsg(self, "Time spent loading Barnacle results: %s" % TimeSpent(start))
  #} end def

  def CalculateEventSpread(self, event): #{
    ExtremeDebugMsg(self, "  Calculating breakpoint spread:")
    bp_coords = (set(), set())
    for candidate in event.members: #{
      sorted_bps = SortBreakpointTuple(candidate.breakpoints)
      for i in range(2): #{
        #ExtremeDebugMsg(self, "    Adding %i to side %i" %
        #    (sorted_bps[i].coord, i))
        bp_coords[i].add(sorted_bps[i].coord)
      #} end for
    #} end for
    #ExtremeDebugMsg(self, "    Side 0: %s\n    Side 1: %s" %
    #  (",".join(map(str, sorted(bp_coords[0]))),
    #   ",".join(map(str, sorted(bp_coords[1])))))
    event.spread_pair = [max(coord_set) - min(coord_set) for
      coord_set in bp_coords]
    ExtremeDebugMsg(self, "  Spread: %s" %
        ",".join(map(str, event.spread_pair)))
  #} end def

  def ChooseBreakpointListForBarnacle(self, member): #{
    #for gene_set in member.main_genes: #{
    #  genes = [g.lower().replace(".non_coding","") for g in gene_set]
    #  ExtremeDebugMsg(self, "    Genes: %s" % ", ".join(genes))
    #  intersection = self.excluded_genes.intersection(genes)
    #  if (0 < len(intersection)): #{
    #    ExtremeDebugMsg(self, "    Excluded genes: %s" %
    #        ", ".join(sorted(intersection)))
    #    return self.excluded_breakpoints
    #  #} end if
    #} end for
    for (gene_set, bp_list) in zip(self.excluded_genes,
        self.excluded_breakpoints): #{
      if (self.BarnacleWithExcludedGene(member, gene_set)): #{
        #return self.excluded_breakpoints
        return bp_list
      #} end if
    #} end for
    return self.breakpoints
  #} end def

  def ChooseLociListForBarnacle(self, member): #{
    ExtremeDebugMsg(self, "  Choosing loci list for member: %s (%s)" %
      (member.IDString(), member.MainGenesString()))
    #if (self.options.separate_local and
    #    member.contig_info.id in self.local_event_ctgs): #{
    #  return self.local_loci
    #} end if
    if (not self.options.allow_bad_commons and
        member.contig_info.id in self.local_event_ctgs): #{
      return self.bad_loci
    #} end if
    loci_list_descriptions = ["FP", "IG", "PSUEDO"].__iter__()
    for (gene_set, loci_list) in zip(self.excluded_genes,
        self.excluded_gene_loci): #{
      list_description = loci_list_descriptions.next()
      #ExtremeDebugMsg(self, "    Checking excluded %s genes" %
      #  list_description)
      if (self.BarnacleWithExcludedGene(member, gene_set)): #{
        ExtremeDebugMsg(self, "    Using excluded genes loci list!")
        #return self.excluded_gene_loci
        return loci_list
      #} end if
    #} end for
    return self.loci
  #} end def

  def BarnacleWithExcludedGene(self, member, excluded_genes): #{
    for gene_sets_pair in [member.main_genes, member.nearby_genes]: #{
      #for gene_set in member.main_genes: #{}
      for gene_set in gene_sets_pair: #{
        genes = [g.lower().replace(".non_coding","") for g in gene_set]
        #ExtremeDebugMsg(self, "    Genes: %s" % ", ".join(genes))
        #intersection = self.excluded_genes.intersection(genes)
        intersection = excluded_genes.intersection(genes)
        if (0 < len(intersection)): #{
          ExtremeDebugMsg(self, "    Excluded genes: %s" %
              ", ".join(sorted(intersection)))
          return True
        #} end if
      #} end for
    #} end for
    return False
  #} end def

  def CheckMultiMapping(self, ctg_id, member, matching_bp): #{
    member_multi_maps = member.meta_fields['multi_mapped']
    if (0 == member_multi_maps or ctg_id not in self.multi_mapped_contigs): #{
      DebugMsg(self, "    Setting primary to true for breakpoint")
      matching_bp.primary = True
    #} end if
    if (0 < member_multi_maps): #{
      if (ctg_id not in self.multi_mapped_contigs): #{
        DebugMsg(self, "    Storing primary breakpoint for multi-mapped "
          "contig")
        self.multi_mapped_contigs[ctg_id] = matching_bp
      #} end if
      self.candidate_contig_map[member.IDString()] = ctg_id
    else:
      DebugMsg(self, "    Setting multi_maps to false for breakpoint")
      matching_bp.multi_maps = False
    #} end if
  #} end def

  def CheckBarnacleRepeats(self, repeat_sets_pair, matching_bp): #{
    if (matching_bp.repeats): #{
      return
    #} end if
    for repeat_set in repeat_sets_pair: #{
      for repeat in repeat_set: #{
        if (repeat.lower() not in ["none", "n/a", "na"] and
            None == re.search(COORD_PATT, repeat, re.IGNORECASE)): #{
          matching_bp.repeats = True
          ExtremeDebugMsg(self, "    contig breakpoint falls within "
              "repeat: %s" % repeat)
          return
        #} end if
      #} end for
    #} end for
  #} end def

  def GetLocus(self, prediction, loci_list): #{
    locus = self.GetBreakpoint(prediction, loci_list, self.options.big_buffer)
    if (not hasattr(locus, "breakpoints")): #{
      locus.breakpoints = dict()
    #} end if
    return locus
  #} end def

  def GetBreakpoint(self, prediction, breakpoint_list, buffer): #{
    DebugMsg(self, "    Getting breakpoint for prediction...")
    #pair_tuple = SortBreakpoints(prediction.breakpointA,
    #  prediction.breakpointB)
    pair_tuple = SortBreakpointTuple((prediction.breakpointA,
      prediction.breakpointB))
    posA = pair_tuple[0].GetPositionKey(buffer)
    posB = pair_tuple[1].GetPositionKey(buffer)
    main_keys = [pair_tuple[0].chr, pair_tuple[1].chr, posA, posB]
    potential_bps = self.GetPotentialBreakpoints(main_keys, breakpoint_list)
    matching_bp = None
    DebugMsg(self, "    Looking for existing breakpoint...\n    %s" %
        ",".join((bp.ToString() for bp in pair_tuple)))
    for bp_pair_set in potential_bps: #{
      DebugMsg(self, "      checking breakpoint pair set: %s" %
          bp_pair_set.ToString())
      for pair in bp_pair_set.pairs: #{
        if (BreakpointPairsAreEqual(pair_tuple, pair,
            buffer=bp_pair_set.buffer)): #{
          if (None == matching_bp): #{
            ExtremeDebugMsg(self, "    Found matching breakpoint!")
            matching_bp = bp_pair_set
          else:
            LogMsg(self, "    Extra matching breakpoint found for %s\n"
              "  %s\n  %s" % (prediction.id, matching_bp.ToString(),
              bp_pair_set.ToString()))
          #} end if
          break
        else:
          DebugMsg(self, "    Breakpoints do not match:\n    %s\n    %s-%s" %
            (bp_pair_set.ToString(), pair_tuple[0].ToString(),
            pair_tuple[1].ToString()))
        #} end if
      #} end for
    #} end for
    if (None == matching_bp): #{
      DebugMsg(self, "    Adding new breakpoint...")
      matching_bp = BreakpointPairCls(buffer, other_type="defuse")
      matching_bp.primary = False
      # assume multi-mapping is true until a non-multi-mapping contig is
      #   encountered, so that a breakpoint will only be considered
      #   multi-mapping if all contigs that support it multi-map
      matching_bp.multi_maps = True
      matching_bp.wide_spread = False
      matching_bp.local_event = False
      matching_bp.repeats = False
      # assume full_repeat is true until a non-full_repeat contig is
      #   encountered, so that a breakpoint will only be considered
      #   a full_repeat if all contigs that support it are full_repeats
      matching_bp.full_repeat = True
      matching_bp.same_gene = False
      matching_bp.bp_keys = main_keys
      try: #{
        SetupMultiDict(breakpoint_list, main_keys, matching_bp,
          list_end=False)
      except MultiDictError, e:
        clashing_bp = AccessMultiDict(breakpoint_list, main_keys)
        raise DefuseComparisonError("Clashing breakpoints with keys "
          "%s:\n    %s\n    %s" % (",".join(map(str,main_keys)),
          matching_bp.ToString(), clashing_bp.ToString()))
      #} end try
    #} end if
    return matching_bp
  #} end def

  def GetPotentialBreakpoints(self, main_keys, breakpoint_list): #{
    potential_breakpoints = list()
    # breakpoints[chromA][chromB][pos_keyA][pos_keyB] = breakpoint_pair
    #for offsetA in [-1,0,1]: #{}
    for offsetA in range(-4,5): #{
      #for offsetB in [-1,0,1]: #{
      for offsetB in range(-4,5): #{
        keys = [main_keys[0], main_keys[1],
          main_keys[2]+offsetA, main_keys[3]+offsetB]
        if (CheckMultiDict(breakpoint_list, keys)): #{
          curr_breakpoint = AccessMultiDict(breakpoint_list, keys)
          potential_breakpoints.append(curr_breakpoint)
        #} end if
      #} end for
    #} end for
    return potential_breakpoints
  #} end def

  def OldBarnacleProcessing(self, event): #{
    # presume event is not common
    event.lcommon = False
    event.gcommon = False
    event.bcommon = False
    # use the breakpoint from the first member
    #breakpoint0 = event.members[0].breakpointA
    #breakpoint0.strand = event.members[0].align_info_A.Strand()
    #breakpoint1 = event.members[0].breakpointB
    #breakpoint1.strand = event.members[0].align_info_B.Strand()
    breakpoints = event.members[0].breakpoints
    #(breakpoint0.genes, breakpoint1.genes) = event.GetGeneSets()
    (breakpoints[0].genes, breakpoints[1].genes) = event.GetGeneSets()
    sorted_breakpoints = SortBreakpointTuple(breakpoints)
    (event.breakpointA, event.breakpointB) = sorted_breakpoints
    #DebugMsg(self, "  %s\n  %s" %
    #  (event.breakpointA.ToString(), event.breakpointB.ToString()))
    # barnacle_by_chrom[chromA][chromB][pos_keyA][pos_keyB] = list
    # chromosome "X" := "Xr" if align to "-" strand
    chromA = event.breakpointA.chr
    chromB = event.breakpointB.chr
    #chromA = GetDirectedChromKey(event.breakpointA)
    #chromB = GetDirectedChromKey(event.breakpointB)
    # pos_key = floor(pos / big_buffer)
    posA = event.breakpointA.GetPositionKey(self.options.big_buffer)
    posB = event.breakpointB.GetPositionKey(self.options.big_buffer)
    #DebugMsg(self, "  chromA: %s, chromB: %s, posA: %i, posB: %i" %
    #  (chromA, chromB, posA, posB))
    SetupMultiDict(self.barnacle_by_chrom, [chromA, chromB, posA, posB],
      event)
    # barnacle_by_gene[geneA][geneB] = list
    # genes sorted by position
    for geneA in event.breakpointA.genes: #{
      for geneB in event.breakpointB.genes: #{
        #genesA = ",".join(sorted(event.breakpointA.genes))
        #genesB = ",".join(sorted(event.breakpointB.genes))
        SetupMultiDict(self.barnacle_by_gene, [geneA, geneB], event)
      #} end for
    #} end for
    #chromA_dict = SetupMultiDict(self.barnacle_by_chrom, chromA)
    #chromB_dict = SetupMultiDict(chromA_dict, chromB)
    #posA_dict = SetupMultiDict(chromB_dict, posA)
    #posB_list = SetupMultiDict(posA_dict, posB, bottom=True)
    #posB_list.append(event)
  #} end def

  def PrepareRepeatOverlapStreams(self): #{
    if (None == self.options.repeat_paths or
        0 == len(self.options.repeat_paths)): #{
      return
    #} end if
    LogMsg(self, "Preparing repeat overlap streams...")
    start = time.time()
    LogMsg(self, "  Creating Defuse coordinates file...")
    coords_path = os.path.join(self.options.output_dir, "defuse_coords.bed")
    coords_file = FileBoxCls(coords_path, "w",
      "cannot create Defuse coordinates file")
    defuse_file = FileBoxCls(self.options.defuse_path, "r",
      "cannot read Defuse results file")
    field_cols = self.GetDefuseFieldColumns(defuse_file)
    num_defuse = 0
    message_printed = False
    for defuse_line in defuse_file: #{
      pred = DefusePredictionCls(defuse_line, field_cols,
        max_rthru=self.options.max_rthru, log_info=self.log_info)
      if (not pred.valid_chrs): #{
        if not message_printed: #{
          LogMsg(self, 'Skipping defuse prediction {} with invalid '
                 'chromosome'.format(pred.id))
          message_printed = True
        #} end if
        #raise DefuseComparisonError(
        #    'Defuse prediction {} has invalid chromsome:\n  {}'.format(
        #        pred.id, defuse_line))
        continue
      #} end if
      for i in range(2): #{
        bp = pred.breakpoints[i]
        coords_file.WriteLine(" ".join(map(str, [bp.chr, bp.coord-5,
          bp.coord+5, pred.id])))
      #} end for
      num_defuse += 1
    #} end for
    coords_file.Close()
    if (not num_defuse): #{
      raise DefuseComparisonError('Could not find any deFuse predictions!')
    #} end if
    LogMsg(self, "Time spent creating Defuse coordinates file: %s" %
        TimeSpent(start))
    for repeat_path in self.options.repeat_paths: #{
      ExtremeDebugMsg(self, "    creating overlap stream for %s" %
          os.path.basename(repeat_path))
      self.overlap_streams['defuse'].append(RunOverlapCode(repeat_path,
        coords_path, STREAM_OUT, dpt=self.options.dpt, log_info=self.log_info))
      self.curr_overlaps['defuse'].append(None)
    #} end for
    LogMsg(self, "Time spent preparing repeat overlap streams: %s" %
        TimeSpent(start))
  #} end def

  def ProcessDefuseResults(self): #{
    LogMsg(self, "Processing Defuse results...")
    start = time.time()
    defuse_file = FileBoxCls(self.options.defuse_path, "r",
      "could not read Defuse results file")
    field_cols = self.GetDefuseFieldColumns(defuse_file)
    # initial processing of Defuse predictions
    for defuse_line in defuse_file: #{
      defuse_prediction = DefusePredictionCls(defuse_line, field_cols,
        max_rthru=self.options.max_rthru, log_info=self.log_info)
      if (not defuse_prediction.valid_chrs): #{
        continue
      #} end if
      self.ProcessDefusePrediction(defuse_prediction)
    #} end for
    # use new alignments to genome to recheck for readthroughs
    if (0 < len(self.defuse_to_align['gen'])): #{
      self.AlignDefuseSequences('gen')
      self.ProcessAlignmentsToGenome()
    #} end if
    # finish processing Defuse predictions, after rechecking for readthroughs
    for dpred in self.dpreds: #{
      self.FinishProcessingDefusePredictions(dpred)
    #} end for
    LogMsg(self, "Time spent processing Defuse results: %s" % TimeSpent(start))
  #} end def

  def GetDefuseFieldColumns(self, defuse_file): #{
    LogMsg(self, "  Determining Defuse field order...")
    desired_fields = ["cluster_id", "sequence", "splitreads_count",
      "adjacent", "altsplice", "deletion", "gene1", "gene2",
      "chromosome1", "chromosome2", "gene_location1", "gene_location2",
      "gene_name1", "gene_name2", "genomic_break_pos1", "genomic_break_pos2",
      "genomic_strand1", "genomic_strand2", "interchromosomal", "inversion",
      "library_name", "readthrough", "defuse_probability",]
    backup_fields = {
        'sequence': 'splitr_sequence',
        'splitreads_count': 'splitr_count',
        'readthrough': 'read_through',
        'defuse_probability': 'probability',
        'chromosome1': 'gene_chromosome1',
        'chromosome2': 'gene_chromosome2'}
    field_cols = dict()
    heading_line = defuse_file.next()
    for (field_col, field_name) in enumerate(heading_line.split()): #{
      #DebugMsg(self, "  %i) %s" % (field_col, field_name))
      if (field_name in desired_fields + backup_fields.values()): #{
        if (field_name in field_cols): #{
          raise DefuseComparisonError("non-unique field name: %s: %i, %i" %
            (field_name, field_col, field_cols[field_name]))
        #} end if
        field_cols[field_name] = field_col
      #} end if
    #} end for
    # check that desired fields were all found
    missing = list()
    for field_name in desired_fields: #{
      if (field_name not in field_cols and
          backup_fields[field_name] not in field_cols): #{
        missing.append(field_name)
      #} end if
    #} end for
    if (0 < len(missing)): #{
      LogMsg(self, "WARNING missing Defuse fields: %s" % ", ".join(missing))
    #} end if
    return field_cols
  #} end def

  def OldProcessDefusePrediction(self, defuse_prediction): #{
    DebugMsg(self, "  Processing Defuse prediction %s (%s/%s)..." % (
      defuse_prediction.id, defuse_prediction.breakpointA.gene_name,
      defuse_prediction.breakpointB.gene_name))
    # set the quality
    self.SetDefuseQuality(defuse_prediction)
    DebugMsg(self, "      Quality: %i" % defuse_prediction.quality)
    self.OldDefuseProcessing(defuse_prediction)
    # track the Defuse breakpoint
    #breakpoint_list = self.breakpoints
    #for index in range(2): #{
    #  if (defuse_prediction.breakpoints[index].gene_name.lower() in
    #      self.excluded_genes): #{
    #    breakpoint_list = self.excluded_breakpoints
    #    break
    #  #} end if
    #} end for
    loci_list = self.ChooseLociListForDefuse(defuse_prediction)
    locus = self.GetLocus(defuse_prediction, loci_list)
    locus.AddOtherPrediction(defuse_prediction)
    #matching_bp = self.GetBreakpoint(defuse_prediction, breakpoint_list,
    #  self.options.small_buffer)
    matching_bp = self.GetBreakpoint(defuse_prediction, locus.breakpoints,
      self.options.small_buffer)
    matching_bp.AddOtherPrediction(defuse_prediction)
    self.dpred_bps[defuse_prediction.id] = matching_bp
    if (defuse_prediction.breakpoints[0].gene_name ==
        defuse_prediction.breakpoints[1].gene_name): #{
      matching_bp.same_gene = True
    #} end if
    #self.ProcessDefuseRepeats(defuse_prediction, matching_bp)
    self.ProcessRepeats("defuse", defuse_prediction)
    if (defuse_prediction.comparison_repeats): #{
      matching_bp.repeats = True
    #} end if
    new_transcribed = 0
    #for index in range(2): #{}
    #  if (defuse_prediction.breakpoints[index].gene_part.lower() in
    for breakpoint in defuse_prediction.breakpoints: #{
      if (breakpoint.gene_part.lower() in
          ["exon", "coding", "utr", "utr3p", "utr5p"]): #{
        new_transcribed += 1
      #} end if
    #} end for
    matching_bp.other_transcribed = max(matching_bp.other_transcribed,
      new_transcribed)
    # if the Defuse prediction is not common
    #if (not common and 2 <= defuse_prediction.quality): #{
    if (matching_bp.OtherOnly()): #{
      self.ProcessDefuseSpecific(defuse_prediction, matching_bp)
    elif (matching_bp.Common() and not matching_bp.primary):
      DebugMsg(self, "  Resetting primary breakpoint for %s" %
          ", ".join(matching_bp.barnacle_preds))
      for barnacle_id in matching_bp.barnacle_preds: #{
        contig_id = self.candidate_contig_map[barnacle_id]
        old_primary = self.multi_mapped_contigs[contig_id]
        if (old_primary.Common()): #{
          LogMsg(self, "Multi-mapped contig has multiple common breakpoints: "
              "%s" % ", ".join(old_primary.barnacle_preds))
        #} end if
        old_primary.primary = False
        matching_bp.primary = True
        self.multi_mapped_contigs[contig_id] = matching_bp
      #} end for
    #} end if
  #} end def

  def ProcessDefusePrediction(self, defuse_prediction): #{
    DebugMsg(self, "  Processing Defuse prediction %s (%s/%s)..." % (
      defuse_prediction.id, defuse_prediction.breakpointA.gene_name,
      defuse_prediction.breakpointB.gene_name))
    # set the quality
    self.SetDefuseQuality(defuse_prediction)
    DebugMsg(self, "      Quality: %i" % defuse_prediction.quality)
    self.OldDefuseProcessing(defuse_prediction)
    # track the Defuse breakpoint
    #breakpoint_list = self.breakpoints
    #for index in range(2): #{
    #  if (defuse_prediction.breakpoints[index].gene_name.lower() in
    #      self.excluded_genes): #{
    #    breakpoint_list = self.excluded_breakpoints
    #    break
    #  #} end if
    #} end for
    loci_list = self.ChooseLociListForDefuse(defuse_prediction)
    defuse_prediction.loci_list = loci_list
    locus = self.GetLocus(defuse_prediction, loci_list)
    #matching_bp = self.GetBreakpoint(defuse_prediction, breakpoint_list,
    #  self.options.small_buffer)
    matching_bp = self.GetBreakpoint(defuse_prediction, locus.breakpoints,
      self.options.small_buffer)
    self.DisplayLoci(loci_list)
    #defuse_prediction.locus = locus
    #defuse_prediction.matching_bp = matching_bp
    defuse_prediction.new_readthrough = False
    # if the breakpoint does not have any Barnacle predictions
    if (0 == len(matching_bp.barnacle_preds)): #{
      # temporarily remove it from the breakpoint list
      ExtremeDebugMsg(self, "Temporarily removing deFuse-specific "
          "prediction from breakpoint list...")
      RemoveFromMultiDict(locus.breakpoints, matching_bp.bp_keys)
      # if the path to a genome file is found, realign the Defuse breakpoint
      # sequences to make sure that they are really not read-throughs
      if (2 <= defuse_prediction.quality and
          None != self.options.genome_path): #{
        ExtremeDebugMsg(self, "  Flagging deFuse prediction to be aligned "
            "against genome (to recheck whether it is a readthrough.")
        self.AddDefusePredictionToAlign(defuse_prediction, 'gen',
          'defuse_seqs', shorten=False)
      #} end if
    #} end if
    # if the locus does not have any Barnacle predictions
    if (0 == len(locus.barnacle_preds)): #{
      # temporarily remove it from the breakpoint list
      ExtremeDebugMsg(self, "Temporarily removing deFuse-specific locus "
          "from loci list...")
      RemoveFromMultiDict(loci_list, locus.bp_keys)
    #} end if
    self.dpreds.append(defuse_prediction)
  #} end def

  def FinishProcessingDefusePredictions(self, defuse_prediction): #{
    ExtremeDebugMsg(self, "Finishing processing of deFuse prediction %s" %
        defuse_prediction.id)
    #if (defuse_prediction.new_readthrough): #{
    #  loci_list = self.ChooseLociListForDefuse(defuse_prediction)
    #  locus = self.GetLocus(defuse_prediction, loci_list)
    #else:
    #  locus = defuse_prediction.locus
    #  #matching_bp = defuse_prediction.matching_bp
    #} end if
    loci_list = self.ChooseLociListForDefuse(defuse_prediction)
    locus = self.GetLocus(defuse_prediction, loci_list)
    locus.AddOtherPrediction(defuse_prediction)
    matching_bp = self.GetBreakpoint(defuse_prediction, locus.breakpoints,
      self.options.small_buffer)
    matching_bp.AddOtherPrediction(defuse_prediction)
    self.dpred_bps[defuse_prediction.id] = matching_bp
    if (defuse_prediction.breakpoints[0].gene_name ==
        defuse_prediction.breakpoints[1].gene_name): #{
      matching_bp.same_gene = True
    #} end if
    #self.ProcessDefuseRepeats(defuse_prediction, matching_bp)
    self.ProcessRepeats("defuse", defuse_prediction)
    if (defuse_prediction.comparison_repeats): #{
      matching_bp.repeats = True
    #} end if
    new_transcribed = 0
    #for index in range(2): #{}
    #  if (defuse_prediction.breakpoints[index].gene_part.lower() in
    for breakpoint in defuse_prediction.breakpoints: #{
      if (breakpoint.gene_part.lower() in
          ["exon", "coding", "utr", "utr3p", "utr5p"]): #{
        new_transcribed += 1
      #} end if
    #} end for
    matching_bp.other_transcribed = max(matching_bp.other_transcribed,
      new_transcribed)
    # if the Defuse prediction is not common
    #if (not common and 2 <= defuse_prediction.quality): #{
    if (matching_bp.OtherOnly()): #{
      self.ProcessDefuseSpecific(defuse_prediction, matching_bp)
    elif (matching_bp.Common() and not matching_bp.primary):
      DebugMsg(self, "  Resetting primary breakpoint for %s" %
          ", ".join(matching_bp.barnacle_preds))
      for barnacle_id in matching_bp.barnacle_preds: #{
        contig_id = self.candidate_contig_map[barnacle_id]
        old_primary = self.multi_mapped_contigs[contig_id]
        if (old_primary.Common()): #{
          LogMsg(self, "Multi-mapped contig has multiple common breakpoints: "
              "%s" % ", ".join(old_primary.barnacle_preds))
        #} end if
        old_primary.primary = False
        matching_bp.primary = True
        self.multi_mapped_contigs[contig_id] = matching_bp
      #} end for
    #} end if
  #} end def

  def ChooseBreakpointListForDefuse(self, dpred): #{
    for (gene_set, bp_list) in zip (self.excluded_genes,
        self.excluded_breakpoints): #{
      if (self.DefuseWithExcludedGene(dpred, gene_set)): #{
        #return self.excluded_breakpoints
        return bp_list
      #} end if
    #} end for
    return self.breakpoints
  #} end def

  def ChooseLociListForDefuse(self, dpred): #{
    ExtremeDebugMsg(self, "  Choosing loci list for deFuse prediction: "
      "%s" % (dpred.ShortString()))
    if (not self.options.allow_bad_commons and 2 > dpred.quality): #{
      self.dpred_loci_types[dpred.id] = "bad"
      ExtremeDebugMsg(self, "    bad loci")
      return self.bad_loci
    #} end if
    for (gene_set, loci_list) in zip (self.excluded_genes,
        self.excluded_gene_loci): #{
      if (self.DefuseWithExcludedGene(dpred, gene_set)): #{
        #return self.excluded_gene_loci
        self.dpred_loci_types[dpred.id] = "excluded"
        ExtremeDebugMsg(self, "    excluded loci")
        return loci_list
      #} end if
    #} end for
    self.dpred_loci_types[dpred.id] = "main"
    ExtremeDebugMsg(self, "    main loci")
    return self.loci
  #} end def

  def DefuseWithExcludedGene(self, dpred, excluded_genes): #{
    for breakpoint in dpred.breakpoints: #{
      #if (breakpoint.gene_name.lower() in self.excluded_genes): #{}
      if (breakpoint.gene_name.lower() in excluded_genes): #{
        return True
      #} end if
    #} end for
    return False
  #} end def

  def ProcessDefuseSpecific(self, dpred, breakpoint): #{
    breakpoint.primary = True
    breakpoint.multi_maps = False
    breakpoint.full_repeat = False
    # if the Defuse prediction is not mitochondrial and not a readthrough
    if (2 <= dpred.quality): #{
      # if the path to a genome file is found, realign the Defuse breakpoint
      # sequences to make sure that they are really not read-throughs
      if (None != self.options.genome_path): #{
        self.AddDefusePredictionToAlign(dpred, 'gen',
            'defuse_seqs', shorten=False)
      else:
        if (self.options.align): #{
          # store the prediction for when processing alignments
          # of Defuse sequences to assembled contigs
          self.AddDefusePredictionToAlign(dpred, 'ctg1',
              'defuse_short_seqs', shorten=True)
        #} end if
        # if the Defuse prediction has a medium or higher quality
        if (3 <= dpred.quality): #{
          # write the Defuse prediction to the "only Defuse" output file
          DebugMsg(self, "    Recording good Defuse only prediction.")
          self.outfiles['defuse'].WriteLine(dpred.ToString())
        #} end if
      #} end if
    #} end if
  #} end def

  def OldDefuseProcessing(self, dpred): #{
    # update the total counts
    self.UpdateDefuseCounts("total", dpred.quality)
    self.UpdateDefusePartners("total", dpred)
    #bp_common = False
    if (self.LocationCommon(dpred)): #{
      self.UpdateDefuseCounts("lcom", dpred.quality)
      self.UpdateDefusePartners("lcom", dpred)
      for barnacle_pred in dpred.barnacle_lcom: #{
        barnacle_pred.lcommon = True
      #} end for
      if (self.PartnerCommon(dpred)): #{
        self.UpdateDefuseCounts("gcom", dpred.quality)
        self.UpdateDefusePartners("gcom", dpred)
      #} end if
      if (self.BreakpointCommon(dpred)): #{
        #bp_common = True
        self.UpdateDefuseCounts("bcom", dpred.quality)
        self.UpdateDefusePartners("bcom", dpred)
      elif (0 < len(dpred.barnacle_gcom)):
        partners = ";".join([dpred.breakpointA.gene_name,
          dpred.breakpointB.gene_name])
        self.breakpoint_partners.add("%s:defuse" % partners)
      #} end if
    elif (self.AltGeneCommon(dpred)):
      LogMsg(self, "Defuse prediction common by gene partners, but not by "
        "general location: %s: %s/%s" % (dpred.id,
        dpred.breakpointA.gene_name,
        dpred.breakpointB.gene_name))
      #common = True
      self.UpdateDefuseCounts("gcom", dpred.quality)
      self.UpdateDefusePartners("gcom", dpred)
      for barnacle_pred in dpred.barnacle_gcom: #{
        barnacle_pred.gcommon = True
      #} end for
    #} end if
  #} end def

  # ptype is prediction type, either barnacle or defuse
  def ProcessRepeats(self, ptype, pred): #{
    pred.comparison_repeats = False
    if (0 == len(self.overlap_streams[ptype])): #{
      return
    #} end if
    #for (stream, overlap) in zip(self.overlap_streams[ptype],
    #    self.curr_overlaps[ptype]): #{}
    for (i, stream) in enumerate(self.overlap_streams[ptype]): #{
      if (None == stream): #{
        break
      #} end if
      overlap = self.curr_overlaps[ptype][i]
      while (OverlapBeforePred(ptype, pred, overlap, self.log_info)): #{
        try: #{
          overlap_line = CleanLine(stream.next())
          #overlap = OverlapResultCls(overlap_line,
          self.curr_overlaps[ptype][i] = OverlapResultCls(overlap_line,
              log_info=self.log_info)
          overlap = self.curr_overlaps[ptype][i]
          ExtremeDebugMsg(self, "    R: %s\n    O: %s" % (overlap_line,
            overlap))
        except StopIteration:
          ExtremeDebugMsg(self, "stream finished")
          stream = None
          break
        #} end try
      #} end while
      if (None == stream): #{
        break
      #} end if
      while (OverlapMatchesPred(ptype, pred, overlap, self.log_info)): #{
        ExtremeDebugMsg(self, "    %s breakpoint overlaps repeat(s): %s" %
          (ptype, ", ".join(overlap.ResultList())))
        pred.comparison_repeats = True
        #if ("defuse" == ptype): #{
        #  pred.repeats = True
        #elif ("barnacle" == ptype):
        #  pred.repeats = False
        #else:
        #  raise DefuseComparisonError("unrecognized prediction type: %s" %
        #    ptype)
        #} end def
        try: #{
          overlap_line = CleanLine(stream.next())
          #overlap = OverlapResultCls(overlap_line,
          self.curr_overlaps[ptype][i] = OverlapResultCls(overlap_line,
              log_info=self.log_info)
          overlap = self.curr_overlaps[ptype][i]
          ExtremeDebugMsg(self, "    R: %s\n    O: %s" % (overlap_line,
            overlap))
        except StopIteration:
          ExtremeDebugMsg(self, "stream finished")
          break
        #} end try
      #} end while
    #} end for
  #} end def

  def ProcessDefuseRepeats(self, defuse_pred, breakpoint_pair): #{
    if (0 == len(self.overlap_streams['defuse'])): #{
      return
    #} end if
    #for (i, stream) in enumerate(self.overlap_streams['defuse']): #{}
    for (stream, overlap) in zip(self.overlap_streams['defuse'],
        self.curr_overlaps['defuse']): #{
      if (None == stream): #{
        break
      #} end if
      while (None == overlap or
          overlap.NumID() < defuse_pred.id): #{
        try: #{
          overlap_line = CleanLine(stream.next())
          overlap = OverlapResultCls(overlap_line,
              log_info=self.log_info)
          ExtremeDebugMsg(self, "    R: %s\n    O: %s" % (overlap_line,
            overlap))
        except StopIteration:
          ExtremeDebugMsg(self, "stream finished")
          #self.overlap_streams['defuse'][i] = None
          stream = None
          break
        #} end try
      #} end while
      if (None == stream): #{
        break
      #} end if
      while (overlap.NumID() == defuse_pred.id): #{
        ExtremeDebugMsg(self, "    Defuse breakpoint overlaps repeat(s): %s" %
            ", ".join(overlap.ResultList()))
        breakpoint_pair.repeats = True
        try: #{
          overlap_line = CleanLine(stream.next())
          overlap = OverlapResultCls(overlap_line,
              log_info=self.log_info)
          ExtremeDebugMsg(self, "    R: %s\n    O: %s" % (overlap_line,
            overlap))
        except StopIteration:
          ExtremeDebugMsg(self, "stream finished")
          #self.overlap_streams['defuse'][i] = None
          break
        #} end try
      #} end while
    #} end for
  #} end def

  def AddDefusePredictionToAlign(self, defuse_prediction, align_to,
      seq_file_key, shorten=False): #{
    if (-1 == defuse_prediction.breakpoint): #{
      LogMsg(self, "WARNING: could not get breakpoint coordinate for "
        "Defuse prediction %s" % defuse_prediction.id)
    else:
      ExtremeDebugMsg(self, "Adding defuse prediction %i to align to %s" %
          (defuse_prediction.id, align_to))
      self.defuse_to_align[align_to][defuse_prediction.id] = defuse_prediction
    #} end if
    # write Defuse sequence to file for aligning Defuse sequences
    # to assembled contigs
    sequence = defuse_prediction.seq.replace("|","")
    if (shorten): #{
      try: #{
        sequence = defuse_prediction.ShortSequence(self.DistFromBreakpoint())
      except DefusePredictionError, e:
        LogMsg(self, "Warning: %s" % e)
      #} end try
    #} end if
    self.outfiles[seq_file_key].WriteLine(">d%s\n%s" %
      (defuse_prediction.id, sequence))
  #} end def

  def SetDefuseQuality(self, defuse_pred): #{
    # qual0: total (mitochondrial)
    # qual1: not mitochondrial (readthrough)
    # qual2: not mitochondrial, not a readthrough-like event (low prob)
    # qual3: not mitochondrial, not readthrough, and medium or better
    #         confidence (based on min value for 'probability' field)
    # qual4: not mitochondrial, not readthrough, and high confidence
    #         (based on min value for 'probability' field)
    # qual5: not mitochondrial, not readthrough, high confidence, and
    #         good coverage
    DebugMsg(self, "    Setting Defuse prediction quality")
    if ("M" in [defuse_pred.breakpointA.chr, defuse_pred.breakpointB.chr]): #{
      DebugMsg(self, "    Mitochondrial prediction!")
      return
    #} end if
    defuse_pred.quality += 1 # qual1
    if (defuse_pred.readthrough): #{
      DebugMsg(self, "    Readthrough-like prediction!")
      return
    #} end if
    defuse_pred.quality += 1 # qual2
    if (self.options.probability_threshold1 >= defuse_pred.prob): #{
      DebugMsg(self, "    Low confidence prediction (prob: %.2f)!" %
        defuse_pred.prob)
      return
    #} end if
    defuse_pred.quality += 1 # qual3
    if (self.options.probability_threshold2 >= defuse_pred.prob): #{
      DebugMsg(self, "    Medium confidence prediction (prob: %.2f)!" %
        defuse_pred.prob)
      return
    #} end if
    defuse_pred.quality += 1 # qual4
    if (None == defuse_pred.coverage): #{
      DebugMsg(self, "    No coverage information for prediction!")
      return
    #} end if
    if (self.options.coverage_threshold >= defuse_pred.coverage): #{
      DebugMsg(self, "    Low coverage prediction (coverage: %i)!" %
        defuse_pred.coverage)
      return
    #} end if
    defuse_pred.quality += 1 # qual5
    #if (defuse_pred.deletion): #{
    #  DebugMsg(self, "    Deletion prediction!")
    #  return
    ##} end if
    #defuse_pred.quality += 1 # qual5
  #} end def

  def UpdateDefuseCounts(self, count_type, pred_quality): #{
    for curr_qual in range(pred_quality+1): #{
      qual_key = "qual%i" % curr_qual
      self.counts['defuse'][count_type][qual_key] += 1
    #} end for
  #} end def

  def UpdateDefusePartners(self, count_type, pred): #{
    partners = ";".join(sorted([pred.breakpointA.gene_name,
      pred.breakpointB.gene_name]))
    for curr_qual in range(pred.quality+1): #{
      qual_key = "qual%i" % curr_qual
      self.partners['defuse'][count_type][qual_key].add(partners)
    #} end for
  #} end def

  def LocationCommon(self, defuse_pred): #{
    #DebugMsg(self, "    Checking whether prediction is common by location")
    defuse_pred.barnacle_lcom = list()
    # barnacle_by_chrom[chromA][chromB][pos_keyA][pos_keyB] = list
    # pos_key = floor(pos / big_buffer)
    chromA = defuse_pred.breakpointA.chr
    chromB = defuse_pred.breakpointB.chr
    posA = defuse_pred.breakpointA.GetPositionKey(self.options.big_buffer)
    posB = defuse_pred.breakpointB.GetPositionKey(self.options.big_buffer)
    #DebugMsg(self, "    chromA: %s, chromB: %s, posA: %i, posB: %i" %
    #  (chromA, chromB, posA, posB))
    if (not CheckMultiDict(self.barnacle_by_chrom, [chromA, chromB])): #{
      #DebugMsg(self, "      Chromosome pair not common")
      return False
    #} end if
    for offsetA in [-1,0,1]: #{
      for offsetB in [-1,0,1]: #{
        keys = [chromA, chromB, posA+offsetA, posB+offsetB]
        #DebugMsg(self, "    Keys: %s" % ",".join(map(str, keys)))
        if (not CheckMultiDict(self.barnacle_by_chrom, keys)): #{
          #DebugMsg(self, "      Keys not common.")
          continue
        #} end if
        barnacle_list = AccessMultiDict(self.barnacle_by_chrom, keys)
        if (0 == offsetA + offsetB): #{
          #DebugMsg(self, "      Common predictions")
          defuse_pred.barnacle_lcom.extend(barnacle_list)
        else:
          for barnacle_pred in barnacle_list: #{
            lcommon = self.BreakpointsMatch(barnacle_pred, defuse_pred,
              self.options.big_buffer)
            if (lcommon): #{
              #DebugMsg(self, "      Common predictions")
              defuse_pred.barnacle_lcom.append(barnacle_pred)
            #} end if
          #} end for
        #} end if
      #} end for
    #} end for
    return (0 < len(defuse_pred.barnacle_lcom))
  #} end def

  def BreakpointsMatch(self, barnacle_pred, defuse_pred, buffer): #{
    diffA = abs(barnacle_pred.breakpointA.coord -
      defuse_pred.breakpointA.coord)
    diffB = abs(barnacle_pred.breakpointB.coord -
      defuse_pred.breakpointB.coord)
    DebugMsg(self, "      Comparing breakpoint coordinates (buffer:%s)\n"
      "      Barnacle: %s; %s\n      Defuse: %s; %s\n      DiffA:%i DiffB:%i" %
      (buffer, barnacle_pred.breakpointA.ToString(),
       barnacle_pred.breakpointB.ToString(),
       defuse_pred.breakpointA.ToString(),
       defuse_pred.breakpointB.ToString(), diffA, diffB))
    if (buffer < diffA or buffer < diffB): #{
      return False
    #} end if
    return True
  #} end def

  def PartnerCommon(self, defuse_pred): #{
    #DebugMsg(self, "    Checking whether prediction is common by "
    #  "gene-partners")
    defuse_pred.barnacle_gcom = list()
    dgeneA = defuse_pred.breakpointA.gene_name
    dgeneB = defuse_pred.breakpointB.gene_name
    #DebugMsg(self, "    Defuse: %s; %s" % (dgeneA, dgeneB))
    for barnacle_pred in defuse_pred.barnacle_lcom: #{
      #DebugMsg(self, "    Barnacle: %s; %s" % (
      #  ",".join(sorted(barnacle_pred.breakpointA.genes)),
      #  ",".join(sorted(barnacle_pred.breakpointB.genes))))
      if (dgeneA in barnacle_pred.breakpointA.genes and
          dgeneB in barnacle_pred.breakpointB.genes): #{
        #DebugMsg(self, "      Common predictions")
        defuse_pred.barnacle_gcom.append(barnacle_pred)
        barnacle_pred.gcommon = True
      #} end if
    #} end for
    return (0 < len(defuse_pred.barnacle_gcom))
  #} end def

  def BreakpointCommon(self, defuse_pred): #{
    #DebugMsg(self, "    Checking whether prediction is common by breakpoint")
    defuse_pred.barnacle_bcom = list()
    for barnacle_pred in defuse_pred.barnacle_lcom: #{
      bcommon = self.BreakpointsMatch(barnacle_pred, defuse_pred,
        self.options.small_buffer)
      if (bcommon): #{
        #DebugMsg(self, "      Common predictions")
        defuse_pred.barnacle_bcom.append(barnacle_pred)
        barnacle_pred.bcommon = True
      #} end if
    #} end for
    return (0 < len(defuse_pred.barnacle_bcom))
  #} end def

  def AltGeneCommon(self, defuse_pred): #{
    #DebugMsg(self, "    Checking whether prediction is common by "
    #  "gene-partners but not location")
    defuse_pred.barnacle_gcom = list()
    keys = [defuse_pred.breakpointA.gene_name,
      defuse_pred.breakpointB.gene_name]
    #DebugMsg(self, "    Keys: %s" % ",".join(map(str, keys)))
    if (CheckMultiDict(self.barnacle_by_gene, keys)): #{
      #DebugMsg(self, "    Common partners!")
      barnacle_list = AccessMultiDict(self.barnacle_by_gene, keys)
      defuse_pred.barnacle_gcom.extend(barnacle_list)
      for barnacle_pred in barnacle_list: #{
        if (defuse_pred.breakpointA.chr != barnacle_pred.breakpointA.chr or
            defuse_pred.breakpointB.chr != barnacle_pred.breakpointB.chr): #{
          LogMsg(self, "WARNING predictions common by genes, but not by "
            "chromosomes!\nBarnacle: %s; %s\nDefuse: %s; %s" %
            (barnacle_pred.breakpointA.ToString(),
             barnacle_pred.breakpointB.ToString(),
             defuse_pred.breakpointA.FullString(),
             defuse_pred.breakpointB.FullString()))
        #} end if
      #} end for
    #} end if
    return (0 < len(defuse_pred.barnacle_gcom))
  #} end def

  def CountBarnacleCommons(self): #{
    LogMsg(self, "Counting common Barnacle predictions...")
    start = time.time()
    # use barnacle_by_chrom
    for barnacle_pred in TraverseMultiDict(self.barnacle_by_chrom): #{
      partners = ";".join([",".join(sorted(barnacle_pred.breakpointA.genes)),
        ",".join(sorted(barnacle_pred.breakpointB.genes))])
      #DebugMsg(self, "  G%i: %s lcom:%s gcom:%s bcom:%s" %
      #  (barnacle_pred.id, partners, barnacle_pred.lcommon,
      #  barnacle_pred.gcommon, barnacle_pred.bcommon))
      self.UpdateBarnacleCounts("total")
      gene_partners = (barnacle_pred.breakpointA.genes,
        barnacle_pred.breakpointB.genes)
      self.UpdateBarnaclePartners("total", gene_partners)
      #DebugMsg(self, "Barnacle gene partners:\n%s" %
      #  " ".join((";".join([",".join(sorted(partner[0])),
      #                      ",".join(sorted(partner[1]))]) for
      #  partner in self.partners['barnacle']['total'])))
      common = False
      if (barnacle_pred.lcommon): #{
        common = True
        self.UpdateBarnacleCounts("lcom")
        self.UpdateBarnaclePartners("lcom", gene_partners)
        #DebugMsg(self, "Barnacle lcommon gene partners:\n%s" %
        #  " ".join((";".join([",".join(sorted(partner[0])),
        #                      ",".join(sorted(partner[1]))]) for
        #  partner in self.partners['barnacle']['lcom'])))
      #} end if
      if (barnacle_pred.gcommon): #{
        common = True
        self.UpdateBarnacleCounts("gcom")
        self.UpdateBarnaclePartners("gcom", gene_partners)
        #DebugMsg(self, "Barnacle gcommon gene partners:\n%s" %
        #  " ".join((";".join([",".join(sorted(partner[0])),
        #                      ",".join(sorted(partner[1]))]) for
        #  partner in self.partners['barnacle']['gcom'])))
      #} end if
      if (barnacle_pred.bcommon): #{
        common = True
        self.UpdateBarnacleCounts("bcom")
        self.UpdateBarnaclePartners("bcom", gene_partners)
        #DebugMsg(self, "Barnacle bcommon gene partners:\n%s" %
        #  " ".join((";".join([",".join(sorted(partner[0])),
        #                      ",".join(sorted(partner[1]))]) for
        #  partner in self.partners['barnacle']['bcom'])))
      elif (barnacle_pred.gcommon):
        self.breakpoint_partners.add("%s:barnacle" % partners)
      #} end if
      # if the Barnacle prediction is not common
      if (not common): #{
        #DebugMsg(self, "    Recording Barnacle only prediction.")
        # write the Barnacle prediction to the "only Barnacle" output file
        self.outfiles['barnacle'].Write(barnacle_pred.FullDataString())
      #} end if
    #} end for
    LogMsg(self, "Time spent counting common Barnacle predictions: %s" %
      TimeSpent(start))
  #} end def

  def UpdateBarnacleCounts(self, count_type): #{
    self.counts['barnacle'][count_type] += 1
  #} end def

  def UpdateBarnaclePartners(self, count_type, new_partners): #{
    #DebugMsg(self, "Updating Barnacle %s gene-partners" % count_type)
    for old_partners in self.partners['barnacle'][count_type]: #{
      commonA = old_partners[0].intersection(new_partners[0])
      commonB = old_partners[1].intersection(new_partners[1])
      if (0 < len(commonA) and 0 < len(commonB)): #{
        if (commonA == new_partners[0] and
            commonB == new_partners[1]): #{
          #DebugMsg(self, "New partners already present")
          return
        #} end if
        #DebugMsg(self, "Updating old partners %s with new partners %s" % (
        #  ";".join([",".join(sorted(old_partners[0])),
        #            ",".join(sorted(old_partners[1]))]),
        #  ";".join([",".join(sorted(new_partners[0])),
        #            ",".join(sorted(new_partners[1]))])))
        new_partners[0].update(old_partners[0])
        new_partners[1].update(old_partners[1])
        self.partners['barnacle'][count_type].remove(old_partners)
        break
      #} end if
    #} end for
    frozen_partners = (frozenset(new_partners[0]), frozenset(new_partners[1]))
    self.partners['barnacle'][count_type].add(frozen_partners)
  #} end def

  def DoAlignmentAnalysis(self): #{
    #if (0 < len(self.defuse_to_align['gen'])): #{
    #  self.AlignDefuseSequences('gen')
    #  self.ProcessAlignmentsToGenome()
    #} end if
    if (0 == len(self.defuse_to_align['ctg1'])): #{
      return
    #} end if
    # align Defuse sequences to Contigs
    self.AlignDefuseSequences('ctg1')
    # process Defuse-to-Contig alignments
    self.ProcessAlignmentsToContigs()
    if (self.options.realign): #{
      # select the Defuse predictions to realign
      self.SelectDefusePredictionsToRealign()
      #if (0 < len(self.defuse_to_realign)): #{}
      if (0 < len(self.defuse_to_align['ctg2'])): #{
        for kvalue in range(self.options.low_k, self.options.high_k+1): #{
          self.DoRealignmentAnalysis(kvalue=kvalue, contig_type="contigs")
          if (self.options.include_bubbles): #{
            self.DoRealignmentAnalysis(kvalue=kvalue, contig_type="bubbles")
          #} end if
          if (self.options.include_indels): #{
            self.DoRealignmentAnalysis(kvalue=kvalue, contig_type="indel")
          #} end if
        #} end for
      #} end if
    #} end if
  #} end def

  def DoRealignmentAnalysis(self, kvalue, contig_type): #{
    key = "%i-%s" % (kvalue, contig_type)
    # align Defuse sequences to Contigs
    self.AlignDefuseSequences(key)
    # process Defuse-to-Contig alignments
    self.ProcessAlignmentsToContigs(key)
    # collate the information and reset for next k-value
    self.FinishProcessingK(key)
  #} end def

  def AlignDefuseSequences(self, key): #{
    start = time.time()
    # get the appropriate target, query, and output paths
    if ('gen' == key): #{
      LogMsg(self, "Aligning Defuse sequences to genome...")
      target_path = self.options.genome_path
      query_file = self.outfiles['defuse_seqs']
      output_path = self.options.align_to_genome_path
    elif ('ctg1' == key):
      LogMsg(self, "Aligning Defuse sequences to assembled contigs...")
      target_path = self.options.contigs_path
      query_file = self.outfiles['defuse_short_seqs']
      output_path = self.options.align_to_contigs_path
    else:
      LogMsg(self, "Realigning unaligned Defuse sequences to unmerged "
        "k%s..." % key)
      target_path = self.options.realign_target_paths[key]
      query_file = self.outfiles['unaligned_defuse_seqs']
      output_path = self.options.realign_output_paths[key]
    #} end if
    # close the Defuse sequences query file
    query_file.Close()
    self.RunBlat(target_path, query_file.path, output_path)
    LogMsg(self, "Time spent aligning Defuse sequences: %s" % TimeSpent(start))
  #} end def

  def RunBlat(self, target_path, query_path, output_path): #{
    # use "-minMatch=1", "stepSize=5", "minScore=4" to allow short matches
    # use "repMatch=2253" to mimic webBLAT results
    command = [GetCommand(self, "blat"), "-minMatch=1", "-stepSize=5",
      "-repMatch=2253", "-minScore=4", target_path, query_path,
      output_path]
    DebugMsg(self, " ".join(command))
    return_code = RunCommandFromList(command, dpt=self.options.dpt)
    if (0 != return_code and not self.log_info['debug']): #{
      LogMsg(self, " ".join(command))
    #} end if
    if (0 > return_code): #{
      raise DefuseComparisonError("Alignment command was terminated by "
        "signal %i" % return_code)
    elif (0 < return_code):
      raise DefuseComparisonError("Error running alignment command: %i" %
        return_code)
    #} end if
  #} end def

  def ProcessAlignmentsToGenome(self): #{
    LogMsg(self, "Processing Defuse-to-genome alignments...")
    start = time.time()
    filters = { 'identity': 40.0 }
    try:
      aligns = ParseAlignmentFile(self.options.align_to_genome_path, filters,
          log_info=self.log_info)
    except NoAlignmentsError, e:
      LogMsg(self, "  Found no alignments")
      return
    #} end if
    num_missed_readthroughs = 0
    for align in map(FixAlign, aligns): #{
      (align_len, align_fraction) = GetAlignLengthAndFraction(align)
      match_fraction = float(align.match) / float(align.query_len)
      if (0.6 < align_fraction and 0.6 < match_fraction): #{
        ExtremeDebugMsg(self, "  Processing alignment %s->%s: %i-%i, M:%i\n"
          "    Align Fraction: %.2f, Match Fraction: %.2f, Min: %.2f" %
          (align.query, align.target, align.qstart, align.qend, align.match,
          align_fraction, match_fraction, self.options.readthrough_threshold))
      #} end if
      if (align_fraction >= self.options.readthrough_threshold and
          match_fraction >= (self.options.readthrough_threshold - 0.05)): #{
        DebugMsg(self, "    Single align: %s" % align.query)
        ExtremeDebugMsg(self, "      %s" % ShortAlignString(align))
        cluster_id = int(align.query[1:])
        DebugMsg(self, "    Defuse prediction %i is missed readthrough" %
            cluster_id)
        if (cluster_id not in self.defuse_to_align['gen']): #{
          raise DefuseComparisonError("Defuse prediction %i not in "
            "\"defuse_to_align\"!\n  %s" % (cluster_id,
            StrJoin(", ", sorted(self.defuse_to_align['gen']))))
        #} end if
        if (not self.defuse_to_align['gen'][cluster_id].readthrough): #{
          num_missed_readthroughs += 1
          self.defuse_to_align['gen'][cluster_id].readthrough = True
          self.defuse_to_align['gen'][cluster_id].new_readthrough = True
          self.defuse_to_align['gen'][cluster_id].quality = 1
          #self.dpred_bps[cluster_id].UpdateOtherQuality(cluster_id, 1)
        #} end if
      #} end if
    #} end for
    LogMsg(self, "%i Defuse predictions have readthrough alignments but "
        "were not marked as readthroughs." % num_missed_readthroughs)
    for pred in self.defuse_to_align['gen'].itervalues(): #{
      if (2 <= pred.quality): #{
        if (self.options.align): #{
          # store the prediction for when processing alignments
          # of Defuse sequences to assembled contigs
          self.AddDefusePredictionToAlign(pred, 'ctg1',
              'defuse_short_seqs', shorten=True)
        #} end if
        # if the Defuse prediction has a high quality
        if (3 <= pred.quality): #{
          # write the Defuse prediction to the "only Defuse" output file
          DebugMsg(self, "    Recording good Defuse only prediction.")
          self.outfiles['defuse'].WriteLine(pred.ToString())
        #} end if
      #} end if
    #} end for
    #if (0 < num_missed_readthroughs): #{
    #  self.MoveNewReadthroughs()
    #} end if
  #} end def

  def MoveNewReadthroughs(self): #{
    old_loci_list = dict()
  #} end def

  def ProcessAlignmentsToContigs(self, key=None): #{
    LogMsg(self, "Processing Defuse-to-contig alignments...")
    start = time.time()
    # get the appropriate list of defuse predictions
    if (None == key): #{
      preds_to_process = self.defuse_to_align['ctg1']
      aligns_path = self.options.align_to_contigs_path
    else:
      #preds_to_process = self.defuse_to_realign
      preds_to_process = self.defuse_to_align['ctg2']
      aligns_path = self.options.realign_output_paths[key]
    #} end if
    filters = { 'identity': 80.0 }
    try:
      aligns = ParseAlignmentFile(aligns_path, filters, log_info=self.log_info)
    except NoAlignmentsError, e:
      LogMsg(self, "  Found no alignments")
      return
    #} end if
    for align in map(FixNonChromAlign, aligns): #{
      (align_len, align_fraction) = GetAlignLengthAndFraction(align)
      match_fraction = float(align.match) / float(align.query_len)
      if (0.6 < align_fraction and 0.6 < match_fraction): #{
        ExtremeDebugMsg(self, "Processing alignment %s->%s: %i-%i, M:%i" %
          (align.query, align.target, align.qstart, align.qend, align.match))
      #} end if
      # skip the 'd' at the beginning of the query name
      cluster_id = int(align.query[1:])
      if (cluster_id not in preds_to_process): #{
        ExtremeDebugMsg(self, "  skipping alignment for prediction %i" %
          cluster_id)
        continue
      #} end if
      defuse_pred = preds_to_process[cluster_id]
      if (3 > align.qstart and 2 > (align.query_len - align.qend) and
          3 > align.qbaseinsert and 3 > align.tbaseinsert): #{
        DebugMsg(self, "Found full Defuse alignment!\n  %s" % align.psl_str)
        defuse_pred.full_align = True
        #defuse_pred.UpdateBestAligns(align)
        defuse_pred.best_aligns['full'].Update(align)
        continue
      #} end if
      breakpoint = self.DistFromBreakpoint()
      overlapA = (breakpoint - align.qstart) + 1
      overlapB = (align.qend - breakpoint) + 1
      if (10 <= min(overlapA, overlapB)): #{
        ExtremeDebugMsg(self, "  Cluster %s, breakpoint: %i overlap: %i/%i" %
          (defuse_pred.ShortString(), breakpoint, overlapA, overlapB))
      #} end if
      if (self.options.align_threshold2 <= min(overlapA, overlapB)): #{
        overlaps = self.GetTrueOverlaps(breakpoint, align)
        tleft = min(align.tstart, align.tend)
        tright = max(align.tstart, align.tend)
        if (3 > tleft and 2 > (align.target_len - tright) and
            3 > align.tbaseinsert and
            self.options.align_threshold2 <= min(overlaps)): #{
          DebugMsg(self, "Found full contig alignment!\n  %s" %
            align.psl_str)
          defuse_pred.full_align = True
          #defuse_pred.UpdateBestAligns(align)
          defuse_pred.best_aligns['tfull'].Update(align)
        elif (self.options.align_threshold1 <= min(overlaps)):
          defuse_pred.part_align = True
          defuse_pred.best_aligns['part'].Update(align)
          #if (not defuse_pred.full_align): #{
          #  defuse_pred.UpdateBestAligns(align)
          #} end if
        #} end if
      #} end if
    #} end for
    LogMsg(self, "Time spent processing alignments: %s" % TimeSpent(start))
  #} end def

  def GetTrueOverlaps(self, breakpoint, align): #{
    ExtremeDebugMsg(self, "    Getting true overlap values")
    # overlaps = [total_before_bp, total_after_bp,
    #   continuous_before_bp, continuous_after_bp]
    overlaps = [0,0,0,0]
    (dprev,cprev) = (None,None)
    continuous_after = True
    for (defuse_block,ctg_block) in zip(align.query_blocks,
        align.target_blocks): #{
      #(dblock_start, dblock_end) = sorted(map(int, defuse_block))
      #dsize = (dblock_end-dblock_start)+1
      dblock = AlignBlockCls(defuse_block)
      cblock = AlignBlockCls(ctg_block)
      if (dblock.span != cblock.span): #{
        raise DefuseComparisonError("defuse sequence and contig alignment "
          "block sizes do not match: %i, %i" % (dblock.span, cblock.span))
      #} end if
      if (None != dprev and None != cprev): #{
        if (dblock.start < breakpoint and
            (CONTINUOUS_THRESHOLD > dblock.Gap(dprev) or
             CONTINUOUS_THRESHOLD > cblock.Gap(cprev))): #{
          overlaps[2] = 0
        #} end if
        if (breakpoint < dblock.start and
            (CONTINUOUS_THRESHOLD > dblock.Gap(dprev) or
             CONTINUOUS_THRESHOLD > cblock.Gap(cprev))): #{
          continuous_after = False
        #} end if
      #} end if
      if (dblock.end < breakpoint): #{
        overlaps[0] += dblock.span
        overlaps[2] += dblock.span
      elif (breakpoint < dblock.start):
        overlaps[1] += dblock.span
        if (continuous_after): #{
          overlaps[3] += dblock.span
        #} end if
      else:
        before = (breakpoint-dblock.start)+1
        overlaps[0] += before
        overlaps[2] += before
        after = (dblock.end-breakpoint)+1
        overlaps[1] += after
        overlaps[3] += after
      #} end if
      ExtremeDebugMsg(self, "  Defuse block: %s, Contig block: %s, "
        "Overlaps: %s (%s)" % (dblock, cblock, StrJoin("/", overlaps[:2]),
        StrJoin("/", overlaps[2:])))
        #"/".join(map(str,overlaps[:2])), "/".join(map(str, overlaps[2:]))))
      (dprev, cprev) = (dblock, cblock)
    #} end for
    return overlaps
  #} end def

  def SelectDefusePredictionsToRealign(self): #{
    for pred in self.defuse_to_align['ctg1'].itervalues(): #{
      # if the defuse prediction sequence did not align to any merged contigs
      if (not pred.full_align and not pred.part_align): #{
        pred.realigned = True
        #self.defuse_to_realign[pred.id] = pred
        #self.defuse_to_align['ctg2'][pred.id] = pred
        # write Defuse sequence to file for realigning Defuse sequences
        # to pre-merging assembled contigs
        self.AddDefusePredictionToAlign(pred, 'ctg2',
            'unaligned_defuse_seqs', shorten=True)
        #try: #{
        #  short_sequence = pred.ShortSequence(self.DistFromBreakpoint())
        #except DefusePredictionError, e:
        #  LogMsg(self, "Warning: %s" % e)
        #  short_sequence = defuse_prediction.seq
        #} end try
        #self.outfiles['unaligned_defuse_seqs'].WriteLine(">d%s\n%s" %
        #  (pred.id, short_sequence))
      #} end if
    #} end for
  #} end def

  def FinishProcessingK(self, key): #{
    #for pred in self.defuse_to_realign.itervalues(): #{}
    for pred in self.defuse_to_align['ctg2'].itervalues(): #{
      if (pred.full_align): #{
        pred.any_full_realign = True
      #} end if
      if (pred.part_align): #{
        pred.any_part_realign = True
      #} end if
      # set the appropriate values for the current k-value and clear the
      # general results to prepare for the next k-value
      pred.full_realign[key] = pred.full_align
      pred.full_align = False
      pred.part_realign[key] = pred.part_align
      pred.part_align = False
      pred.best_realigns[key] = dict()
      for align_type in ['full', 'tfull', 'part']: #{
        align_list = pred.best_aligns[align_type].aligns
        pred.best_realigns[key][align_type] = align_list
        pred.best_aligns[align_type].Reset()
      #} end for
    #} end for
  #} end def

  def WriteCounts(self): #{
    #LogMsg(self, " ".join(self.partners['defuse']['total']['qual0']))
    LogMsg(self, "%s\n%s" % (COUNTS_HEADER1, COUNTS_HEADER2))
    for qual in range(6): #{
      qual_key = "qual%i" % qual
      data_list = ["defuse_%i" % qual]
      for count_type in ["total","lcom","gcom","bcom"]: #{
        data_list.append(self.counts['defuse'][count_type][qual_key])
      #} end for
      for count_type in ["total","lcom","gcom","bcom"]: #{
        data_list.append(len(self.partners['defuse'][count_type][qual_key]))
      #} end for
      data_str = "\t".join(map(str, data_list))
      self.outfiles['counts'].WriteLine(data_str)
      LogMsg(self, data_str)
    #} end for
    data_list = ["barnacle"]
    for count_type in ["total","lcom","gcom","bcom"]: #{
      data_list.append(self.counts['barnacle'][count_type])
    #} end for
    for count_type in ["total","lcom","gcom","bcom"]: #{
      data_list.append(len(self.partners['barnacle'][count_type]))
    #} end for
    data_str = "\t".join(map(str, data_list))
    self.outfiles['counts'].WriteLine(data_str)
    LogMsg(self, "%s\n%s" % (self.options.lib, data_str))
    # write out common gene-partners with tool-specific breakpoints
    for partners in sorted(self.breakpoint_partners): #{
      self.outfiles['u_breakpoints'].WriteLine(partners)
    #} end for
  #} end def

  def WriteAlignCounts(self): #{
    total = len(self.defuse_to_align['ctg1'])
    full_aligns = part_aligns = no_aligns = 0
    #realign_total = len(self.defuse_to_realign)
    realign_total = len(self.defuse_to_align['ctg2'])
    full_realigns = part_realigns = no_realigns = 0
    for defuse_pred in self.defuse_to_align['ctg1'].itervalues(): #{
      if (defuse_pred.realigned): #{
        no_aligns += 1
        if (defuse_pred.any_full_realign): #{
          full_realigns += 1
        elif (defuse_pred.any_part_realign):
          part_realigns += 1
        else:
          no_realigns += 1
        #} end if
      else:
        if (defuse_pred.full_align): #{
          full_aligns += 1
        elif (defuse_pred.part_align):
          part_aligns += 1
        else:
          no_aligns += 1
        #} end if
      #} end if
      self.outfiles['align_results'].WriteLine(defuse_pred.AlignString())
    #} end for
    data_str = ("Attempted to align %s Defuse sequences\n  Full: %i, "
      "Partial: %i, None: %i" % (total, full_aligns, part_aligns, no_aligns))
    if (self.options.realign): #{
      data_str += ("\nAttempted to realign %s Defuse sequences\n  Full: %i, "
        "Partial: %i, None: %i" % (realign_total, full_realigns, part_realigns,
        no_realigns))
    #} end if
    self.outfiles['counts'].WriteLine(data_str)
    LogMsg(self, data_str)
  #} end def

  def WriteBreakpointAnalysis(self, description, breakpoint_list,
      output_file_key): #{
    LogMsg(self, "Writing %s breakpoint analysis..." % description)
    if (0 == len(breakpoint_list)): #{
      LogMsg(self, "  No %s breakpoints." % description)
      return
    #} end if
    num = dict(((type, 0) for type in ["common", "comm_mm", "comm_repeat",
      "comm_frepeat", "comm_untran", "comm_local", "comm_secondary",
      "barnacle", "barn_mm", "barn_repeat", "barn_frepeat", "barn_untran",
      "barn_local", "barn_wide", "barn_secondary", "defuse",
      "defuse_repeat_good", "defuse_untran_good", "defuse_poor",
      "defuse_repeat_poor", "defuse_untran_poor", "defuse_same_gene",
      "defuse_dup_gene",]))
    breakpoint_iter = TraverseMultiDict(breakpoint_list, sort=True,
        sort_key=BreakpointSortKeyCls(self.log_info))
    for (index, breakpoint) in enumerate(breakpoint_iter): #{
      if (breakpoint.Common()): #{
        self.CountCommonBreakpoint(index, breakpoint, num)
      elif (breakpoint.BarnacleOnly()):
        self.CountBarnacleBreakpoint(index, breakpoint, num)
      elif (breakpoint.OtherOnly()):
        self.CountDefuseBreakpoint(index, breakpoint, num)
      else:
        LogMsg(self, "  MYSTERY BREAKPOINT! %s" % breakpoint.ToString())
      #} end if
      DebugMsg(self, "    %s" % breakpoint.ToString())
      self.outfiles[output_file_key].WriteLine("%i %s" % (index,
        breakpoint.ToString()))
    #} end for
    #self.outfiles['breakpoints'].WriteLine("\nCommon breakpoints: %i" %
    #  num["common"])
    #self.outfiles['breakpoints'].WriteLine("Barnacle specific breakpoints: "
    #  "%i" % num["barnacle"])
    #self.outfiles['breakpoints'].WriteLine("Defuse specific breakpoints "
    #  "with qual > 2: %i" % num["defuse"])
    self.outfiles[output_file_key].WriteLine("\n".join(["",
      "Common breakpoints: %i (%s)" % (num["common"], StrJoin(", ",
        ("%i %s" % (num[type], desc) for (type, desc) in
        [("comm_mm", "multi-map"),
         ("comm_repeat", "near repeats"),
         ("comm_frepeat", "fully in repeats"),
         ("comm_untran", "untranscribed"),
         ("comm_local", "local event"),
         ("comm_secondary", "secondary"),
        ]))),
      "Barnacle specific breakpoints: %i (%s)" % (num["barnacle"],
        StrJoin(", ", ("%i %s" % (num[type], desc) for (type, desc) in
        [("barn_mm", "multi-map"),
         ("barn_repeat", "near repeats"),
         ("barn_frepeat", "fully in repeats"),
         ("barn_untran", "untranscribed"),
         ("barn_local", "local event"),
         ("barn_wide", "wide-spread"),
         ("barn_secondary", "secondary"),
        ]))),
      "Defuse specific breakpoints:\n  with qual > 2: %i (%s)\n  "
        "with qual <= 2: %i (%s)\n  in single gene: %i (%s)" % (num["defuse"],
        StrJoin(", ", ("%i %s" % (num[type], desc) for (type, desc) in
        [("defuse_untran_good", "untranscribed"),
         ("defuse_repeat_good", "in repeats"),
        ])), num["defuse_poor"], StrJoin(", ", ("%i %s" %
        (num[type], desc) for (type, desc) in
        [("defuse_untran_poor", "untranscribed"),
         ("defuse_repeat_poor", "in repeats"),
        ])), num["defuse_same_gene"], StrJoin(", ", ("%i %s" %
        (num[type], desc) for (type, desc) in
        [("defuse_dup_gene", "interchromosomal"), ]))),
      #"Common breakpoints: %i (%i multi-map, %i untranscribed)" %
      #  (num["common"], num["comm_mm"], num["comm_untran"]),
      #"Barnacle specific breakpoints: %i (%i multi-map, %i untranscribed, "
      #  "%i wide)" % (num["barnacle"], num["barn_mm"],
      #    num["barn_untran"], num["barn_wide"]),
      #"Defuse specific breakpoints with qual > 2: %i (%i untranscribed)" %
      #  (num["defuse"], num["barn_untran"]),
    ]))
  #} end def

  def DisplayLoci(self, loci_list): #{
    if not self.log_info.get('extreme_debug', False): #{
      return
    #} end if
    LogMsg(self, "Displaying loci...")
    loci_iter = TraverseMultiDict(loci_list, sort=True,
        sort_key=BreakpointSortKeyCls())
    breakpoint_index = 0
    for (locus_index, locus) in enumerate(loci_iter): #{
      locus.SetComparisonResult()
      LogMsg(self, "Displaying locus %i (%s)\n  %s" % (locus_index,
        locus.ltype, locus.ToString()))
      breakpoint_iter = TraverseMultiDict(locus.breakpoints, sort=True,
          sort_key=BreakpointSortKeyCls())
      for (breakpoint_index, breakpoint) in enumerate(
          breakpoint_iter, start=breakpoint_index): #{
        LogMsg(self, "  Breakpoint %i.%i:\n    %s" % (locus_index,
          breakpoint_index, breakpoint.ToString()))
      #} end for
    #} end for
  #} end def

  # outfiles[ltype] = FileBoxCls
  def WriteLocusAnalysis(self, description, loci_list, outfiles): #{
    LogMsg(self, "Writing %s locus analysis..." % description)
    if (0 == len(loci_list)): #{
      LogMsg(self, "  No %s loci." % description)
      return
    #} end if
    num = self.SetupBreakpointCounts()
    max_bp_per_locus = dict((ltype,0) for ltype in COMPARE_RESULTS)
    loci_iter = TraverseMultiDict(loci_list, sort=True,
        sort_key=BreakpointSortKeyCls())
    breakpoint_index = 0
    for (index, locus) in enumerate(loci_iter): #{
      locus.index = index
      locus.SetComparisonResult()
      ExtremeDebugMsg(self, "Writing locus %i (%s)" % (locus.index,
        locus.ltype))
      breakpoint_index = self.AnalyseLocusBreakpoints(outfiles, locus,
        breakpoint_index, num)
      self.AnalyseLocus(locus, num, max_bp_per_locus)
    #} end for
    self.WriteLocusCounts(outfiles, num, max_bp_per_locus)
  #} end def

  def SetupBreakpointCounts(self): #{
    num = dict(((type, dict()) for type in COMPARE_RESULTS))
    for ltype in COMPARE_RESULTS: #{
      num[ltype]["loci"] = 0
    #} end for
    common_types = ["clean", "mm", "repeat", "frepeat", "notmat", "local",
      "secondary", "total"]
    for breakpoint_type in common_types: #{
      num["common"]["common_%s" % breakpoint_type] = 0
    #} end for
    #barnacle_types = ["clean", "mm", "repeat", "frepeat", "untran", "local",
    #  "wide", "secondary", "total"]
    barnacle_types = common_types + ["wide"]
    for breakpoint_type in barnacle_types: #{
      num["common"]["barnacle_%s" % breakpoint_type] = 0
      num["barnacle"]["barnacle_%s" % breakpoint_type] = 0
    #} end for
    #defuse_types = ["good_clean", "good_repeat", "good_notmat", "poor_clean",
    #  "poor_repeat", "poor_notmat", "same_gene", "dup_gene",]
    defuse_types1 = ["clean", "repeat", "notmat", "total"]
    for breakpoint_type in defuse_types1: #{
      qual_list = ["good"] + map(str, range(3))
      for qual in qual_list: #{
        num["common"]["defuse_%s_%s" % (qual, breakpoint_type)] = 0
        num["defuse"]["defuse_%s_%s" % (qual, breakpoint_type)] = 0
      #} end for
    #} end for
    defuse_types2 = ["same_gene", "dup_gene"]
    for breakpoint_type in defuse_types2: #{
      num["common"]["defuse_%s" % breakpoint_type] = 0
      num["defuse"]["defuse_%s" % breakpoint_type] = 0
    #} end for
    #num = dict(((type, 0) for type in ["common", "comm_mm", "comm_repeat",
    #  "comm_frepeat", "comm_untran", "comm_local", "comm_secondary",
    #  "barnacle", "barn_mm", "barn_repeat", "barn_frepeat", "barn_untran",
    #  "barn_local", "barn_wide", "barn_secondary", "defuse",
    #  "defuse_repeat_good", "defuse_untran_good", "defuse_poor",
    #  "defuse_repeat_poor", "defuse_untran_poor", "defuse_same_gene",
    #  "defuse_dup_gene",]))
    common_loci_types = ["one_comm_bp", "all_comm_bp", "one_diff_bp",
      "all_diff_bp"]
    for loci_type in common_loci_types: #{
      num["common"][loci_type] = 0
    #} end for
    other_loci_types = ["single_bp", "multi_bp"]
    for ltype in COMPARE_RESULTS: #{
      for other_loci_type in other_loci_types: #{
        num[ltype][other_loci_type] = 0
      #} end for
    #} end for
    return num
  #} end def

  def AnalyseLocusBreakpoints(self, outfiles, locus,
      breakpoint_index, num): #{
    #outfile_key = "%s_%s" % (outfile_prefix, locus.ltype)
    breakpoint_iter = TraverseMultiDict(locus.breakpoints, sort=True,
        sort_key=BreakpointSortKeyCls())
    (common_found, distinct_found) = (False, False)
    locus.num_bp = 0
    lnum = num[locus.ltype]
    for (breakpoint_index, breakpoint) in enumerate(breakpoint_iter,
        start=breakpoint_index): #{
      breakpoint.SetBreakpointQuality()
      locus.num_bp += 1
      full_id = "%i.%i" % (locus.index, breakpoint_index)
      if (breakpoint.Common()): #{
        self.CountCommonBreakpoint(full_id, breakpoint, lnum)
        common_found = True
      elif (breakpoint.BarnacleOnly()):
        self.CountBarnacleBreakpoint(full_id, breakpoint, lnum)
        distinct_found = True
      elif (breakpoint.OtherOnly()):
        self.CountDefuseBreakpoint(full_id, breakpoint, lnum)
        distinct_found = True
      else:
        LogMsg(self, "  MYSTERY BREAKPOINT! %s" % breakpoint.ToString())
      #} end if
      DebugMsg(self, "    %s" % breakpoint.ToString())
      #self.outfiles[outfile_key].WriteLine("%s %s" % (full_id,
      #  breakpoint.ToString()))
      outfiles[locus.ltype].WriteLine("%s %s" %
        (full_id, breakpoint.ToString()))
    #} end for
    if ("common" == locus.ltype): #{
      if (common_found): #{
        lnum["one_comm_bp"] += 1
        if (not distinct_found): #{
          lnum["all_comm_bp"] += 1
        #} end if
      #} end if
      if (distinct_found): #{
        lnum["one_diff_bp"] += 1
        if (not common_found): #{
          lnum["all_diff_bp"] += 1
        #} end if
      #} end if
    #} end if
    return breakpoint_index
  #} end def

  def CountCommonBreakpoint(self, full_id, breakpoint, num): #{
    LogMsg(self, "  Breakpoint %s is common" % full_id)
    num["common_total"] += 1
    if (not breakpoint.primary): #{
      LogMsg(self, "    but not primary!")
      num["common_secondary"] += 1
      return
    #} end if
    if (breakpoint.local_event): #{
      ExtremeDebugMsg(self, "    but contig can be explained by "
          "local event!")
      num["common_local"] += 1
      return
    #} end if
    if (0 == breakpoint.barnacle_transcribed or
        0 == breakpoint.other_transcribed): #{
      DebugMsg(self, "    but should not be present in mature transcripts "
          "(b:%i, d:%i)" % (breakpoint.barnacle_transcribed,
            breakpoint.other_transcribed))
      num["common_notmat"] += 1
      return
    #} end if
    if (breakpoint.full_repeat): #{
      ExtremeDebugMsg(self, "    but aligns completely within repeat!")
      num["common_frepeat"] += 1
      return
    #} end if
    if (breakpoint.repeats): #{
      ExtremeDebugMsg(self, "    but overlaps or is near repeat!")
      num["common_repeat"] += 1
      return
    #} end if
    if (breakpoint.multi_maps): #{
      ExtremeDebugMsg(self, "    but contig multi-maps!")
      num["common_mm"] += 1
      return
    #} end if
    num["common_clean"] += 1
  #} end def

  def OldCountCommonBreakpoint(self, index, breakpoint, num): #{
    DebugMsg(self, "  Breakpoint %i is common" % index)
    if (breakpoint.primary): #{
      num["common"] += 1
      if (breakpoint.multi_maps): #{
        num["comm_mm"] += 1
      #} end if
      if (breakpoint.local_event): #{
        ExtremeDebugMsg(self, "    contig can be explained by "
            "local event!")
        num["comm_local"] += 1
      #} end if
      if (breakpoint.repeats): #{
        ExtremeDebugMsg(self, "    breakpoint falls within repeat!")
        num["comm_repeat"] += 1
      #} end if
    else:
      LogMsg(self, "  Breakpoint %i is common, but not primary!" % index)
    #} end if
    if (0 == breakpoint.barnacle_transcribed or
        0 == breakpoint.other_transcribed): #{
      DebugMsg(self, "    common breakpoint is untranscribed "
          "(b:%i, d:%i)" % (breakpoint.barnacle_transcribed,
            breakpoint.other_transcribed))
      num["comm_untran"] += 1
    #} end if
  #} end def

  def CountBarnacleBreakpoint(self, full_id, breakpoint, num): #{
    DebugMsg(self, "  Breakpoint %s is Barnacle-specific" % full_id)
    num["barnacle_total"] += 1
    breakpoint.untranscribed = (0 == breakpoint.barnacle_transcribed)
    if (not breakpoint.primary): #{
      DebugMsg(self, "    but not primary!")
      num["barnacle_secondary"] += 1
      for bp_attr in BP_ATTRS: #{
        CheckBreakpointAttribute(breakpoint, bp_attr, self)
      #} end for
      #if (breakpoint.wide_spread): #{
      #  DebugMsg(self, "      (and wide spread!)")
      #} end if
      #if (breakpoint.local_event): #{
      #  ExtremeDebugMsg(self, "     (and contig can be explained by "
      #      "local event!)")
      #} end if
      #if (0 == breakpoint.barnacle_transcribed): #{
      #  DebugMsg(self, "      (and fully untranscribed!)")
      #} end if
      #if (breakpoint.repeats): #{
      #  ExtremeDebugMsg(self, "     (and falls within repeat!)")
      #} end if
      #if (breakpoint.multi_maps): #{
      #  ExtremeDebugMsg(self, "     (and contig multi-maps!)")
      #} end if
      return
    #} end if
    if (breakpoint.wide_spread): #{
      DebugMsg(self, "    but wide spread!")
      num["barnacle_wide"] += 1
      for bp_attr in BP_ATTRS[1:]: #{
        CheckBreakpointAttribute(breakpoint, bp_attr, self)
      #} end for
      #if (breakpoint.local_event): #{
      #  ExtremeDebugMsg(self, "     (and contig can be explained by "
      #      "local event!)")
      #} end if
      #if (0 == breakpoint.barnacle_transcribed): #{
      #  DebugMsg(self, "      (and fully untranscribed!)")
      #} end if
      #if (breakpoint.repeats): #{
      #  ExtremeDebugMsg(self, "     (and falls within repeat!)")
      #} end if
      #if (breakpoint.multi_maps): #{
      #  ExtremeDebugMsg(self, "     (and contig multi-maps!)")
      #} end if
      return
    #} end if
    if (breakpoint.local_event): #{
      ExtremeDebugMsg(self, "    but contig can be explained by "
          "local event!")
      num["barnacle_local"] += 1
      for bp_attr in BP_ATTRS[2:]: #{
        CheckBreakpointAttribute(breakpoint, bp_attr, self)
      #} end for
      #if (0 == breakpoint.barnacle_transcribed): #{
      #  DebugMsg(self, "      (and fully untranscribed!)")
      #} end if
      #if (breakpoint.repeats): #{
      #  ExtremeDebugMsg(self, "     (and falls within repeat!)")
      #} end if
      #if (breakpoint.multi_maps): #{
      #  ExtremeDebugMsg(self, "     (and contig multi-maps!)")
      #} end if
      return
    #} end if
    if (0 == breakpoint.barnacle_transcribed): #{
      DebugMsg(self, "    but should not be present in mature transcripts "
          "(primary: %s)" % breakpoint.primary)
      num["barnacle_notmat"] += 1
      for bp_attr in BP_ATTRS[3:]: #{
        CheckBreakpointAttribute(breakpoint, bp_attr, self)
      #} end for
      #if (breakpoint.repeats): #{
      #  ExtremeDebugMsg(self, "     (and falls within repeat!)")
      #} end if
      #if (breakpoint.multi_maps): #{
      #  ExtremeDebugMsg(self, "     (and contig multi-maps!)")
      #} end if
      return
    #} end if
    if (breakpoint.full_repeat): #{
      ExtremeDebugMsg(self, "    but %s!" %
          BP_ATTR_DESCRIPTIONS["full_repeat"])
      num["barnacle_frepeat"] += 1
      for bp_attr in BP_ATTRS[4:]: #{
        CheckBreakpointAttribute(breakpoint, bp_attr, self)
      #} end for
      return
    #} end if
    if (breakpoint.repeats): #{
      ExtremeDebugMsg(self, "    but falls within repeat!")
      num["barnacle_repeat"] += 1
      for bp_attr in BP_ATTRS[5:]: #{
        CheckBreakpointAttribute(breakpoint, bp_attr, self)
      #} end for
      #if (breakpoint.multi_maps): #{
      #  ExtremeDebugMsg(self, "     (and contig multi-maps!)")
      #} end if
      return
    #} end if
    if (breakpoint.multi_maps): #{
      ExtremeDebugMsg(self, "    but contig multi-maps!")
      num["barnacle_mm"] += 1
      return
    #} end if
    num["barnacle_clean"] += 1
  #} end def

  def OldCountBarnacleBreakpoint(self, index, breakpoint, num): #{
    DebugMsg(self, "  Breakpoint %i is Barnacle-specific" % index)
    # do not count wide spread breakpoints
    if (breakpoint.wide_spread): #{
      DebugMsg(self, "    and wide spread! (not counting)")
      num["barn_wide"] += 1
      return
    #} end if
    if (0 == breakpoint.barnacle_transcribed): #{
      DebugMsg(self, "    Barnacle breakpoint is untranscribed "
          "(primary: %s)" % breakpoint.primary)
      num["barn_untran"] += 1
    #} end if
    if (breakpoint.primary): #{
      DebugMsg(self, "    and primary!")
      num["barnacle"] += 1
      if (breakpoint.multi_maps): #{
        num["barn_mm"] += 1
      #} end if
      if (breakpoint.local_event): #{
        ExtremeDebugMsg(self, "    contig can be explained by "
            "local event!")
        num["barn_local"] += 1
      #} end if
      if (breakpoint.repeats): #{
        ExtremeDebugMsg(self, "    breakpoint falls within repeat!")
        num["barn_repeat"] += 1
      #} end if
    #} end if
  #} end def

  def CountDefuseBreakpoint(self, full_id, breakpoint, num): #{
    DebugMsg(self, "  Breakpoint %s is Defuse-specific" % full_id)
    # TODO: Add deFuse full total here.
    if (breakpoint.same_gene): #{
      if (breakpoint.interchromosomal): #{
        DebugMsg(self, "    but involves an interchromosomal gene duplication")
        num["defuse_dup_gene"] += 1
        return
      #} end if
      DebugMsg(self, "    but involves only a single gene")
      num["defuse_same_gene"] += 1
      return
    #} end if
    # count high and low quality Defuse specific breakpoints separately
    if (None == breakpoint.other_quality or
        3 <= breakpoint.other_quality):
      DebugMsg(self, "    and good quality (>2)!")
      num["defuse_good_total"] += 1
      if (0 == breakpoint.other_transcribed): #{
        DebugMsg(self, "    but should not be present in mature transcripts")
        num["defuse_good_notmat"] += 1
        return
      #} end if
      if (breakpoint.repeats): #{
        ExtremeDebugMsg(self, "    but falls within repeat!")
        num["defuse_good_repeat"] += 1
        return
      #} end if
      num["defuse_good_clean"] += 1
    else:
      DebugMsg(self, "    and poor quality (%i < 3)!" %
        breakpoint.other_quality)
      num["defuse_%i_total" % breakpoint.other_quality] += 1
      if (0 == breakpoint.other_transcribed): #{
        DebugMsg(self, "    but should not be present in mature transcripts")
        num["defuse_%i_notmat" % breakpoint.other_quality] += 1
        return
      #} end if
      if (breakpoint.repeats): #{
        ExtremeDebugMsg(self, "    but falls within repeat!")
        num["defuse_%i_repeat" % breakpoint.other_quality] += 1
        return
      #} end if
      num["defuse_%i_clean" % breakpoint.other_quality] += 1
    #} end if
  #} end def

  def OldCountDefuseBreakpoint(self, index, breakpoint, num): #{
    DebugMsg(self, "  Breakpoint %i is Defuse-specific" % index)
    if (0 == breakpoint.other_transcribed): #{
      DebugMsg(self, "    Defuse breakpoint is untranscribed"
          "(quality: %i)" % breakpoint.other_quality)
      num["defuse_untran"] += 1
    # only count high quality Defuse specific breakpoints
    elif (None == breakpoint.other_quality or
        3 <= breakpoint.other_quality):
      DebugMsg(self, "  and good quality!")
      num["defuse"] += 1
      if (breakpoint.repeats): #{
        ExtremeDebugMsg(self, "    breakpoint falls within repeat!")
        num["defuse_repeat"] += 1
      #} end if
    #} end if
  #} end def

  def AnalyseLocus(self, locus, num, max_bp_per_locus): #{
    # update locus counts
    lnum = num[locus.ltype]
    lnum["loci"] += 1
    #num_bp = len(locus.breakpoints)
    #ExtremeDebugMsg(self, "  NUM BREAKPOINTS IN LOCUS: %i" % locus.num_bp)
    if (1 == locus.num_bp): #{
      lnum["single_bp"] += 1
    elif (1 < locus.num_bp):
      lnum["multi_bp"] += 1
    else:
      raise DefuseComparisonError("Locus %i does not have any breakpoints! "
        "(%i)" % locus.num_bp)
    #} end if
    max_bp_per_locus[locus.ltype] = max(max_bp_per_locus[locus.ltype],
      locus.num_bp)
  #} end def

  def WriteLocusCounts(self, outfiles, num, max_bp_per_locus): #{
    for ltype in COMPARE_RESULTS: #{
      #outfile_key = "%s_%s" % (outfile_prefix, ltype)
      lnum = num[ltype]
      data = [""]
      data.append("Loci: %i (%s), Max breakpoints/locus: %i" % (lnum["loci"],
        StrJoin(", ", ("%i %s" % (lnum[type], desc) for (type, desc) in [
          ("single_bp", "single breakpoint"),
          ("multi_bp", "multiple breakpoint")])),
        max_bp_per_locus[ltype]))
      if ("common" == ltype): #{
        self.AddCommonCounts(data, lnum)
      #} end if
      if (ltype in ["common", "barnacle"]): #{
        self.AddBarnacleCounts(data, lnum)
      #} end if
      if (ltype in ["common", "defuse"]): #{
        self.AddDefuseCounts(data, lnum)
      #} end if
      #self.outfiles[outfile_key].WriteLine("\n".join(["",
      #  "Common breakpoints: %i (%s)" % (lnum["common_total"], StrJoin(", ",
      #    ("%i %s" % (lnum["common_%s" % type], desc) for (type, desc) in
      #    [("clean", "clean"), ("mm", "multi-map"), ("repeat", "near repeats"),
      #     ("frepeat", "fully in repeats"), ("notmat", "not in mature transcripts"),
      #     ("local", "local event"), ("secondary", "secondary"),
      #    ]))),
      #  "Barnacle specific breakpoints: %i (%s)" % (lnum["barnacle_total"],
      #    StrJoin(", ", ("%i %s" % (lnum["barnacle_%s" % type], desc) for
      #    (type, desc) in [("clean", "clean"), ("mm", "multi-map"),
      #     ("repeat", "near repeats"), ("frepeat", "fully in repeats"),
      #     ("notmat", "untranscribed"), ("local", "local event"),
      #     ("wide", "wide-spread"), ("secondary", "secondary"),
      #    ]))),
      #  "Defuse specific breakpoints:",
      #  "  with qual > 2: %i (%s)" % (lnum["defuse_good_total"],
      #    StrJoin(", ", ("%i %s" % (lnum["defuse_good_%s" % type], desc) for
      #    (type, desc) in [("clean", "clean"),
      #     ("notmat", "not in mature transcripts"), ("repeat", "in repeats"),
      #    ]))),
      #  "  with qual <= 2: %i (%s)" % (lnum["defuse_poor_total"],
      #    StrJoin(", ", ("%i %s" % (lnum["defuse_poor_%s" % type], desc) for
      #    (type, desc) in [("clean", "clean"),
      #     ("notmat", "not in mature transcripts"), ("repeat", "in repeats"),
      #    ]))),
      #  "  in single gene: %i (%s)" % (lnum["defuse_same_gene"],
      #    StrJoin(", ", ("%i %s" % (lnum["defuse_%s" % type], desc) for
      #    (type, desc) in [("dup_gene", "interchromosomal"), ]))),
      #]))
      #self.outfiles[outfile_key].WriteLine("\n".join(data))
      outfiles[ltype].WriteLine("\n".join(data))
    #} end for
      #"Common breakpoints: %i (%i multi-map, %i untranscribed)" %
      #  (num["common"], num["comm_mm"], num["comm_untran"]),
      #"Barnacle specific breakpoints: %i (%i multi-map, %i untranscribed, "
      #  "%i wide)" % (num["barnacle"], num["barn_mm"],
      #    num["barn_untran"], num["barn_wide"]),
      #"Defuse specific breakpoints with qual > 2: %i (%i untranscribed)" %
      #  (num["defuse"], num["barn_untran"]),
  #} end def

  def AddCommonCounts(self, data, lnum): #{
    data.extend(["  %i with common breakpoints (%i with only common "
        "breakpoints)" % (lnum["one_comm_bp"], lnum["all_comm_bp"]),
      "  %i with distinct breakpoints (%i with only distinct "
        "breakpoints)" % (lnum["one_diff_bp"], lnum["all_diff_bp"]),
      "Common breakpoints: %i (%s)" % (lnum["common_total"],
        StrJoin(", ", ["%i %s" % (lnum["common_%s" % type], desc) for
        (type, desc) in [("clean", "clean"), ("mm", "multi-map"),
         ("repeat", "near repeats"), ("frepeat", "fully in repeats"),
         ("notmat", "not in mature transcripts"), ("local", "local event"),
         ("secondary", "secondary"),
        ]])
      )
    ])
  #} end def

  def AddBarnacleCounts(self, data, lnum): #{
    data.append("Barnacle specific breakpoints: %i (%s)" %
      (lnum["barnacle_total"], StrJoin(", ", ["%i %s" %
        (lnum["barnacle_%s" % type], desc) for (type, desc) in
        [("clean", "clean"), ("mm", "multi-map"),
         ("repeat", "near repeats"), ("frepeat", "fully in repeats"),
         ("notmat", "not in mature transcripts"), ("local", "local event"),
         ("wide", "wide-spread"), ("secondary", "secondary"),
        ]])
      )
    )
  #} end def

  def AddDefuseCounts(self, data, lnum): #{
    # TODO: Add count of total number of deFuse-specific breakpoints
    #       for the current locus type.
    #data.extend(["Defuse specific breakpoints: (%i)" % ,
    data.extend(["Defuse specific breakpoints:",
      "  with qual > 2: %i (%s)" % (lnum["defuse_good_total"],
        StrJoin(", ", ["%i %s" % (lnum["defuse_good_%s" % type], desc) for
        (type, desc) in [("clean", "clean"), ("repeat", "in repeats"),
         ("notmat", "not in mature transcripts"),
        ]])
      ),
    ])
    for qual in reversed(range(3)): #{
      data.append("  with qual %i (%s): %i (%s)" % (qual,
        QUALITY_DESCRIPTIONS[qual], lnum["defuse_%i_total" % qual],
        StrJoin(", ", ["%i %s" % (lnum["defuse_%i_%s" % (qual, type)], desc)
         for (type, desc) in [("clean", "clean"), ("repeat", "in repeats"),
          ("notmat", "not in mature transcripts"),]
        ])
      ))
    #} end for
    data.append("  in single gene: %i (%s)" % (lnum["defuse_same_gene"],
      StrJoin(", ", ["%i %s" % (lnum["defuse_%s" % type], desc) for
      (type, desc) in [("dup_gene", "interchromosomal"), ]])
    ))
  #} end def

  def DistFromBreakpoint(self): #{
    return int(self.options.align_threshold1 * 1.5)
  #} end def
#} end class

def OverlapBeforePred(ptype, pred, overlap, log_info=None): #{
  if (None == overlap):
    ExtremeDebugMsg(log_info, "Curr overlap not yet set!!")
    return True
  #} end if
  if ("defuse" == ptype): #{
    ExtremeDebugMsg(log_info, "Comparing defuse prediction %i to "
      "overlap %i" % (pred.id, overlap.NumID()))
    return (overlap.NumID() < pred.id)
  elif ("barnacle" == ptype):
    ParseFullEventID(overlap, overlap.query_id)
    ExtremeDebugMsg(log_info, "Comparing barnacle prediction %i%s to "
      "overlap %i%s" % (pred.group_id, pred.candidate_id,
      overlap.group_id, overlap.candidate_id))
    return (overlap.group_id < pred.group_id or
      (overlap.group_id == pred.group_id and
       overlap.candidate_id < pred.candidate_id))
  else:
    raise DefuseComparisonError("unrecognized prediction type: %s" % ptype)
  #} end def
#} end def

def OverlapMatchesPred(ptype, pred, overlap, log_info=None): #{
  if ("defuse" == ptype): #{
    ExtremeDebugMsg(log_info, "Comparing defuse prediction %i to "
      "overlap %i" % (pred.id, overlap.NumID()))
    return (overlap.NumID() == pred.id)
  elif ("barnacle" == ptype):
    ParseFullEventID(overlap, overlap.query_id)
    ExtremeDebugMsg(log_info, "Comparing barnacle prediction %i%s to "
      "overlap %i%s" % (pred.group_id, pred.candidate_id,
      overlap.group_id, overlap.candidate_id))
    return (overlap.group_id == pred.group_id and
       overlap.candidate_id == pred.candidate_id)
  else:
    raise DefuseComparisonError("unrecognized prediction type: %s" % ptype)
  #} end def
#} end def

# breakpoints[chromA][chromB][pos_keyA][pos_keyB] = breakpoint_pair
class BreakpointSortKeyCls(MultiDictSortKeyCls): #{
  def __init__(self, log_info=None): #{
    MultiDictSortKeyCls.__init__(self)
    self.log_info = log_info
  #} end def

  def __call__(self, arg1): #{
    if (2 > self.level): #{
      key = ChrToInt(arg1)
    else:
      key = arg1
    #} end if
    ExtremeDebugMsg(self, "%sLevel: %i, Raw: %s, Returning: %i" %
      (" "*self.level, self.level, arg1, key))
    return key
  #} end def
#} end class

def GetDirectedChromKey(breakpoint): #{
  # chromosome "X" := "Xr" if align to "-" strand
  chrom_key = breakpoint.chr
  if ("-" == breakpoint.strand): #{
    chrom_key += "r"
  #} end if
  return chrom_key
#} end def

def GetBreakpointCoordsTuple(candidate): #{
  sorted_breakpoints = SortBreakpointTuple(candidate.breakpoints)
  return tuple(breakpoint.coord for breakpoint in sorted_breakpoints)
#} end def

def GetContigPart(target_align, other_align): #{
  # get the contig coordinates of the part of the target alignment
  # that does not overlap with the other alignment
  target_coords = CoordPairCls(target_align.ctg_start, target_align.ctg_end)
  other_coords = CoordPairCls(other_align.ctg_start, other_align.ctg_end)
  if (target_coords.min < other_coords.min): #{
    target_coords.SetMax(min(target_coords.max, other_coords.min))
  else:
    target_coords.SetMin(max(target_coords.min, other_coords.max))
  #} end if
  return target_coords
#} end def

def CheckBreakpointAttribute(breakpoint, attr_key, log_info): #{
  if (getattr(breakpoint, attr_key)): #{
    ExtremeDebugMsg(log_info, "      (and %s!)" %
        BP_ATTR_DESCRIPTIONS[attr_key])
  #} end if
#} end def

#### EXCEPTION CLASSES ####
class DefuseComparisonError(MyError): #{
  pass
#} end class

def SetupOptionsParser(): #{
  description_string = ("") #TODO
  args = [ "LIB", "BARNACLE_FILE", "DEFUSE_FILE", "CONTIGS_FILE" ]
  usage_string = "%prog " + " ".join(args) + " [ OPTIONS ]"
  parser = OptionParser(description=description_string,
                        version="%prog " + VERSION,
                        usage=usage_string)
  parser.num_args = len(args)
  parser.add_option("--allow-bad-commons",
    action="store_true",
    help="Allow \"bad\" predictions to be common with \"good\" predictions.")
  parser.add_option("-B", "--big-buffer",
    type="int", metavar="N",
    help="Consider events similar if their breakpoint "
      "coordinates are within Nbp of each other. "
      "[default:%default]")
  parser.add_option("-b", "--small-buffer",
    type="int", metavar="N",
    help="Consider events identical if their breakpoint "
      "coordinates are within Nbp of each other. "
      "[default:%default]")
  parser.add_option("-p", "--probability-threshold1",
    type="float", metavar="F",
    help="Consider Defuse events to be \"medium confidence\" "
      "if the probability field has a value greater than F. "
      "[default:%default]")
  parser.add_option("-P", "--probability-threshold2",
    type="float", metavar="F",
    help="Consider Defuse events to be \"high confidence\" if "
      "the probability field has a value greater than F. "
      "[default:%default]")
  parser.add_option("-c", "--coverage-threshold",
    type="int", metavar="N",
    help="Consider Defuse events to have \"good coverage\" if "
      "the splitr_count field has a value greater than N. "
      "[default:%default]")
  parser.add_option("--genome",
    dest="genome_path",
    help="The path to a fasta or 2bit formatted genome sequences file.")
  parser.add_option("--readthrough-threshold",
    type='float', dest="readthrough_threshold", metavar="FRACTION",
    help="If any single alignment for a Defuse sequence to the genome uses "
      "at least FRACTION of the total contig length, consider that Defuse "
      "sequence to be a potential readthrough. [default: %default]")
  parser.add_option("--readthrough-max-len",
    type="int", dest="max_rthru", metavar="N",
    help="If a Defuse prediction is marked as a deletion of less than N "
      "nt, consider it a readthrough. [default: %default]")
  parser.add_option("-a", "--align",
    action="store_true",
    help="Align Defuse prediction sequences with unique "
      "breakpoints to assembled contigs.")
  parser.add_option("-A", "--align-threshold1",
    type="int", metavar="N",
    help="Consider a Defuse-to-contig alignment \"partial\" "
      "if less than the full Defuse sequence is aligned, but "
      "the alignment overlaps the breakpoint by at least Nbp "
      "on either side. [default:%default]")
  parser.add_option("--align-threshold2",
    type="int", metavar="N",
    help="Consider a Defuse-to-contig alignment \"partial\" "
      "if less than the full Defuse sequence is aligned, but "
      "the full contig is aligned and it overlaps the "
      "breakpoint by at least Nbp on either side. "
      "[default:%default]")
  parser.add_option("--realign-unaligned",
    action="store_true", dest="realign",
    help="Realign Defuse prediction sequences with unique "
      "breakpoints that do not align to the merged contig "
      "assembly to a pre-merging assembly.")
  parser.add_option("--realign-base",
    dest="realign_dir", metavar="DIR",
    help="The base directory holding pre-merging assembled "
      "contig sequence files to use during realignment.")
  parser.add_option("--low-k",
    type="int",
    help="The minimum k-value pre-merging assembly to use "
      "when realigning. [default: %default]")
  parser.add_option("--high-k",
    type="int",
    help="The maximum k-value pre-merging assembly to use "
      "when realigning. [default: %default]")
  parser.add_option("--include-bubbles",
    action="store_true",
    help="When realigning unique Defuse prediction sequences "
      "to pre-merging assembled contigs, also align to "
      "\"-bubbles.fa\" file. [default]")
  parser.add_option("--exclude-bubbles",
    action="store_false", dest="include_bubbles",
    help="When realigning unique Defuse prediction sequences "
      "to pre-merging assembled contigs, do not align to "
      "\"-bubbles.fa\" file.")
  parser.add_option("--include-indels",
    action="store_true",
    help="When realigning unique Defuse prediction sequences "
      "to pre-merging assembled contigs, also align to "
      "\"-indels.fa\" file. [default]")
  parser.add_option("--exclude-indels",
    action="store_false", dest="include_indels",
    help="When realigning unique Defuse prediction sequences "
      "to pre-merging assembled contigs, do not align to "
      "\"-indels.fa\" file.")
  parser.add_option("--exclude-genes",
    action="append", dest="exclude_genes_paths", metavar="PATH",
    help="Set aside predictions involving genes in provided gene "
      "list file. This option can appear multiple times to split "
      "predictions into multiple subsets (based on genes).")
  parser.add_option("--local-event-ctgs",
    dest="local_event_ctgs_path", metavar="PATH",
    help="Discard predictions involving contigs in provided contig "
      "list file.")
  parser.add_option("--separate-local",
    action="store_true",
    help="Do not consider contigs that could be explained by local events "
      "when determining common loci.")
  parser.add_option("--repeat-coords",
    action="append", dest="repeat_paths", metavar="PATH",
    help="The path to a BED-formatted file of repeat coordinates to use to "
      "screen Defuse predictions. This option can appear multiple times to "
      "screen multiple repeat-coordinate files.")
  parser.add_option("--repeat-seqs",
    dest="repeat_seqs_path", metavar="PATH",
    help="The path to a fasta-formatted file of repeat sequences to align "
      "the contig parts against, to check whether any contig parts align "
      "fully within a repeat.")
  parser.add_option("--disable-profiling-timer",
    action="store_true", dest="dpt",
    help="Sometimes this script can hang when trying to spawn "
         "child processes, due to the kernel's profiling "
         "timer. Use this option to disable the profiling "
         "timer if the script seems to be hanging.")
  parser.add_option("-f", "--force",
    action="store_true",
    help="Force comparison to take place, even if the output "
         "directory already exists.")
  parser.add_option("-d", "--debug",
    action="store_true",
    help="Print debug information while the program runs.")
  parser.add_option("--extreme-debug",
    action="store_true", dest="extreme_debug",
    help="Print extremely in-depth debug information while "
      "the program runs. Not recommended for large jobs.")
  parser.set_defaults(allow_bad_commons=False,
                      big_buffer=500000,
                      small_buffer=50,
                      probability_threshold1=0.75,
                      probability_threshold2=0.9,
                      coverage_threshold=3,
                      readthrough_threshold=0.95,
                      max_rthru=200000,
                      align=False,
                      align_threshold1=23,
                      align_threshold2=19,
                      realign=False,
                      low_k=26,
                      high_k=50,
                      include_bubbles=True,
                      include_indels=True,
                      separate_local=False,
                      dpt=False,
                      force=False,
                      debug=False,
                      extreme_debug=False)
  return parser
#} end def

def CheckPaths(options): #{
  opts_good = True
  if (not opts_good): #{
    ErrMsg("bad option") #TODO
    opts_good = False
  #} end if
  if (options.align_threshold1 < options.align_threshold2): #{
    ErrMsg("Alignment threshold 1 must be larger than alignment threshold 2")
    opts_good = False
  #} end if
  path_errors = list()
  CheckFilePath(options.barnacle_path, "Barnacle results",  path_errors)
  CheckFilePath(options.defuse_path,   "Defuse results",    path_errors)
  CheckFilePath(options.contigs_path,  "assembled contigs", path_errors)
  # get and check the output path
  options.output_dir = GetOutDir(os.path.dirname(options.barnacle_path),
    "defuse_comparison")
  if (options.realign): #{
    # need to do original alignment if realigning
    options.align = True
    if (options.low_k > options.high_k): #{
      ErrMsg("minimum k-value (%i) must not be greater than maximum k-value "
        "(%i)" % (options.low_k, options.high_k))
      opts_good = False
    #} end if
    if (None == options.realign_dir): #{
      ErrMsg("you must provide the path to a directory holding pre-merging "
        "assembly contig sequence files (--realign-base) when using the "
        "\"--realign\" option.")
      opts_good = False
    else:
      CheckDirPath(options.realign_dir, "pre-merging assembly base",
        path_errors)
    #} end if
    realign_out_dir = os.path.join(options.output_dir, "unaligned_realign")
    options.realign_target_paths = dict()
    options.realign_output_paths = dict()
    for kvalue in range(options.low_k, options.high_k+1): #{
      SetupRealignPath(options, realign_out_dir, kvalue, "contigs",
        path_errors)
      if (options.include_bubbles): #{
        SetupRealignPath(options, realign_out_dir, kvalue, "bubbles",
          path_errors)
      #} end if
      if (options.include_indels): #{
        SetupRealignPath(options, realign_out_dir, kvalue, "indel",
          path_errors)
      #} end if
    #} end for
  #} end if
  if (None != options.genome_path): #{
    options.align_to_genome_path = os.path.join(options.output_dir,
      "defuse_to_genome.psl")
  #} end if
  if (options.align): #{
    options.align_to_contigs_path = os.path.join(options.output_dir,
      "defuse_to_contigs.psl")
  #} end if
  if (None != options.local_event_ctgs_path): #{
    CheckFilePath(options.local_event_ctgs_path, "local event contig list",
        path_errors)
  #} end if
  if (None != options.repeat_paths): #{
    for repeat_path in options.repeat_paths: #{
      CheckFilePath(repeat_path, "repeat coordinates path", path_errors)
    #} end for
  #} end if
  if (opts_good and 0 == len(path_errors)): #{
    CheckDirPath(options.output_dir, "comparison output", path_errors,
      create=True, replace=options.force)
    # get the log file name
    options.log_file_name = GetLogPath(options.barnacle_path,
      "defuse_comparison", options.output_dir)
  #} end if
  if (opts_good and 0 == len(path_errors) and options.realign): #{
    CheckDirPath(realign_out_dir, "realignment output", path_errors,
      create=True, replace=True)
  #} end if
  if (0 < len(path_errors)): #{
    ErrMsg("Errors in input arguments:\n  %s" % "\n  ".join(path_errors))
  #} end if
  # the paths are good if there are no path errors and no conflicting options
  return (opts_good and 0 == len(path_errors))
#} end def

def SetupRealignPath(options, realign_out_dir, kvalue, contig_type,
    path_errors): #{
  key = "%i-%s" % (kvalue, contig_type)
  options.realign_target_paths[key] = os.path.join(options.realign_dir,
    "k%i" % kvalue, "%s-%s.fa" % (options.lib, contig_type))
  CheckFilePath(options.realign_target_paths[key],
    "k%i %s assembly" % (kvalue, contig_type), path_errors)
  options.realign_output_paths[key] = os.path.join(realign_out_dir,
    "defuse_to_k%i-%s.psl" % (kvalue, contig_type))
#} end def

def Main(): #{
  # get options and arguments
  parser = SetupOptionsParser()
  (options, args) = parser.parse_args()
  # if the right number of args was used
  if (parser.num_args == len(args)): #{
    options.lib           = args[0]
    options.barnacle_path = EnsureAbsPath(args[1])
    options.defuse_path   = EnsureAbsPath(args[2])
    options.contigs_path  = EnsureAbsPath(args[3])
    if (CheckPaths(options)): #{
      try: #{
        compare_boss = DefuseComparisonCls(options)
        WriteCommand(compare_boss, sys.argv)
        compare_boss.RunComparison()
      except (MyError), e:
        ErrMsg("ERROR while comparing results:\n  %s" % e)
        return ES_RUN_ERR
      #} end try
    else:
      return ES_PATH_ERR
    #} end if
  else:
    parser.error("you must specify a library name (LIB); the path to a "
      "Barnacle data file for that library (BARNACLE_FILE); the path to a "
      "Defuse results file for that library (DEFUSE_FILE); and the path to "
      "a fasta-formatted assembled contig sequences file (CONTIGS_FILE).")
    return ES_OPT_ERR
  #} end if
  return ES_SUCCESS
#} end def

if __name__ == '__main__': #{
  try: #{
    exit_status = Main()
  except Exception, e:
    traceback.print_exc()
    exit_status = ES_EXCEPTION
  #} end try
  if (ES_SUCCESS == exit_status): #{
    print MSG_SUCCESS
  else:
    print MSG_FAIL
  #} end if
  sys.exit(exit_status)
#} end if
