#! /usr/bin/env python
"""
alignment.py

Created by Readman Chiu
Edited by Lucas Swanson
Copyright (c) 2012 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

#import standard modules
from __future__ import division
import math, re, sys
from operator import attrgetter

# import custom modules
from utils.error import MyError
from utils.general import ReverseComplement
from utils.messages import ErrMsg, ExtremeDebugMsg, LogMsg
from utils.files_paths import FileBoxCls
from utils.common_constants import FIRST, LAST
from common.aligned_coord_pair import AlignedCoordPairListCls


# TODO: Change all log messages in member functions to extreme debug messages.
# TODO: Some of these member functions should move to the Alignment class.
class ContigAlignmentParser:
    def __init__(
            self, file_path, filters, noblocks, minimum, noline, refseq,
            log_info=None):
        self.file = FileBoxCls(file_path, 'r')
        self.filters = clean_filters(filters)
        self.noblocks = noblocks
        self.minimum = minimum
        self.noline = noline
        self.refseq = refseq
        self.log_info = log_info
        self.curr_line = None
        self.curr_contig_id = None
        # TODO: Remove aligns data member, once properly using 'yield'.
        self.aligns = list()
        self.raw_aligns_for_contig = list()

    def __del__(self):
        if hasattr(self, 'file'):
            self.file.close()

    def is_header(self, line):
        """Return true if line is a header line."""
        raise NotImplementedError

    def parse_header(self):
        """Parse any header lines that may exist in the file.

        May raise StopIteration.

        """
        raise NotImplementedError

    # TODO: Refactor this so that 'self.file.next()' is not called
    #       in so many places.
    def process_body_line(self, robust_mode=False):
        """Process one body line (alignment record) from the file."""
        ExtremeDebugMsg(self, self.curr_line)
        align_cols = self.curr_line.split("\t")
        if self.align_is_incomplete(align_cols):
            raise ContigAlignmentParserError(
                'Incomplete record: {0}'.format(self.curr_line))
        # Skip queries that are not actually aligned anywhere.
        if self.query_is_unaligned(align_cols):
            ExtremeDebugMsg(self, 'Query is unaligned!')
            self.curr_line = self.file.next()
            return
        try:
            align = self.create_align(self.curr_line)
        except DiscardAlignmentError, e:
            LogMsg(self, 'Discarding alignment.\n{0}'.format(e))
            self.curr_line = self.file.next()
            return
        except Exception, e:
            if robust_mode:
                LogMsg(
                    self, 'WARNING: Cannot parse alignment, it will be '
                    'discarded:\n{0}\n'.format(self.curr_line, e))
                self.curr_line = self.file.next()
                return
            else:
                raise ContigAlignmentParserError(
                    'Error creating alignment for record:\n{0}\n{1}'.format(
                        self.curr_line, e))
        if align is None:
            raise ContigAlignmentParserError(
                'Could not create alignment from line:\n{0}'.format(align))
        if not self.filters:
            # TODO: 'yield' align, rather than appending.
            self.aligns.append(align)
            self.curr_line = self.file.next()
            return
        if self.query_should_be_excluded(align):
            self.curr_line = self.file.next()
            return
        # This checks if we are starting to load alignments for the 'next'
        # contig, so we need to finish everything for the 'current' contig.
        if self.all_records_for_query_are_loaded(align):
            self.process_alignment_group()
        self.raw_aligns_for_contig.append(align)
        self.curr_contig_id = align.query
        self.curr_line = self.file.next()

    def align_is_incomplete(self, align_cols):
        """Return true if the current record is obviously an incomplete
        alignment (indicating that the alignment process was interrupted).

        """
        raise NotImplementedError

    def query_is_unaligned(self, align_cols):
        """Return true if the current line is for a query that is not
        actually aligned to any position in the target.

        """
        raise NotImplementedError

    def query_should_be_excluded(self, align):
        """Return true if the 'exclude' filter is being used and the query
        for the current alignment is in the list to be excluded.

        """
        if ('exclude' in self.filters and
            align.query in self.filters['exclude']):
            return True
        return False

    def all_records_for_query_are_loaded(self, align):
        """Return true if all alignment records for the current query have
        been loaded from the alignment file, indicated by the query id for
        the most recently loaded alignment being different from the query
        id for the previous alignment.

        """
        if (self.curr_contig_id is not None and
            self.curr_contig_id != align.query):
            return True
        return False

    # TODO: Change all log messages in apply_filters to extreme debug messages.
    def apply_filters(self):
        # Return a list of the alignments from raw_aligns_for_contig that
        # pass all the provided filter settings. This list will be sorted
        # by the 'rank by' values of the alignments.
        #
        # We should not be calling 'apply_filters' if filters is None or
        # an empty list.
        assert self.filters is not None and len(self.filters)
        self.prepare_to_rank_alignments()
        # If the 'unique' filter is enabled and the current query has more than
        # one alignment with the best 'rank by' value, then no alignments are
        # returned for the current query.
        if self.query_fails_unique_filter():
            return list()
        # Associate a rank with each alignment line.
        self.rank_alignments()
        # TODO: Will not need 'keep', once using 'yield'.
        keep = []
        # Not using enumerate for count, because we only want to increment count
        # when an alignment passes all filters.
        count = 1
        # Loop through the sorted alignments, from highest to lowest ranked.
        for align in self.raw_aligns_for_contig:
            ExtremeDebugMsg(self, 'Processing alignment of {0}'.format(
                align.query))
            ExtremeDebugMsg(self, '  LINE: {0}'.format(align.raw_record))
            # Alignments are sorted by rank (actually, by 'rank by' value), so
            # once we see one alignment that is ranked lower than the filter,
            # no other alignments will pass the 'bestn' filter.
            ExtremeDebugMsg(self, 'RANK: {0}'.format(align.rank))
            if self.count_limit_met(count):
                break
            if self.align_fails_bestn_filter(align):
                break
            if self.align_fails_qlen_filter(align):
                break
            # For the remaining filters, only fail the current alignment.
            if self.align_fails_match_filter(align):
                continue
            if self.align_fails_ungapped_filter(align):
                # Use continue here to skip gapped alignments, which do not
                # necessarily come after all the ungapped ones!
                continue
            if self.align_fails_identity_filter(align):
                # Use continue here, since alignments are not necessarily
                # sorted by identity.
                continue
            count += 1
            ExtremeDebugMsg(self, 'Keeping #{0}'.format(align.input_order))
            # TODO: This should probably be 'yield align'.
            keep.append(align)
        return keep

    def prepare_to_rank_alignments(self):
        for i, align in enumerate(self.raw_aligns_for_contig):
            align.input_order = i
            align.rank = -1
        ExtremeDebugMsg(self, 'SCORES: {0}'.format([
            align.rank_by for align in self.raw_aligns_for_contig]))
        ExtremeDebugMsg(self, 'META_GROUP INDICES pre-sort: {0}'.format([
            align.input_order for align in self.raw_aligns_for_contig]))
        # Sort the alignments by their 'rank by' values.
        self.raw_aligns_for_contig.sort(
            key=attrgetter('rank_by'), reverse=True)
        ExtremeDebugMsg(self, 'META_GROUP INDICES post-sort: {0}'.format([
            align.input_order for align in self.raw_aligns_for_contig]))

    def rank_alignments(self):
        # Assign a rank to each alignment, where the alignment with the
        # highest 'rank by' value has rank 1, and alignments with the same
        # 'rank by' value have the same rank.
        #
        # Assumes that alignments have been sorted by their 'rank_by' values.
        #
        # The alignments only need to be ranked if the 'bestn' filter is
        # enabled and there are enough alignments for it to matter.
        if ('bestn' not in self.filters or
            len(self.raw_aligns_for_contig) < self.filters['bestn']):
            return
        rank = 1
        prev_rank_by_value = None
        for align in self.raw_aligns_for_contig:
            # Do not bother applying ranks higher than the 'bestn' filter.
            if 'bestn' in self.filters and rank > self.filters['bestn']:
                break
            ExtremeDebugMsg(self, 'PREV: {0}'.format(prev_rank_by_value))
            if (prev_rank_by_value is not None and
                align.rank_by != prev_rank_by_value):
                rank += 1
            align.rank = rank
            prev_rank_by_value = align.rank_by
        ExtremeDebugMsg(self, 'RANKS: {0}'.format([
            align.rank for align in sorted(
                self.raw_aligns_for_contig, key=attrgetter('input_order'))]))

    def query_fails_unique_filter(self):
        # If the 'unique' filter is enabled and the current query has more
        # than one alignment with the best 'rank by' value, then no
        # alignments are returned for the current query.
        #
        # This function assumes that the alignments have already been
        # sorted by their 'rank by' values.
        if 1 < len(self.raw_aligns_for_contig):
            ExtremeDebugMsg(self, '1st: {0}\n2nd: {1}'.format(
                self.raw_aligns_for_contig[0].rank_by,
                self.raw_aligns_for_contig[1].rank_by))
        if ('unique' in self.filters and
            1 < len(self.raw_aligns_for_contig) and
            self.raw_aligns_for_contig[0].rank_by ==
            self.raw_aligns_for_contig[1].rank_by):
            return True
        return False

    def count_limit_met(self, count):
        # This filter sets a limit on the number of alignments returned
        # for each query. Since the alignments are sorted by 'rank by'
        # value, every alignment returned will have a 'rank by' value at
        # least as good as that of any alignments not returned.
        if ('count' in self.filters and
            self.filters['count'] < count):
            return True
        return False

    def align_fails_bestn_filter(self, align):
        # Alignments are sorted by rank (actually, by 'rank by' value), so
        # once we see one alignment that is ranked lower than the filter,
        # no other alignments will pass the 'bestn' filter.
        if ('bestn' in self.filters and
            self.filters['bestn'] < align.rank):
            return True
        return False

    def align_fails_qlen_filter(self, align):
        # Simply compare the query length (which will be the same for all
        # alignments for a given query) to the filter threshold.
        if ('qlen' in self.filters and
            self.filters['qlen'] > align.query_len):
            return True
        return False

    def align_fails_match_filter(self, align):
        # Compare the percent of query bases that are actually paired with
        # target bases in the alignment to the provided threshold.
        percent_bases_aligned = (
            align.num_bases_aligned / align.query_len) * 100
        if ('match' in self.filters and
            self.filters['match'] > percent_bases_aligned):
            return True
        return False

    def align_fails_identity_filter(self, align):
        # Calculate the percent identity of the alignment and compare it to
        # the threshold, but have a special case for when the threshold is
        # 100.0%.
        if 'identity' in self.filters:
            # When the threshold is 100%, just look at number of mismatches.
            if (100.0 == self.filters['identity'] and
                0 < align.mismatch):
                return True
            if self.filters['identity'] > align.identity:
                return True
        return False

    def align_fails_ungapped_filter(self, align):
        # Checking whether an alignment has 'gaps' is completely different
        # for PSL and SAM formatted alignments, so this member function is
        # abstract.
        raise NotImplementedError

    def create_align(self, record):
        """Create and return an appropriate alignment object from the
        provided record, which should be a string read from the alignment
        file.

        """
        raise NotImplementedError

    def process_alignment_group(self):
        if not self.raw_aligns_for_contig:
            return
        try:
            passing_alignments = self.apply_filters()
        except ValueError, e:
            raise ContigAlignmentParserError(
                'Error filtering contig alignments:\n  {0}\n{1}'.format(
                    '\n  '.join((
                        align.raw_record for align in
                        self.raw_aligns_for_contig)),
                    e))
        if passing_alignments:
            self.aligns.extend(passing_alignments)
        # Clear the list of alignments for the current query.
        self.raw_aligns_for_contig = list()


# TODO: Replace blocks with target_blocks.
class Alignment(AlignedCoordPairListCls): #{

    line = None
    method = None
    query = target = None
    query_len = target_len = None
    target_strand = None
    query_strand = None
    qstart = qend = tstart = tend = None
    blocks = None
    splice_sites = None
    query_blocks = None
    num_bases_aligned = matches = mismatches = None
    identity = None
    score = None
    rank = None
    model = None
    pairwise = None
    psl_str = None
    #transcript orientation inferred from Exonerate alignment
    orient = None
    contig = None

    # TODO: Replace 'method' data member with 'input_format'.
    def __init__(self, method):
        self.method = method
        self.blocks = list()
        self.query_blocks = list()
        self.splice_sites = list()

    def SetIdentity(self):
      self.identity = calc_identity(self.qstart, self.qend, self.tstart,
          self.tend, self.qnuminsert, self.mismatch, self.num_bases_aligned)

    def SetScore(self):
      self.score = calc_score(self.match, self.mismatch,
          self.qnuminsert, self.tnuminsert)

    def set_num_bases_aligned(self):
        # Calculate the number of bases 'paired-up' in the alignment, i.e.
        # matches, mismatches, and repetitive matches. Note that some
        # alignment formats do not have repetitive matches.
        if not hasattr(self, 'repmatch'):
            self.repmatch = 0
        self.num_bases_aligned = self.match + self.mismatch + self.repmatch

    def details(self):
        if self.blocks:
            return "%s %s-%s %s %s-%s %s" % (self.query, self.qstart, self.qend, self.target, self.tstart, self.tend, len(self.blocks))
        else:
            return "%s %s-%s %s %s-%s" % (self.query, self.qstart, self.qend, self.target, self.tstart, self.tend)

    def gff(self, type):
        if not self.blocks:
            print "no blocks", self.query, self.target
            return ""
        #chromosome
        if 'chr' in self.target:
            chrom = self.target
        else:
            chrom = 'chr' + self.target
        #contig
        contig = self.query.split(" ")[0]
        if self.contig and self.contig.num: #{
            contig = str(self.contig.num)
        #strand
        if self.target_strand: #{
            strand = "+"
        else:
            strand = "-"
        #gene orientation
        orient = "?"
        if self.orient: #{
            if self.orient == 'forward': #{
                orient = "+"
            elif self.orient == 'backward': #{
                orient = "-"
        #coverage
        cov = None
        if self.contig: #{
            cov = str(self.contig.normalized_coverage())
        if cov: #{
            gff_group = ":".join([self.target, str(self.tstart), "%s%s" % (strand, orient), contig, str(cov)])
        else:
            gff_group = ":".join([self.target, str(self.tstart), "%s%s" % (strand, orient), contig])
        #do this after recording strand info in gff_group
        if orient != "?": #{
            strand = orient
        out = ""
        for i in range(len(self.blocks)): #{
            qstart, qend = self.query_blocks[i]
            tstart, tend = self.blocks[i]
            splice_site = ""
            if len(self.blocks) > 1 and i < len(self.blocks)-1 and len(self.blocks): #{
                splice_site = self.splice_sites[i]
            #hide query block and splice sites in "program" field
            annot = ":".join([str(qstart), str(qend), splice_site])
            out += "\t".join([chrom, annot, type, tstart, tend, ".", strand, ".", gff_group]) + "\n"
        return out


    def psl(self): #{
        "output ucsc psl format"

        if self.psl_str: #{
            psl_str = self.psl_str
            cols = self.psl_str.split("\t")
            #make sure target has chr in it for UCSC
            if not re.match('^(chr|scaffold)', cols[13], re.IGNORECASE): #{
                cols[13] = 'chr' + cols[13]
            if self.contig: #{
                cov = self.contig.normalized_coverage()
                if cov: #{
                    contig = ":".join([str(self.contig.num), str(cov)])
                    cols[9] = contig
            #gene orientation
            if self.orient: #{
                if self.orient == 'forward': #{
                    orient = "+"
                elif self.orient == 'backward': #{
                    orient = "-"
                cols[8] = orient
            psl_str = "\t".join(cols)
        else:
            #first few fields are not accurate
            fields = []
            #print 'test', self.query, self.target, self.num_bases_aligned, self.matches, self.mismatches, self.tnuminsert, self.bases_del, self.qnuminsert, self.qbaseinsert, self.target_len
            fields.append(str(self.match))
            fields.append(str(self.mismatch))
            #repMatches = not available, use 0
            fields.append('0')
            ncount = 0
            #if refseq:
            #    tseq = refseq.GetSequence(self.target, int(self.tstart), int(self.tend))
            #    ncount = len(re.findall('n', tseq, re.IGNORECASE))
            fields.append(str(ncount))
            fields.append(str(self.qnuminsert))
            fields.append(str(self.qbaseinsert))
            fields.append(str(self.tnuminsert))
            fields.append(str(self.tbaseinsert))
            target = self.target
            if self.target[:3] != 'chr':
                target = 'chr' + self.target
            fields.extend([self.query_strand, self.query, str(self.query_len), str(int(self.qstart)-1), str(self.qend), target, str(self.target_len)])
            #if ("+" == self.query_strand): #{
            #  fields.extend([str(int(self.tstart)-1), str(self.tend)])
            #else:
            #  fields.extend([str(int(self.tend)-1), str(self.tstart)])
            tcoords = sorted(map(int, [self.tstart, self.tend]))
            fields.extend(map(str, [tcoords[0]-1, tcoords[1]]))
            fields.append(str(len(self.blocks)))
            block_sizes = [str(t[1] - t[0] + 1) for t in self.blocks]
            fields.append(','.join(block_sizes) + ',')
            if self.query_strand == '+':
                qstarts = [str(q[0] - 1) for q in self.query_blocks]
            else:
                qstarts = [str(self.query_len - q[0]) for q in self.query_blocks]
            fields.append(','.join(qstarts) + ',')
            tstarts = [str(t[0] - 1) for t in self.blocks]
            fields.append(','.join(tstarts) + ',')
            psl_str = "\t".join(fields)
        return psl_str

    def exon(self): #{
        "output original exonerate record"

        if self.record: #{
            return self.record + "\n"


    def correct_blocks(self, splice_motifs, target_seq, query_seq):
        if not self.splice_sites or not splice_motifs: #{
            return False

        #fix fake insertion
        self.correct_unaligned(query_seq, target_seq)

        #fix single gaps
        self.correct_single_gaps(splice_motifs, target_seq)

        #fix neighboring gaps
        self.correct_neighbor_gaps(splice_motifs, target_seq)


    def correct_single_gaps(self, splice_motifs, target_seq): #{
        gaps = {}

        for i in range(len(self.blocks)-1): #{
            ss = self.splice_sites[i]

            if ss and not splice_motifs.has_key(ss) and not splice_motifs.has_key(ReverseComplement(ss).lower()): #{
                if abs(self.query_blocks[i+1][0] - self.query_blocks[i][1]) == 1: #{
                    gaps[i] = 0

        if gaps: #{
            target_blocks = self.blocks[:]
            query_blocks = self.query_blocks[:]
            splice_sites = self.splice_sites[:]

            gap_indices = gaps.keys()
            gap_indices.sort(lambda x,y: x-y)

            for i in gap_indices: #{
                tblock1 = target_blocks[i][:]
                tblock2 = target_blocks[i+1][:]
                qblock1 = query_blocks[i][:]
                qblock2 = query_blocks[i+1][:]
                splice_site = self.fix_single_gap(tblock1, tblock2, qblock1, qblock2, splice_motifs, target_seq, self.query_strand)

                if splice_site: #{
                    sys.stderr.write("%s changed blocks %s %s to %s %s\n" % (self.query, self.target, target_blocks[i], tblock1, splice_site))
                    sys.stderr.write("%s changed blocks %s %s to %s %s\n" % (self.query, self.target, target_blocks[i+1], tblock2, splice_site))
                    target_blocks[i] = tblock1
                    target_blocks[i+1] = tblock2
                    query_blocks[i] = qblock1
                    query_blocks[i+1] = qblock2
                    splice_sites[i] = splice_site

            if target_blocks != self.blocks: #{
                self.blocks = target_blocks[:]
                self.query_blocks = query_blocks[:]
                self.splice_sites = splice_sites
                if not self.mismatch or int(self.mismatch) == 0: #{
                    self.mismatch = 1

    def correct_neighbor_gaps(self, splice_motifs, target_seq): #{
        gaps = {}

        for i in range(len(self.blocks)-1): #{
            ss = self.splice_sites[i]

            if ss and not splice_motifs.has_key(ss) and not splice_motifs.has_key(ReverseComplement(ss).lower()): #{
                if abs(self.query_blocks[i+1][0] - self.query_blocks[i][1]) == 1: #{
                    gaps[i] = 0

        if gaps: #{
            target_blocks = self.blocks[:]
            query_blocks = self.query_blocks[:]
            splice_sites = self.splice_sites[:]

            gap_indices = gaps.keys()
            gap_indices.sort(lambda x,y: x-y)

            #fix by moving exon and then shuffle
            replaced = {}
            replaced_ordered = []
            for i in range(len(gap_indices)-1): #{
                i1 = gap_indices[i]
                i2 = i1 + 1

                if gaps.has_key(i2) and gaps[i2] != 'fixed' and i2 + 1 < len(target_blocks): #{
                    tblock1 = target_blocks[i1][:]
                    tblock2 = target_blocks[i2][:]
                    tblock3 = target_blocks[i2+1][:]
                    qblock1 = query_blocks[i1]
                    qblock2 = query_blocks[i2][:]
                    qblock3 = query_blocks[i2+1][:]

                    splice_site = self.fix_neighbor_gaps(tblock1, tblock2, tblock3, qblock1, qblock2, qblock3, splice_motifs, target_seq, self.query_strand)

                    if splice_site: #{
                        idx = ' '.join((str(i1), str(i2), str(i2+1)))
                        replaced[idx] = tblock1, tblock3, qblock1, qblock3, splice_site
                        gaps[i1] += 1
                        gaps[i2] += 1
                        replaced_ordered.append(idx)

            #make sure delete from back to front
            replaced_ordered.reverse()
            for indices in replaced_ordered: #{
                new_blocks = replaced[indices]

                ok = True
                for index in indices.split(' '): #{
                    if gaps.has_key(int(index)) and gaps[int(index)] > 1: #{
                        ok = False
                        break

                if ok: #{
                    idx = [int(i) for i in indices.split(' ')]
                    sys.stderr.write("%s changed blocks %s %s to %s %s\n" % (self.query, self.target, self.blocks[idx[0]], new_blocks[0], new_blocks[-1]))
                    sys.stderr.write("%s changed blocks %s %s to %s %s\n" % (self.query, self.target, self.blocks[idx[2]], new_blocks[1], new_blocks[-1]))
                    sys.stderr.write("%s removed block %s %s\n" % (self.query, self.target, self.blocks[idx[1]]))
                    target_blocks[idx[0]] = new_blocks[0]
                    target_blocks[idx[2]] = new_blocks[1]
                    query_blocks[idx[0]] = new_blocks[2]
                    query_blocks[idx[2]] = new_blocks[3]
                    splice_sites[idx[0]] = new_blocks[-1]
                    del target_blocks[idx[1]]
                    del query_blocks[idx[1]]
                    del splice_sites[idx[1]]

            if target_blocks != self.blocks: #{
                self.blocks = target_blocks[:]
                self.query_blocks = query_blocks[:]
                self.splice_sites = splice_sites

                if not self.mismatch or int(self.mismatch) == 0: #{
                    self.mismatch = 1


    def correct_unaligned(self, query_seq, target_seq, max_diff=5): #{
        """tries to fix unaligned query sequence because of snvs"""

        #print self.query_blocks
        #print self.blocks
        #print self.splice_sites

        expand= {}
        for i in range(len(self.blocks)-1): #{
            if abs(self.query_blocks[i+1][0] - self.query_blocks[i][1]) != 1 and abs(self.blocks[i+1][0] - self.blocks[i][1]) != 1: #{
                if self.query_strand == '+': #{
                    qseq = query_seq[self.query_blocks[i][1]:self.query_blocks[i+1][0]-1]
                else:
                    qseq = ReverseComplement(query_seq[self.query_blocks[i+1][0]:self.query_blocks[i][1]-1]).lower()

                tseq = target_seq[self.blocks[i][1]:self.blocks[i+1][0]-1].lower()

                gap_seq = tseq[len(qseq):]
                ss_start = gap_seq[:2] + gap_seq[-2:]
                gap_seq = tseq[:-1*len(qseq)]
                ss_end = gap_seq[:2] + gap_seq[-2:]

                #if gap in target sequence greater than gap in query sequence, method doesn't apply
                #also don't apply if target gap is less than 20bp(arbitrary) of query seq
                #print "test", i, self.blocks[i], self.blocks[i][1], self.blocks[i+1][0]-1, qseq, tseq
                if len(qseq) > len(tseq) or len(tseq) - len(qseq) < 20: #{
                    continue

                #find differences
                start_diff = 0
                for j in range(len(qseq)): #{
                    if qseq[j].lower() != tseq[j].lower(): #{
                        start_diff += 1

                end_diff = 0
                for j in range(len(qseq)): #{
                    if qseq[j].lower() != tseq[-1*len(qseq):][j].lower(): #{
                        end_diff += 1

                if start_diff < end_diff and start_diff < max_diff: #{
                    expand[i] = ['left', len(qseq)]
                    self.mismatch = start_diff
                elif end_diff < start_diff and end_diff < max_diff: #{
                    self.mismatch = end_diff
                    expand[i] = ['right', len(qseq)]

        for (idx, where) in expand.iteritems(): #{
            if where[0] == 'left': #{
                sys.stderr.write("%s changed blocks %s\n" % (self.query, self.blocks[idx]))
                if self.query_strand == '+': #{
                    self.query_blocks[idx][1] += where[1]
                else:
                    self.query_blocks[idx][1] -= where[1]
                self.blocks[idx][1] += where[1]
                self.splice_sites[idx] = ss_start
            else:
                sys.stderr.write("%s changed blocks %s\n" % (self.query, self.blocks[idx+1]))
                if self.query_strand == '+': #{
                    self.query_blocks[idx+1][0] -= where[1]
                else:
                    self.query_blocks[idx+1][0] += where[1]
                self.blocks[idx+1][0] -= where[1]
                self.splice_sites[idx] = ss_end

        #print self.query_blocks
        #print self.blocks
        #print self.splice_sites


    def fix_single_gap(self, tblock1, tblock2, qblock1, qblock2, splice_motifs, target_seq, query_strand): #{
        min_size = 10
        max_size = 10000
        shuffle_sizes = [-2, -1, 1, 2]

        tgap = [tblock1[1]+1, tblock2[0]-1]
        tsize = tgap[1] - tgap[0] + 1

        if tsize < min_size or tsize > max_size: #{
            return False

        artefact = False
        ambiguous = False
        splice_site = None
        for size in shuffle_sizes: #{
            #if shuffling size is greater than or equal to block size, can't shuffle
            if abs(size) >= (tblock1[1] - tblock1[0]) or abs(size) >= (tblock2[1] - tblock2[0]): #{
                continue

            coord = tgap[0] + size, tgap[1] + size
            gap_seq = target_seq[coord[0]-1:coord[1]]

            ss = gap_seq[:2] + gap_seq[-2:]

            if splice_motifs.has_key(ss.lower()) or splice_motifs.has_key(ReverseComplement(ss).lower()): #{
                if not artefact: #{
                    artefact = True
                else:
                    ambiguous = True

                if not ambiguous: #{
                    tblock1[1] += size
                    tblock2[0] += size

                    if query_strand == '+': #{
                        qblock1[1] += size
                        qblock2[0] += size
                    else:
                        qblock1[1] -= size
                        qblock2[0] -= size

                    splice_site = ss

        if artefact and not ambiguous: #{
            return splice_site
        else:
            return False

    def fix_neighbor_gaps(self, tblock1, tblock2, tblock3, qblock1, qblock2, qblock3, splice_motifs, target_seq, query_strand): #{
        tgap1 = [tblock1[1]+1, tblock2[0]-1]
        tgap2 = [tblock2[1]+1, tblock3[0]-1]

        tgap1_size = tgap1[1] - tgap1[0] + 1
        tgap2_size = tgap2[1] - tgap2[0] + 1

        tblock2_size = tblock2[1] - tblock2[0] + 1

        max_shuffle_size = 10
        if tblock2_size > max_shuffle_size: #{
            return False

        small_cutoff = 10
        big_cutoff = 20

        big = None

        if tgap1_size < small_cutoff and tgap2_size > big_cutoff: #{
            big = 'right'

        elif tgap2_size < small_cutoff and tgap1_size > big_cutoff: #{
            big = 'left'

        if not big: #{
            return False

        if big == 'right': #{
            tblock3[0] -= tblock2_size
            qblock3[0] = qblock2[0]
        else:
            tblock1[1] += tblock2_size
            qblock1[1] = qblock2[1]

        tgap = [tblock1[1]+1, tblock3[0]-1]
        gap_seq = target_seq[tgap[0]-1:tgap[1]]
        ss = gap_seq[:2] + gap_seq[-2:]

        merged_gaps = None
        splice_site = None

        if splice_motifs.has_key(ss.lower()) or splice_motifs.has_key(ReverseComplement(ss).lower()): #{
            splice_site = ss
        else:
            #print "testing11", tblock1, tblock3, qblock1, qblock3
            splice_site = self.fix_single_gap(tblock1, tblock3, qblock1, qblock3, splice_motifs, target_seq, query_strand)
            #print "testing22", tblock1, tblock3, qblock1, qblock3

        return splice_site

    def NumExonsOverlapped(self): #{
      return len(getattr(self, "overlap", ()))
    #} end def

    def HasExonOverlap(self): #{
      return (0 < self.NumExonsOverlapped())
    #} end def
#} end class


def calc_identity(
        qstart, qend, tstart, tend, qnuminsert, mismatch, num_bases_aligned):
    """Calculate percent identity as defined by UCSC webblat.

    Assumes start coordinates are 0-based inclusive, and end coordinates are
    0-based exclusive. i.e. adding 1 to start coordinates will convert to
    using 1-based fully inclusive coordinates.

    """
    qspan = abs(int(qend) - int(qstart))
    tspan = abs(int(tend) - int(tstart))
    maxspan = min(qspan, tspan)
    # Why would the maximum span ever be zero or less?
    if maxspan <= 0:
        return 0
    # Assume query span is larger than target span. Otherwise, pretend that
    # the spans are the same size.
    sizediff = qspan - tspan
    if sizediff < 0:
        sizediff = 0
    sizediff_factor = round(3 * math.log(1 + sizediff))
    raw_numerator = (
        int(mismatch) + int(qnuminsert) + sizediff_factor)
    millibad = (1000 * raw_numerator) / num_bases_aligned
    return round(100.0 - millibad * 0.1, 1)


def calc_score(match, mismatch, qnuminsert, tnuminsert): #{
  "ucsc score"
  return int(match) - int(mismatch) - int(qnuminsert) - int(tnuminsert)
#} end def


def clean_filters(filters):
    """Ensure that all filter values are the correct types."""
    # 'int' filters:
    for fkey in ['bestn','count','qlen']:
        if fkey in filters:
            filters[fkey] = int(filters[fkey])
    # 'float' filters:
    for fkey in ['match','identity']:
        if fkey in filters:
            filters[fkey] = float(filters[fkey])
    return filters


#### EXCEPTION CLASSES ####
class ContigAlignmentParserError(MyError): #{
  """Exception raised for errors encountered while parsing contig-to-genome
  alignments.

  """
  pass
#} end class

class DiscardAlignmentError(ContigAlignmentParserError):
    """Exception raised when an alignment should be completely discarded for
    any reason.

    For example, when fixing an alignment's CIGAR and MD strings reveals that
    the alignment is so bad that it never should have been reported in the
    first place.

    """
    pass
