#! /usr/bin/env python

# import standard modules
from optparse import OptionParser, OptionGroup
import os, re, sys, time, traceback

# import custom modules
from version import VERSION
from utils.log import GetLogPath, CloseLogFile
from utils.error import MyError
from utils.general import (SetupMainClass, TimeSpent, WriteCommand,
  NormalizeGeneID)
from utils.messages import ErrMsg, LogMsg, DebugMsg, ExtremeDebugMsg
from utils.files_paths import (CheckFilePath, CheckDirPath, EnsureAbsPath,
  GetOutDir, FileBoxCls)
from parsers.candidate_group_parser import CandidateGroupParserCls
from parsers.fasta import FastaFileCls
from prediction.realigner import RealignerCls
from common.candidate_group import CandidateGroupCls

# CONSTANTS
ES_SUCCESS = 0
ES_RUN_ERR = 1
ES_OPT_ERR = 2
ES_PATH_ERR = 3
ES_EXCEPTION = 4
MSG_SUCCESS = "SUCCESS"
MSG_FAIL = "FAIL"

class MainClassCls: #{
  def __init__(self, options): #{
    SetupMainClass(self, options)
  #} end def

  def __del__(self): #{
    CloseLogFile(self)
  #} end def

  def Run(self): #{
    LogMsg(self, "Post-processing PTD predictions with BLAT...")
    start = time.time()
    realigner = RealignerCls(
        self.options, self.log_info, robust_mode=self.options.robust_mode)
    realigner.options.run_fine_align = False
    #realigner.min_score = 0
    #realigner.min_identity = 0
    self.LoadPredictions(realigner)
    if (0 < len(realigner.contigs)): #{
      # realigner automatically uses "tran_seq_path" as target for alignment
      realigner.options.tran_seq_path = self.options.genome_path
      realigner.RealignContigs()
      self.CheckRealignments(realigner.contigs)
    #} end if
    LogMsg(self, "Time spent post-processing PTD predictions with BLAT: %s" %
      TimeSpent(start))
  #} end def

  def LoadPredictions(self, realigner): #{
    suspect_file = FileBoxCls(self.options.suspect_path, "w")
    for group in CandidateGroupParserCls(self.options.barnacle_path): #{
      suspect_event = CandidateGroupCls(group.DataString())
      suspect_event.suspect_reasons.add("UNMATCHED_EXONS")
      good_members = list()
      for (index, member) in enumerate(group.members): #{
        if (self.CheckMemberExons(member)): #{
          good_members.append(index)
        else:
          suspect_event.members.append(member)
        #} end if
      #} end for
      if (0 < len(suspect_event.members)): #{
        suspect_file.Write(suspect_event.FullDataString())
      #} end if
      if (0 < len(good_members)): #{
        realigner.UpdateContigs(group, good_members, "ptd")
      #} end if
    #} end for
  #} end def

  def CheckMemberExons(self, member): #{
    ExtremeDebugMsg(self, "Checking exon genes for member with contig %s\n"
      "  GENES: %s" % (member.contig_info.id, member.MainGenesString()))
    exon_genes_pair = (set(), set())
    for (exon_genes, all_genes) in zip(exon_genes_pair, member.main_genes): #{
      for gene in all_genes: #{
        parts_list = all_genes[gene]
        if ("exon" in parts_list): #{
          exon_genes.add(NormalizeGeneID(gene).lower())
      #} end for
    #} end for
    ExtremeDebugMsg(self, "  Exon genes: %s" % ";".join(("%s" %
      ",".join(exon_genes) for exon_genes in exon_genes_pair)))
    if (exon_genes_pair[0].isdisjoint(exon_genes_pair[1])): #{
      ExtremeDebugMsg(self, "  No matching exon genes!")
      return False
    #} end if
    return True
  #} end def

  def CheckRealignments(self, realigned_contigs): #{
    pass_file = FileBoxCls(self.options.pass_path, "w")
    fail_file = FileBoxCls(self.options.fail_path, "w")
    for group in CandidateGroupParserCls(self.options.barnacle_path): #{
      DebugMsg(self, "Reprocessing group %i:" % group.id)
      pass_event = CandidateGroupCls(group.DataString())
      fail_event = CandidateGroupCls(group.DataString())
      fail_event.fail_reasons.add("FULL_LENGTH_ALIGN")
      for member in group.members: #{
        if (member.contig_info.id not in realigned_contigs): #{
          continue
        #} end if
        DebugMsg(self, "%s %s %s(%ibp):%.2f %s(%ibp):%.2f %s" % (
          member.IDString(), member.contig_info.ToString(with_kform=False),
          member.align_info_A.ContigCoordsString(),
          member.align_info_A.ContigSpan(), member.align_info_A.identity,
          member.align_info_B.ContigCoordsString(),
          member.align_info_B.ContigSpan(), member.align_info_B.identity,
          member.MainGenesString()))
        realigned_contig = realigned_contigs[member.contig_info.id]
        DebugMsg(self, "%s" % realigned_contig.best_align)
        if (self.CheckMemberRealignment(member, realigned_contig)): #{
          pass_event.members.append(member)
        else:
          fail_event.members.append(member)
        #} end if
      #} end for
      if (0 < len(pass_event.members)): #{
        DebugMsg(self, "Group %i has good members after realignment" %
            group.id)
        pass_file.Write(pass_event.FullDataString())
      #} end if
      if (0 < len(fail_event.members)): #{
        DebugMsg(self, "Group %i has bad members after realignment" %
            group.id)
        fail_file.Write(fail_event.FullDataString())
      #} end if
    #} end for
  #} end def

  def CheckMemberRealignment(self, member, realigned_contig): #{
    contig_alignment = realigned_contig.best_align
    if (None == contig_alignment): #{
      DebugMsg(self, "  Contig cannot be realigned!")
      return True
    #} end if
    if (contig_alignment.perfect): #{
      DebugMsg(self, "  False positive: perfect alignment (%i)" %
        contig_alignment.match)
      return False
    #} end if
    if (5 > contig_alignment.ctg_start and
        5 > (contig_alignment.ctg_length - contig_alignment.ctg_end)): #{
      DebugMsg(self, "  False positive: full-length alignment (%i-%i, "
        "contig length: %i)" % (contig_alignment.ctg_start,
          contig_alignment.ctg_end, contig_alignment.ctg_length))
      return False
    #} end if
    return True
  #} end def
#} end class

#### EXCEPTION CLASSES ####
class MainClassError(MyError): #{
  pass
#} end class

def SetupOptionsParser(): #{
  description_string = ("") #TODO
  args = [ "LIB", "BARNACLE_FILE", "CONTIG_SEQS_FILE", "GENOME_SEQS_FILE" ]
  usage_string = "%prog " + " ".join(args) + " [ OPTIONS ]"
  parser = OptionParser(description=description_string,
                        version="%prog " + VERSION,
                        usage=usage_string)
  parser.num_args = len(args)
  misc_group = OptionGroup(parser, "Miscellaneous Options")
  misc_grp.add_option(
      '--robust-mode', action='store_true', help='If a problem is encountered '
      'while parsing the contig-to-genome realignments, ignore that alignment '
      'and keep going.')
  misc_group.add_option("--disable-profiling-timer",
                    action="store_true", dest="dpt",
                    help="Sometimes this script can hang when trying to spawn "
                         "child processes, due to the kernel's profiling "
                         "timer. Use this option to disable the profiling "
                         "timer if the script seems to be hanging.")
  misc_group.add_option("-f", "--force",
                    action="store_true",
                    help="Force filtering to take place, even if the output "
                         "directory already exists.")
  misc_group.add_option("-d", "--debug",
                    action="store_true",
                    help="Print debug information while the program runs.")
  misc_group.add_option("--extreme-debug",
                    action="store_true", dest="extreme_debug",
                    help="Print extremely in-depth debug information while "
                      "the program runs. Not recommended for large jobs.")
  parser.add_option_group(misc_group)
  parser.set_defaults(robust_mode=False,
                      dpt=False,
                      force=False,
                      debug=False,
                      extreme_debug=False)
  return parser
#} end def

def CheckPaths(options): #{
  opts_good = True
  if (not options.barnacle_path.endswith(".ptd")): #{
    ErrMsg("This post-processor is only for PTD predictions!")
    opts_good = False
  #} end if
  path_errors = list()
  CheckFilePath(options.barnacle_path, "Barnacle predictions", path_errors)
  CheckFilePath(options.ctg_seq_path, "contig sequences", path_errors)
  CheckFilePath(options.genome_path, "genome sequences", path_errors)
  # get and check the output path
  #options.output_dir = GetOutDir(os.path.dirname(options.barnacle_path),
  #  "TASK_DESC")
  (options.output_dir, infilename) = os.path.split(options.barnacle_path)
  options.suspect_path = "%s.suspect" % options.barnacle_path
  options.pass_path = "%s.pass" % options.barnacle_path
  options.fail_path = "%s.fail" % options.barnacle_path
  options.realign_dir = os.path.join(options.output_dir, "ptd_realign")
  #options.query_path = os.path.join(options.realign_dir,
  #  "%s.ptd_contigs.fa" % options.lib)
  #options.align_path = os.path.join(options.realign_dir,
  #  "%s.ptd_realign.psl" % options.lib)
  if (opts_good and 0 == len(path_errors)): #{
    CheckDirPath(options.output_dir, "output", path_errors,
      create=False, replace=True)
    CheckDirPath(options.realign_dir, "realignment", path_errors,
      create=True, replace=True)
    # get the log file name
    options.log_file_name = GetLogPath(options.barnacle_path,
      "ptd_post_process", options.output_dir)
  #} end if
  if (0 < len(path_errors)): #{
    ErrMsg("Errors in input arguments:\n  %s" % "\n  ".join(path_errors))
  #} end if
  # the paths are good if there are no path errors and no conflicting options
  return (opts_good and 0 == len(path_errors))
#} end def

def Main(): #{
  # get options and arguments
  parser = SetupOptionsParser()
  (options, args) = parser.parse_args()
  # if the right number of args was used
  if (parser.num_args == len(args)): #{
    options.lib           = args[0]
    options.barnacle_path = EnsureAbsPath(args[1])
    options.ctg_seq_path  = EnsureAbsPath(args[2])
    options.genome_path   = EnsureAbsPath(args[3])
    if (CheckPaths(options)): #{
      try: #{
        main_class_object = MainClassCls(options)
        WriteCommand(main_class_object, sys.argv)
        main_class_object.Run()
      except (MyError), e:
        ErrMsg("ERROR while post-processing PTD predictions with BLAT:\n  %s" % e)
        return ES_RUN_ERR
      #} end try
    else:
      return ES_PATH_ERR
    #} end if
  else:
    parser.error("you must specify a library name (LIB); the path to a "
      "Barnacle data file for that library (BARNACLE_FILE); the path to a "
      "file holding the sequences of the contigs in the Barnacle predictions "
      "(CONTIG_SEQS_FILE); and the path to a file holding genomic sequences "
      "(GENOME_SEQS_FILE).")
    return ES_OPT_ERR
  #} end if
  return ES_SUCCESS
#} end def

if __name__ == '__main__': #{
  try: #{
    exit_status = Main()
  except Exception, e:
    traceback.print_exc()
    exit_status = ES_EXCEPTION
  #} end try
  if (ES_SUCCESS == exit_status): #{
    print MSG_SUCCESS
  else:
    print MSG_FAIL
  #} end if
  sys.exit(exit_status)
#} end if
