#! /usr/bin/env python
"""
quick_summary.py

Created by Lucas Swanson
Copyright (c) 2012 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

# import standard modules
from optparse import OptionParser, OptionGroup
import itertools, os, re, sys, time, traceback

# import custom modules
from version import VERSION
from utils.log import GetLogPath, CloseLogFile
from utils.error import MyError
from utils.general import (SetupMainClass, TimeSpent, WriteCommand, PowerSet,
    NormalizeGeneID, StrJoin)
from utils.messages import ErrMsg, LogMsg, DebugMsg, ExtremeDebugMsg
from utils.files_paths import (CheckFilePath, CheckDirPath, EnsureAbsPath,
  FileBoxCls)
from utils.multi_dict import AddToMultiDict
from utils.library_iterator import LibIteratorCls
from parsers.candidate_group_parser import CandidateGroupParserCls

# CONSTANTS
ES_SUCCESS = 0
ES_RUN_ERR = 1
ES_OPT_ERR = 2
ES_PATH_ERR = 3
ES_EXCEPTION = 4
MSG_SUCCESS = "SUCCESS"
MSG_FAIL = "FAIL"

EVENT_TYPES_EXT = [
  ("standard fusion", "fus"),                       # 0
  ("mixed-sense fusion", "fus.mix_dirs"),           # 1
  ("total fusion", "fus.all"),                      # 2
  ("ptd", "ptd"),                                   # 3
  ("itd-2f", "itd"),                                # 4
  ("itd-2f multi", "itd.multi_exon"),               # 5
  ("itd-1f", "itd.edge_gap"),                       # 6
  ("itd-1f multi", "itd.edge_gap.multi_exon"),      # 7
  ("itd-0f", "itd.full_ctg_dup"),                   # 8
  ("itd-0f multi", "itd.full_ctg_dup.multi_exon"),  # 9
  ("itd-flanked", "itd.flanked"),                   # 10
  ("itd-all", "itd.all"),                           # 11
]
EVENT_TYPES = [type[0] for type in EVENT_TYPES_EXT]
BASE_TYPES = [EVENT_TYPES[i] for i in [0,1,3,4,5,6,7,8,9]]
ITD_ALL_INDEX = 11

class PredictionSummarizerCls: #{
  def __init__(self, options): #{
    SetupMainClass(self, options)
    #ExtremeDebugMsg(self, "In main class constructor")
    #self.event_counters = dict(
    #  [(event_type[0], EventCounterCls(event_type[0], event_type[1])) for
    #    event_type in EVENT_TYPES])
    #self.event_counters = [EventCounterCls(event_type, self.log_info) for
    #  event_type in EVENT_TYPES]
    if (self.options.itd_exons): #{
      events_to_check = [EVENT_TYPES_EXT[ITD_ALL_INDEX]]
    else:
      events_to_check = EVENT_TYPES_EXT
    #} end if
    self.event_counters = dict(([event_type[0],
      EventCounterCls(event_type, self.log_info, self.options.itd_exons)] for
      event_type in events_to_check))
    #self.event_counters = dict(([event_type[0],
    #  EventCounterCls(event_type, self.log_info)] for
    #  event_type in EVENT_TYPES_EXT))
    #ExtremeDebugMsg(self, " Created event counters dict")
    #self.output_file = None
    # genes[gene_id][event_type][lib_id] = list(r2c_values)
    self.genes = dict()
    self.gene_names = dict()
    self.libs = list()
  #} end def

  def __del__(self): #{
    #if (None != self.output_file): #{
    #  self.output_file.Close()
    #} end if
    CloseLogFile(self)
  #} end def

  def SummarizePredictions(self): #{
    LogMsg(self, "Summarizing Barnacle predictions...")
    start = time.time()
    # initialize the library iterator for loading predictions
    lib_iterator = LibIteratorCls(self.options.input_path,
       self.LoadLibrary, self.options, self.log_info)
    # do not try to find event paths
    lib_iterator.options.get_paths = False
    # load the predictions from each library
    lib_iterator.IterateOverAllLibs()
    # create output file (OR JUST USE LOG FILE??)
    #self.CreateOutputFile()
    #for event_counter in self.event_counters: #{}
    if (self.options.itd_exons): #{
      events_to_check = [EVENT_TYPES[ITD_ALL_INDEX]]
    else:
      events_to_check = EVENT_TYPES
    #} end if
    #for etype in EVENT_TYPES: #{}
    for etype in events_to_check: #{
      event_counter = self.event_counters[etype]
      #DebugMsg(self, "%s:\n%s" % (event_counter.key,
      #  "\n".join(["%s: %s -- %s" % (name, event_counter.gene_keys[name],
      #  ",".join(sorted(event_counter.genes[event_counter.gene_keys[name]])))
      #  for name in sorted(event_counter.gene_keys)])))
      # amalgamate gene keys
      event_counter.AmalgamateGeneKeys()
      #DebugMsg(self, "%s:\n%s" % (event_counter.key,
      #  "\n".join(["%s: %s -- %s" % (name, event_counter.gene_keys[name],
      #  ",".join(sorted(event_counter.genes[event_counter.gene_keys[name]])))
      #  for name in sorted(event_counter.gene_keys)])))
      # summarize event type
      #raise PredictionSummarizerError("not implemented!")
      if (self.options.gene_list): #{
        self.LoadGeneRecurrenceDict(event_counter)
      #} end if
      if (0 < event_counter.num_predictions): #{
        self.SummarizeEventType(event_counter)
      #} end if
    #} end for
    if (self.options.gene_list): #{
      self.WriteGeneRecurrence()
    #} end if
    if (not self.options.itd_exons): #{
      self.WriteLibCounts()
    #} end if
    if (self.options.write_events): #{
      # initialize the library iterator for writing recurrent predictions
      lib_iterator = LibIteratorCls(self.options.input_path,
         self.WriteEvents, self.options, self.log_info)
      # do not try to find event paths
      lib_iterator.options.get_paths = False
      # write the recurrent predictions for each library
      lib_iterator.IterateOverAllLibs()
      DebugMsg(self, "Num libs processed: %i" % lib_iterator.num_libs)
    #} end if
    LogMsg(self, "Total time spent summarizing Barnacle predictions: %s" %
      TimeSpent(start))
  #} end def

  def LoadLibrary(self, lib_info): #{
    start = time.time()
    LogMsg(self, "Loading predictions for library %s, output version: %s..." %
      (lib_info.lib_name, lib_info.barnacle_ver))
    self.libs.append(lib_info.lib_name)
    #for event_counter in self.event_counters: #{}
    if (self.options.itd_exons): #{
      events_to_check = [EVENT_TYPES[ITD_ALL_INDEX]]
    else:
      events_to_check = EVENT_TYPES
    #} end if
    #for etype in EVENT_TYPES: #{}
    for etype in events_to_check: #{
      event_counter = self.event_counters[etype]
      DebugMsg(self, "Loading %s predictions..." % etype)
      event_counter.AddLibrary(lib_info.lib_name)
      event_file_name = "%s.barnacle.%s" % (lib_info.lib_name,
        event_counter.ext)
      event_path = os.path.join(lib_info.lib_dir, "8_predicted_events",
        event_file_name)
      DebugMsg(self, "PATH: %s" % event_path)
      if (not os.path.isfile(event_path)): #{
        DebugMsg(self, "No %s predictions for library." % etype)
        continue
      #} end if
      event_parser = CandidateGroupParserCls(event_path)
      for event in event_parser: #{
        DebugMsg(self, "Loading event %i" % event.id)
        event_counter.LoadEvent(lib_info.lib_name, event)
      #} end for
    #} end for
    LogMsg(self, "Time loading predictions: %s" % TimeSpent(start))
  #} end def

  #def CreateOutputFile(self): #{
  #  input_file_name = os.path.basename(self.options.input_path)
  #  input_root = os.path.splitext(input_file_name)[0]
  #  output_file_name = "%s.barnacle_summary.txt" % input_root
  #  output_path = os.path.join(self.options.output_dir, output_file_name)
  #  self.output_file = FileBoxCls(output_path, "w",
  #    "cannot create output path")
  #} end def

  def LoadGeneRecurrenceDict(self, event_counter): #{
    start = time.time()
    LogMsg(self, "Loading %s gene recurrence..." % event_counter.key)
    # self.genes[gene_id][event_type][lib_id] = list(r2c_values)
    # event_counter.genes[gene_key][lib_id] = r2c_list
    for (gene_key, libs) in event_counter.genes.iteritems(): #{
      for (lib_key, r2c_list) in libs.iteritems(): #{
        keys = [gene_key, event_counter.key, lib_key]
        AddToMultiDict(self.genes, keys, r2c_list, doextend=True)
        # for fusions, add both sides separately as well
        if (";" in gene_key): #{
          for sub_key in gene_key.split(";", 1): #{
            keys = [sub_key, event_counter.key, lib_key]
            AddToMultiDict(self.genes, keys, r2c_list, doextend=True)
          #} end for
        #} end if
      #} end for
    #} end for
    for (gene_key, gene_names_str) in event_counter.gene_names.iteritems(): #{
      if (gene_key in self.gene_names and
          gene_names_str != self.gene_names[gene_key]): #{
        raise PredictionSummarizerError("inconsistent gene names for %s: "
          "%s and %s" % (gene_key, gene_names_str, self.gene_names[gene_key]))
      #} end if
      self.gene_names[gene_key] = gene_names_str
    #} end for
    LogMsg(self, "Time loading gene recurrence: %s" % TimeSpent(start))
  #} end def

  def SummarizeEventType(self, event_counter): #{
    start = time.time()
    LogMsg(self, "\nSummarizing %s predictions..." % event_counter.key)
    total_gene_set = set()
    total_gene_set_rec = set()
    for lib_key in sorted(event_counter.libraries.keys()): #{
      lib_counts = event_counter.counts[lib_key]
      lib_counts["preds"]  = len(event_counter.libraries[lib_key])
      lib_counts["rpreds"] = 0
      gene_set = set()
      recurrent_gene_set = set()
      for gene_names_str in event_counter.libraries[lib_key]: #{
        gene_key = event_counter.gene_keys[gene_names_str]
        gene_set.add(gene_key)
        total_gene_set.add(gene_key)
        if (1 < len(event_counter.genes[gene_key])): #{
          lib_counts["rpreds"] += 1
          recurrent_gene_set.add(gene_key)
          total_gene_set_rec.add(gene_key)
        #} end if
      #} end for
      lib_counts["upreds"] = (lib_counts["preds"] - lib_counts["rpreds"])
      lib_counts["genes"]  = len(gene_set)
      lib_counts["rgenes"] = len(recurrent_gene_set)
      lib_counts["ugenes"] = (lib_counts["genes"] - lib_counts["rgenes"])
      LogMsg(self, "%s: %i predictions (%i recurrent); %i genes "
        "(%i recurrent)" % (lib_key, lib_counts["preds"],
        lib_counts["rpreds"], lib_counts["genes"],
        lib_counts["rgenes"]))
    #} end for
    LogMsg(self, "Total genes: %i (%i recurrent)" % (len(total_gene_set),
      len(total_gene_set_rec)))
    LogMsg(self, "Time summarizing predictions: %s" % TimeSpent(start))
  #} end def

  def WriteGeneRecurrence(self): #{
    start = time.time()
    LogMsg(self, "\nWriting gene recurrence...")
    recur_file = FileBoxCls(self.options.recurrent_genes_path, "w",
        "cannot create gene recurrence output file")
    # self.genes[gene_id][event_type][lib_id] = list(r2c_values)
    by_gid = sorted(self.genes)
    #for gid in sorted(self.genes, key=lambda gid:
    #for gid in sorted(by_gid, key=lambda gid:
    #    #(sum([len(self.genes[gid][etype]) for etype in self.genes[gid]]), gid),
    #    sum([len(self.genes[gid][etype]) for etype in self.genes[gid]]),
    #    reverse=True): #{}
    for gid in sorted(by_gid, key=self.SumPredTypesOverAllLibs,
        reverse=True): #{
      #total_num_preds = sum([len(self.genes[gid][etype]) for
      #  etype in self.genes[gid]])
      total_num_preds = self.SumPredTypesOverAllLibs(gid)
      gene_name = self.gene_names[gid]
      out_line = "%s (%i)" % (gene_name, total_num_preds)
      ExtremeDebugMsg(self, out_line)
      recur_file.WriteLine(out_line)
      if (self.options.itd_exons): #{
        events_to_write = [EVENT_TYPES[ITD_ALL_INDEX]]
      else:
        events_to_write = BASE_TYPES
      #} end if
      #for etype in EVENT_TYPES: #{}
      for etype in events_to_write: #{
        if (etype in self.genes[gid]): #{
          libs = self.genes[gid][etype]
          num_preds = len(libs)
          libs_str = "; ".join(["%s (%s)" % (lib_id,
            StrJoin(", ", sorted(set(r2c_list)))) for (lib_id, r2c_list) in
            sorted(libs.iteritems())])
          out_line = "  %s (%i): %s" % (etype, num_preds, libs_str)
          ExtremeDebugMsg(self, out_line)
          recur_file.WriteLine(out_line)
        #} end if
      #} end for
      recur_file.WriteLine("")
    #} end for
    recur_file.Close()
    LogMsg(self, "Time writing gene recurrence: %s\nSee file:\n  %s\n" %
        (TimeSpent(start), self.options.recurrent_genes_path))
  #} end def

  def SumPredTypesOverAllLibs(self, gid): #{
    if (self.options.itd_exons): #{
      return len(self.genes[gid][EVENT_TYPES[ITD_ALL_INDEX]])
    #} end if
    return sum([len(self.genes[gid][etype]) for etype in
      set(self.genes[gid]).intersection(BASE_TYPES)])
  #} end def

  def WriteLibCounts(self): #{
    start = time.time()
    LogMsg(self, "Writing library counts table...")
    lib_counts_file = FileBoxCls(self.options.lib_counts_path, "w",
        "cannot create library counts table file")
    #header = "Library\t%s" % ("\t".join((ec.key for ec in
    #  self.event_counters)))
    header = "Library\t%s" % ("\t".join((etype.replace(" ", "_") for
      etype in EVENT_TYPES)))
    if (None != self.options.targets and 0 < len(self.options.targets)): #{
      #header += "\t%s" % "\t".join((t.split(":",1)[1] for t in
      header += "\t%s" % "\t".join(("%s(%s)" % (t[1], t[0].replace(" ","_"))
        for t in self.options.targets))
    #} end if
    lib_counts_file.WriteLine(header)
    for lib in sorted(self.libs): #{
      ExtremeDebugMsg(self, "Writing count for %s..." % lib)
      lib_line = "%s\t%s" % (lib, "\t".join(("%i(%i)" %
        (self.event_counters[etype].counts[lib]["preds"],
        self.event_counters[etype].counts[lib]["upreds"]) for
        etype in EVENT_TYPES)))
      if (None != self.options.targets and 0 < len(self.options.targets)): #{
        #for target in self.options.targets: #{}
        #  (etype, tgene) = target.split(":",1)
        for (etype, tgene) in self.options.targets: #{
          etype = etype.replace("_", " ")
          ExtremeDebugMsg(self, "  Looking for target: %s (of type %s)" %
              (tgene, etype))
          tgene = tgene.lower().replace("/",";")
          target_found = "N"
          # libraries[lib_name] = list of gene_names strings
          for gene_names in self.event_counters[etype].libraries[lib]: #{
            gene_keys = self.event_counters[etype].gene_keys[gene_names]
            ExtremeDebugMsg(self, "    Names: %s, Keys: %s" %
                (gene_names, gene_keys))
            if (";" in tgene): #{
              tsets = [set(tset.split(",")) for tset in tgene.split(";")]
              psets = [set(pset.split(",")) for pset in gene_keys.split(";")]
              #ExtremeDebugMsg(self, "TSET: %s\nPSET: %s\nISECT: %s" %
              #    (", ".join(tset), ", ".join(pset),
              #    ", ".join(tset.intersection(pset))))
              if ((0 < len(tsets[0].intersection(psets[0])) and
                   0 < len(tsets[1].intersection(psets[1]))) or
                  (0 < len(tsets[0].intersection(psets[1])) and
                    0 < len(tsets[1].intersection(psets[0])))): #{
                target_found = "Y"
                break
              #} end if
            else:
              tset = set(tgene.split(","))
              # will create a nested list of lists if gene_keys contains a ";"
              pset_nested = (sub_keys.split(",") for sub_keys in
                  gene_keys.split(";"))
              # flatten the list of lists and convert to a set
              pset = set(itertools.chain.from_iterable(pset_nested))
              ExtremeDebugMsg(self, "TSET: %s\nPSET: %s\nISECT: %s" %
                  (", ".join(tset), ", ".join(pset),
                  ", ".join(tset.intersection(pset))))
              if (0 < len(tset.intersection(pset))): #{
                target_found = "Y"
                break
              #} end if
            #} end if
            #if (tgene.lower() in gene_keys): #{
            #  target_found = "Y"
            #  break
            #} end if
          #} end for
          ExtremeDebugMsg(self, "    Found: %s" % target_found)
          lib_line += "\t%s" % target_found
        #} end for
      #} end if
      lib_counts_file.WriteLine(lib_line)
    #} end for
    LogMsg(self, "Time writing library counts table: %s\nSee file:\n  %s\n" %
        (TimeSpent(start), self.options.lib_counts_path))
  #} end def

  def WriteEvents(self, lib_info): #{
    start = time.time()
    LogMsg(self, "Writing recurrent and library-specific predictions for "
      "library %s, output version: %s..." % (lib_info.lib_name,
      lib_info.barnacle_ver))
    #for event_counter in self.event_counters: #{}
    if (self.options.itd_exons): #{
      events_to_check = [EVENT_TYPES[ITD_ALL_INDEX]]
    else:
      events_to_check = EVENT_TYPES
    #} end if
    #for etype in EVENT_TYPES: #{}
    for etype in events_to_check: #{
      event_counter = self.event_counters[etype]
      DebugMsg(self, "Writing %s predictions..." % etype)
      event_file_name = "%s.barnacle.%s" % (lib_info.lib_name,
        event_counter.ext)
      event_path = os.path.join(lib_info.lib_dir, "8_predicted_events",
        event_file_name)
      #DebugMsg(self, "PATH: %s" % event_path)
      if (not os.path.isfile(event_path)): #{
        DebugMsg(self, "No %s predictions for library." % etype)
        continue
      #} end if
      event_parser = CandidateGroupParserCls(event_path)
      recurrent_path = "%s.rec" % event_path
      recurrent_file = FileBoxCls(recurrent_path, "w",
        "cannot create recurrent prediction file")
      lib_specific_path = "%s.lsp" % event_path
      lib_specific_file = FileBoxCls(lib_specific_path, "w",
        "cannot create library-specific prediction file")
      for event in event_parser: #{
        # get the gene_names_str for the event
        gene_names_str = event_counter.GetGeneNamesString(event)
        # get the appropriate gene_key
        gene_key = event_counter.gene_keys[gene_names_str]
        # check the number of libraries with predictions
        num_libs = len(event_counter.genes[gene_key])
        DebugMsg(self, "G%i: %s = %s (%s) = %i" % (event.id,
          gene_names_str, gene_key, ",".join(event_counter.genes[gene_key]),
          num_libs))
        # write the event if it is recurrent
        if (1 < num_libs): #{
          recurrent_file.Write(event.FullDataString())
        else:
          lib_specific_file.Write(event.FullDataString())
        #} end if
      #} end for
    #} end for
    LogMsg(self, "Time loading predictions: %s" % TimeSpent(start))
  #} end def
#} end class

class EventCounterCls: #{
  def __init__(self, event_info, log_info, itd_exons=False): #{
    self.key = event_info[0]
    self.ext = event_info[1]
    # libraries[lib_name] = list of gene_names strings
    self.libraries = dict()
    # gene_keys[gene_names string] = gene_key
    self.gene_keys = dict()
    # gene_names[gene_key] = gene_names string
    self.gene_names = dict()
    # genes[gene_key] = lib_set
    # genes[gene_key][lib_id] = r2c_list
    self.genes = dict()
    self.log_info = log_info
    self.itd_exons = itd_exons
    self.num_predictions = 0
    self.counts = dict()
  #} end def

  def AddLibrary(self, lib_key): #{
    self.libraries[lib_key] = list()
    self.counts[lib_key] = dict(([ctype, 0] for ctype in ["preds", "rpreds",
      "upreds", "genes", "rgenes", "ugenes"]))
  #} end def

  def LoadEvent(self, lib_key, event): #{
    gene_names_str = self.GetGeneNamesString(event)
    self.libraries[lib_key].append(gene_names_str)
    gene_key = gene_names_str.lower()
    DebugMsg(self, "Genes: %s (key: %s)" % (gene_names_str, gene_key))
    if (gene_names_str in self.gene_keys and
        gene_key != self.gene_keys[gene_names_str]): #{
      raise PredictionSummarizerError("inconsistent keys for %s: %s and %s" %
        (gene_names_str, gene_key, self.gene_keys[gene_names_str]))
    #} end if
    self.gene_keys[gene_names_str] = gene_key
    self.AddGeneNamesString(gene_key, gene_names_str)
    if (";" in gene_key): #{
      for (sub_key, sub_name) in zip(gene_key.split(";",1),
          gene_names_str.split(";",1)): #{
        self.AddGeneNamesString(sub_key, sub_name)
      #} end for
    #} end if
    #if (gene_key not in self.genes): #{
    #  self.genes[gene_key] = set()
    #} end if
    #self.genes[gene_key].add(lib_key)
    # genes[gene_key][lib_id] = r2c_list
    for candidate in event.members: #{
      r2c = candidate.read_to_ctg_unique
      AddToMultiDict(self.genes, [gene_key, lib_key], r2c)
    #} end for
    self.num_predictions += 1
  #} end def

  def AddGeneNamesString(self, gene_key, gene_names_str): #{
    if (gene_key in self.gene_names and
        gene_names_str != self.gene_names[gene_key]): #{
      raise PredictionSummarizerError("inconsistent gene names for %s: "
        "%s and %s" % (gene_key, gene_names_str, self.gene_names[gene_key]))
    #} end if
    self.gene_names[gene_key] = gene_names_str
  #} end def

  def GetGeneNamesString(self, event): #{
    if ("fusion" in self.key): #{
      gene_sets = self.GetGeneSets(event)
      gene_sets[0].discard("no_exons")
      gene_sets[1].discard("no_exons")
      gene_names_str_1 = ",".join(sorted(gene_sets[0]))
      gene_names_str_2 = ",".join(sorted(gene_sets[1]))
      return ";".join(sorted([gene_names_str_1, gene_names_str_2]))
    else:
      gene_set = self.GetGeneSet(event)
      gene_set.discard("no_exons")
      return ",".join(sorted(gene_set))
    #} end if
  #} end def

  def GetGeneSets(self, event): #{
    clean_lists = (map(NormalizeGeneID, getattr(event.members[0], set_id)) for
      set_id in ("genes_A", "genes_B"))
    #gene_set_1 = set(event.members[0].genes_A)
    #gene_set_2 = set(event.members[0].genes_B)
    gene_sets = map(set, clean_lists)
    DebugMsg(self, "Initializing gene sets with: %s" %
      event.members[0].MainGenesString())
    if (1 == len(event.members)): #{
      #return (gene_set_1, gene_set_2)
      return gene_sets
    #} end if
    for member in event.members[1:]: #{
      DebugMsg(self, "Updating gene sets with: %s" %
        member.MainGenesString())
      clean_lists = list(map(NormalizeGeneID, getattr(member, set_id)) for
        set_id in ("genes_A", "genes_B"))
      #overlap = (len(gene_set_1.intersection(member.genes_A)) +
      #  len(gene_set_2.intersection(member.genes_B)))
      overlap = sum((len(gene_set.intersection(clean_list)) for
        (gene_set, clean_list) in zip(gene_sets, clean_lists)))
      #swap_overlap = (len(gene_set_1.intersection(member.genes_B)) +
      #  len(gene_set_2.intersection(member.genes_A)))
      swap_overlap = sum((len(gene_set.intersection(clean_list)) for
        (gene_set, clean_list) in zip(gene_sets, reversed(clean_lists))))
      DebugMsg(self, "Overlap: %i, Swap overlap: %i" % (overlap, swap_overlap))
      if (overlap > swap_overlap): #{
        DebugMsg(self, "Not swapping")
        #gene_set_1.update(member.genes_A)
        #gene_set_2.update(member.genes_B)
      else:
        DebugMsg(self, "Swapping")
        #gene_set_1.update(member.genes_B)
        #gene_set_2.update(member.genes_A)
        clean_lists.reverse()
      #} end if
      for i in range(2): #{
        gene_sets[i].update(clean_lists[i])
      #} end for
    #} end for
    #return (gene_set_1, gene_set_2)
    return gene_sets
  #} end def

  def GetGeneSet(self, event): #{
    gene_set = set()
    for member in event.members: #{
      if (self.itd_exons): #{
        #gene_set.update(self.GetMemberExons(member))
        gene_set.update(self.GetITDMemberExons(member))
      else:
        gene_set.update(map(NormalizeGeneID, member.genes_A))
      #} end if
    #} end for
    return gene_set
  #} end def

  def GetITDMemberExons(self, member): #{
    #ErrMsg("Getting ITD Member Exons:\n  %s" %
    #    member.BPGenesString())
    exon_set = set()
    for gene in member.breakpoint_genes_A: #{
      #ErrMsg("  --Curr gene: %s" % gene)
      exon_dict = dict()
      transcript_dict = member.breakpoint_genes_A[gene]
      for transcript in transcript_dict: #{
        parts_dict = transcript_dict[transcript]
        if ("exon" in parts_dict): #{
          exon_list = parts_dict["exon"]
          for exon in exon_list: #{
            if (exon not in exon_dict): #{
              exon_dict[exon] = set()
            #} end if
            exon_dict[exon].add(transcript)
          #} end for
        #} end if
      #} end for
      #ErrMsg("    Exons: %s" % exon_dict)
      common_exons = None
      transcript_count = 0
      for exon in exon_dict: #{
        curr_transcripts = exon_dict[exon]
        curr_count = len(curr_transcripts)
        if (transcript_count < curr_count): #{
          common_exons = dict()
          transcript_count = curr_count
        #} end if
        if (transcript_count == curr_count): #{
          common_exons[exon] = sorted(curr_transcripts)[0]
        #} end if
      #} end for
      if (None == common_exons): #{
        continue
      #} end if
      #ErrMsg("    Common exons: %s" % common_exons)
      exon_set.update(("_".join((NormalizeGeneID(gene), transcript, exon))
        for (exon, transcript) in common_exons.iteritems()))
    #} end for
    return exon_set
  #} end def

  def GetMemberExons(self, member): #{
    exon_set = set()
    for gene in member.breakpoint_genes_A: #{
      transcript_dict = member.breakpoint_genes_A[gene]
      for transcript in transcript_dict: #{
        parts_dict = transcript_dict[transcript]
        if ("exon" in parts_dict): #{
          exon_list = parts_dict["exon"]
          exon_set.update(("_".join((NormalizeGeneID(gene), transcript, exon))
            for exon in exon_list))
        #} end if
      #} end for
    #} end for
    return exon_set
  #} end def

  def AmalgamateGeneKeys(self): #{
    start = time.time()
    LogMsg(self, "Amalgamating %s gene keys..." % self.key)
    processed = set()
    for gene_names_str in sorted(self.gene_keys, key=lambda x: x.count(","),
        reverse=True): #{
      full_key = self.gene_keys[gene_names_str]
      DebugMsg(self, "%s: %s" % (gene_names_str, full_key))
      if (gene_names_str in processed): #{
        DebugMsg(self, "Skipping processed key")
        continue
      #} end if
      if (0 == gene_names_str.count(",")): #{
        DebugMsg(self, "Done")
        break
      #} end if
      if ("fusion" in self.key): #{
        self.AmalgamateGenePairKeys(gene_names_str, full_key, processed)
      else:
        for subset in PowerSet(gene_names_str.split(",")): #{
          DebugMsg(self, "Subset: %s" % ",".join(sorted(subset)))
          if (0 == len(subset) or
              len(gene_names_str.split(",")) == len(subset)): #{
            continue
          #} end if
          self.CheckSubKey(subset, None, full_key, processed)
        #} end for
      #} end if
    #} end for
    LogMsg(self, "Time amalgamating gene keys: %s" % TimeSpent(start))
  #} end def

  def AmalgamateGenePairKeys(self, gene_names_str, full_key, processed): #{
    (genes_1, genes_2) = sorted(gene_names_str.split(";"),
      key=lambda x: x.count(","), reverse=True)
    for subset_1 in PowerSet(map(NormalizeGeneID, genes_1.split(","))): #{
      #DebugMsg(self, "Subset1: %s" % ",".join(sorted(subset_1)))
      if (0 == len(subset_1)): #{
        continue
      #} end if
      #if ("," not in genes_2): #{
      #  self.CheckSubKey(subset_1, [genes_2], processed)
      #  continue
      #} end if
      for subset_2 in PowerSet(map(NormalizeGeneID, genes_2.split(","))): #{
        #DebugMsg(self, "Subset2: %s" % ",".join(sorted(subset_2)))
        if (0 == len(subset_2) or
            (len(genes_1.split(",")) == len(subset_1) and
             len(genes_2.split(",")) == len(subset_2))): #{
          continue
        #} end if
        self.CheckSubKey(subset_1, subset_2, full_key, processed)
      #} end for
    #} end for
  #} end def

  def CheckSubKey(self, subset_1, subset_2, full_key, processed): #{
    if ("fusion" in self.key): #{
      sub_names_str_1 = ",".join(sorted(subset_1))
      sub_names_str_2 = ",".join(sorted(subset_2))
      sub_names_str = ";".join(sorted([sub_names_str_1, sub_names_str_2]))
    else:
      sub_names_str = ",".join(sorted(subset_1))
    #} end if
    if (sub_names_str in processed): #{
      DebugMsg(self, "WARNING: sub key already processed: %s" % sub_names_str)
      return
    #} end if
    processed.add(sub_names_str)
    sub_key = sub_names_str.lower()
    DebugMsg(self, "SUBKEY %s: %s" % (sub_names_str, sub_key))
    if (sub_names_str not in self.gene_keys): #{
      DebugMsg(self, "Sub Key not present")
      return
    #} end if
    DebugMsg(self, "Amalgamating sub key with %s" % full_key)
    self.gene_keys[sub_names_str] = full_key
    # genes[gene_key][lib_id] = r2c_list
    if (sub_key in self.genes): #{
      libs = self.genes[sub_key]
      for (lib_key, r2c_list) in libs.iteritems(): #{
        AddToMultiDict(self.genes, [full_key, lib_key], r2c_list,
            doextend=True)
      #} end for
      #self.genes[full_key].update(self.genes[sub_key])
      del self.genes[sub_key]
    #} end if
  #} end def
#} end class

#### EXCEPTION CLASSES ####
class PredictionSummarizerError(MyError): #{
  pass
#} end class

def SetupOptionsParser(): #{
  description_string = ("Counts number of predictions, number of recurrent "
    "predictions, number of genes with predictions, and number of genes with "
    "recurrent predictions for each library in the input list.")
  args = [ "LIBRARY_LIST", "LIBS_DIR", ]
  usage_string = "%prog " + " ".join(args) + " [ OPTIONS ]"
  parser = OptionParser(description=description_string,
                        version="%prog " + VERSION,
                        usage=usage_string)
  parser.num_args = len(args)
  parser.add_option("-t", "--target-gene",
      action="append", dest="targets", nargs=2,
      metavar="EVENT_TYPE GENE_NAME",
      help="Use this option to provide the name of a gene of particular "
        "interest, to be explicitly checked for presence in each library. "
        "Use this option multiple times to provide multiple genes of "
        "interest.")
  parser.add_option("--gene-list",
      action="store_true", dest="gene_list",
      help="Write a file containing a list of genes with their recurrence "
        "information. [default]")
  parser.add_option("--no-gene-list",
      action="store_false", dest="gene_list",
      help="Do not create a gene recurrence file.")
  parser.add_option("--write-events",
      action="store_true", dest="write_events",
      help="Write events to recurrent or library-specific "
        "output files.")
  parser.add_option("--count-only",
      action="store_false", dest="write_events",
      help="Do not write events to recurrent or "
        "library-specific output files. [default]")
  parser.add_option("--template",
      metavar="TEMPLATE", dest="lib_path_template",
      help="The template to use to construct the path to the event data "
        "files for a library. \"%{base}\", \"%{lib}\", \"%{assembly_ver}\", "
        "and \"%{barnacle_ver}\" will be replaced with the appropriate "
        "values from LIB_LIST_FILE.")
  parser.add_option("--itd-exons",
      action="store_true",
      help="Only check ITD events, and determine exon-level recurrence.")
  parser.add_option("-d", "--debug",
      action="store_true",
      help="Print debug information while the program runs.")
  parser.add_option("--extreme-debug",
      action="store_true", dest="extreme_debug",
      help="Print extremely in-depth debug information while "
        "the program runs. Not recommended for large jobs.")
  parser.set_defaults(gene_list=True,
                      write_events=False,
                      force=False,
                      debug=False,
                      extreme_debug=False)
  return parser
#} end def

def CheckPaths(options): #{
  opts_good = True
  if (None != options.targets and 0 < len(options.targets)): #{
    invalid_target_types = set()
    for (etype, tgene) in options.targets: #{
      etype = etype.replace("_", " ")
      if (etype not in EVENT_TYPES): #{
        invalid_target_types.add(etype)
      #} end if
    #} end for
    if (0 < len(invalid_target_types)): #{
      ErrMsg("Invalide target event types provided: %s\n"
          "  Valid types are: %s" % (", ".join(invalid_target_types),
          ", ".join(EVENT_TYPES)))
      opts_good = False
    #} end if
  #} end if
  path_errors = list()
  CheckFilePath(options.input_path, "libraries list", path_errors)
  options.output_dir = os.path.dirname(options.input_path)
  # check the libraries path
  CheckDirPath(options.libs_dir, "libraries", path_errors,
    create=False)
  if (options.itd_exons): #{
    task = "itd_exon_summary"
    gene_filename = "itd_exon_recurrence.txt"
  else:
    task = "summary"
    gene_filename = "gene_recurrence.txt"
  #} end if
  # get the log file name
  options.log_file_name = GetLogPath(options.input_path,
    task, options.output_dir)
  if (options.gene_list): #{
    options.recurrent_genes_path = os.path.join(options.output_dir,
        gene_filename)
        #"gene_recurrence.txt")
  #} end if
  options.lib_counts_path = os.path.join(options.output_dir,
        "library_counts.txt")
  if (0 < len(path_errors)): #{
    ErrMsg("Errors in input arguments:\n  %s" % "\n  ".join(path_errors))
  #} end if
  # the paths are good if there are no path errors and no conflicting options
  return (opts_good and 0 == len(path_errors))
#} end def

def Main(): #{
  # get options and arguments
  parser = SetupOptionsParser()
  (options, args) = parser.parse_args()
  # if the right number of args was used
  if (parser.num_args == len(args)): #{
    options.input_path = EnsureAbsPath(args[0])
    options.libs_dir   = EnsureAbsPath(args[1])
    if (CheckPaths(options)): #{
      try: #{
        main_class_object = PredictionSummarizerCls(options)
        WriteCommand(main_class_object, sys.argv)
        main_class_object.SummarizePredictions()
      except (MyError), e:
        ErrMsg("ERROR while summarizing Barnacle predictions:\n  %s" % e)
        return ES_RUN_ERR
      #} end try
    else:
      return ES_PATH_ERR
    #} end if
  else:
    parser.error("you must specify the path to a file containing a list of "
      "libraries with Barnacle results (LIBRARY_LIST); and the path to the "
      "base directory containing the data for those libraries (LIBS_DIR).")
    return ES_OPT_ERR
  #} end if
  return ES_SUCCESS
#} end def

if __name__ == '__main__': #{
  try: #{
    exit_status = Main()
  except Exception, e:
    traceback.print_exc()
    exit_status = ES_EXCEPTION
  #} end try
  if (ES_SUCCESS == exit_status): #{
    print MSG_SUCCESS
  else:
    print MSG_FAIL
  #} end if
  sys.exit(exit_status)
#} end if
