#! /usr/bin/env python
"""
create_report.py

Created by Lucas Swanson
Copyright (c) 2013 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

# import standard modules
from argparse import ArgumentParser
import os, sys, time, traceback

# ensure that the sys.path is set up to allow importing Barnacle modules
barnacle_dir = sys.path[0]
if not barnacle_dir:
    barnacle_dir = os.getcwd()
while not os.path.isfile(os.path.join(barnacle_dir, 'barnacle.pl')):
    barnacle_dir = os.path.dirname(barnacle_dir)
    if not barnacle_dir or '/' == barnacle_dir:
        print 'cannot find Barnacle install directory'
        raise Exception
barnacle_dir = os.path.expanduser(barnacle_dir)
barnacle_dir = os.path.abspath(barnacle_dir)
if barnacle_dir not in sys.path:
    sys.path.insert(0, barnacle_dir)

# import custom modules
from common.breakpoint import SortBreakpointTuple, BreakpointPairsAreEqual
from parsers.two_bit import TwoBitFileCls
from parsers.tokenizer import GetFieldAndValue, TokenizerCls
from parsers.candidate_group_parser import CandidateGroupParserCls
from utils.log import GetLogPath, CloseLogFile
from utils.error import MyError
from utils.general import (SetupMainClass, TimeSpent, WriteCommand, StrJoin)
from utils.messages import ErrMsg, LogMsg, DebugMsg, ExtremeDebugMsg
from utils.files_paths import (
    CheckDirPath, CheckFilePath, CheckNewFilePath, EnsureAbsPath,
    FileBoxCls, GetFilePath, GetOutDir)
from utils.argparse_extensions import common_parser

# CONSTANTS
ES_SUCCESS = 0
ES_RUN_ERR = 1
ES_OPT_ERR = 2
ES_PATH_ERR = 3
ES_EXCEPTION = 4
MSG_SUCCESS = 'SUCCESS'
MSG_FAIL = 'FAIL'
ITD = 'itd'
PTD = 'ptd'
FUSION = 'fusion'
EXTENSIONS = {
    ITD: ('itd itd.multi_exon itd.edge_gap itd.edge_gap.multi_exon '
          'itd.full_ctg_dup itd.full_ctg_dup.multi_exon'.split()),
    PTD: ['ptd'],
    FUSION: 'fus fus.mix_dirs'.split(),
}
PRESENT = 'present'
ABSENT = 'absent'
OF_GENE = 'GENE'
OF_GENES = 'GENES'
OF_TYPE = 'TYPE'
OF_PRESENCE = 'PRESENCE'
OF_EVENT_READS = 'EVENT_READS'
OF_BREAKPOINTS = 'BREAKPOINTS'
OF_SIZE = 'SIZE'
OF_SEQUENCE = 'SEQUENCE'
OF_ENDS = 'ENDS'
ALL_OUTPUT_FIELDS = [
    OF_GENE, OF_TYPE, OF_PRESENCE, OF_EVENT_READS, OF_BREAKPOINTS, OF_SIZE,
    OF_SEQUENCE,
]
DEFAULT_OUTPUT_FIELDS=','.join((OF_GENE, OF_TYPE, OF_PRESENCE))
UNKNOWN = 'unrecognized_field'
BREAKPOINT_EQUALITY_BUFFER = 5

class ReportGeneratorCls(object):

    """A class for generating custom-formatted reports from Barnacle
    results.

    """

    def __init__(self, options):
        SetupMainClass(self, options)
        self._report = ReportCls()

    def __del__(self):
        CloseLogFile(self)

    def create_report(self):
        """Generate a report."""
        LogMsg(self, 'Generating report...')
        start = time.time()
        self._load_report_template()
        self._load_report_data()
        self._write_report()
        LogMsg(self, 'Time spent generating report: {0}'.format(
            TimeSpent(start)))

    def _load_report_template(self):
        # Open the report template file and use it to initialize the report.
        LogMsg(self, 'Loading report template...')
        report_template = FileBoxCls(self.options.template_path, 'r')
        genome_file = None
        if self.options.twobit_genome is not None:
            LogMsg(self, 'Loading genome...')
            genome_file = TwoBitFileCls(self.options.twobit_genome)
        self._report.load_template(report_template, genome_file)

    def _load_report_data(self):
        # Load the report data for all the records loaded from the report
        # template.
        for event_type in self._report.event_types():
            LogMsg(self, 'Loading data for {0} events...'.format(event_type))
            data_paths = self._get_data_paths(event_type)
            for data_path in data_paths:
                self._report.load_data_file(event_type, data_path)

    def _get_data_paths(self, event_type):
        # Return the paths to all possible Barnacle output files for the
        # given event type.
        extensions = EXTENSIONS[event_type.lower()]
        data_paths = [GetFilePath(
            self.options.barnacle_dir, self.options.lib, extension) for
            extension in extensions]
        return data_paths

    def _write_report(self):
        # Create the report output file, and write the output for each record
        # to the report.
        LogMsg(self, 'Writing report...')
        report_file = FileBoxCls(self.options.report_path, 'w')
        self._report.write(report_file)
        report_file.close()


class ReportCls(object):

    """A report object, with structure defined by a report template
    and contents defined by Barnacle results.

    """

    def __init__(self):
        self.records_in_order = list()
        self.records_by_type_and_gene = dict()

    def load_template(self, template_file, genome_file):
        """Setup the report structure, using the provided report template."""
        for template_line in template_file:
            record = RecordCls(genome_file)
            record.parse(template_line)
            self.records_in_order.append(record)
            typed_records_by_gene = self.records_by_type_and_gene.setdefault(
                record.type, dict())
            typed_records_by_gene[record.gene] = record

    def event_types(self):
        return self.records_by_type_and_gene.keys()

    def load_data_file(self, event_type, data_path):
        """Load the Barnacle data for a file."""
        if not os.path.exists(data_path):
            return
        typed_records_by_gene = self.records_by_type_and_gene[event_type]
        parser = CandidateGroupParserCls(data_path)
        for group in parser:
            target_genes = group.GeneSet(retain_case=True).intersection(
                typed_records_by_gene.keys())
            for gene in target_genes:
                record = typed_records_by_gene[gene]
                if record.both_genes_match(group.GetGeneSets()):
                    record.load_data(group)

    def write(self, report_file):
        """Write the formatted data for each event in each record to the
        report file.

        """
        for record in self.records_in_order:
            report_file.WriteLine(record.output_events())


class RecordCls(object):

    """A record within a report, providing information about an event type
    within a gene.

    """

    def __init__(self, genome_file):
        self.type = None
        self.gene = None
        # Note: gene2 is only used for fusions.
        self.gene2 = None
        # _events data member is a list of EventCls objects.
        self._events = list()
        self._genome = genome_file
        self._output_fields = None
        # Functions for parsing record template fields.
        self._parse_field = {
            'type': self._parse_type,
            'gene': self._parse_gene,
            'gene2': self._parse_gene2,
            'output_fields': self._parse_output_fields,
        }

    def __str__(self):
        data_list = [
            'type:{0}'.format(self.type),
            'gene:{0}'.format(self.gene),
        ]
        if self.gene2 is not None:
            data_list.append('gene2:{0}'.format(self.gene2))
        data_list.append('output_fields:{0}'.format(','.join(self._output_fields)))
        return ' '.join(data_list)

    def parse(self, template_line):
        """Setup the record structure, using the provided record template."""
        fields = TokenizerCls(template_line, delimiter=' ')
        for field_name, field_value in map(GetFieldAndValue, fields):
            try:
                self._parse_field[field_name](field_value)
            except KeyError:
                LogMsg(
                    self, 'Warning: unrecognized field type in report '
                    'template: {0}'.format(field_name))
        if self.type is None:
            raise ReportGeneratorError(
                'Cannot create record from template line without type '
                'field: {0}'.format(template_line))
        if self.gene is None:
            raise ReportGeneratorError(
                'Cannot create record from template line without gene '
                'field: {0}'.format(template_line))
        if FUSION == self.type and self.gene2 is None:
            raise ReportGeneratorError(
                'Cannot create fusion record from template line without '
                'gene2 field: {0}'.format(template_line))
        if self._output_fields is None:
            self._parse_output_fields(DEFAULT_OUTPUT_FIELDS)

    def present(self):
        """Return true if at least one Barnacle prediction indicates the
        presence of the event this record represents.

        """
        if self._events:
            return True
        return False

    def load_data(self, group):
        """Load the data from a Barnacle result group into the record."""
        if self.present():
            LogMsg(
                self, 'WARNING: multiple Barnacle predictions found for '
                '{0} {1}.'.format(self.gene, self.type))
        candidates = self._select_candidates(group)
        for candidate in candidates:
            self._add_event(candidate)

    def output_events(self):
        """Return the formatted data for each event associated with this
        record, or a string of default values if no events are associated
        with this record.

        """
        if self._events:
            return '\n'.join(
                event.output_string(self._output_fields) for
                event in self._events)
        else:
            return StrJoin(
                ',', (self._default_value(field_name) for field_name in
                self._output_fields))

    def both_genes_match(self, gene_sets_pair):
        """Return true if both genes for this record are represented by a
        different one of the two gene sets provided, false otherwise.

        """
        if self.gene2 is None:
            return (self.gene in gene_sets_pair[0] or
                    self.gene in gene_sets_pair[1])
        if (self.gene not in gene_sets_pair[0] or
            self.gene2 not in gene_sets_pair[1]):
            gene_sets_pair = list(reversed(gene_sets_pair))
        return (self.gene in gene_sets_pair[0] and
                self.gene2 in gene_sets_pair[1])

    def _parse_type(self, value):
        # Set the event type.
        self.type = value

    def _parse_gene(self, value):
        # Set the target gene.
        self.gene = value

    def _parse_gene2(self, value):
        # Set the second target gene (for fusions).
        self.gene2 = value

    def _parse_output_fields(self, value):
        # Set the desired output fields, in the order they should appear.
        self._output_fields = value.split(',')
        if OF_SEQUENCE in self._output_fields and self._genome is None:
            LogMsg(
                self, 'WARNING: cannot get sequence data for {0} {1} '
                'without 2bit genome sequence file.'.format(
                    self.gene, self.type))

    def _default_value(self, field_name):
        # Return the default value for the output field with the given name.
        if OF_GENE == field_name:
            return self.gene
        if OF_GENES == field_name:
            return '/'.join(sorted((self.gene, self.gene2)))
        if OF_TYPE == field_name:
            return self.type
        if OF_PRESENCE == field_name:
            return 'absent'
        return 'N/A'

    def _select_candidates(self, group):
        # Select a candidate from a Barnacle prediction group to represent
        # each distinct event present in the candidates for that group.
        # For now, regroup the candidates by contig, then pick the first
        # split-candidate encountered for each contig.
        candidates_by_contig = dict()
        for candidate in group.candidates:
            candidates_by_contig.setdefault(
                candidate.contig_info.id, list()).append(candidate)
        selected_candidates = list()
        for candidates_for_contig in candidates_by_contig.itervalues():
            selected_candidates.append(
                self._select_candidate_for_contig(candidates_for_contig))
        return selected_candidates

    def _select_candidate_for_contig(self, potential_candidates):
        # Pick the first split-candidate from the list of potential
        # candidates, or the the gap-candidate representing the largest
        # duplication if there are no split-candidates.
        for candidate in potential_candidates:
            if not candidate.gap:
                return candidate
        # Return gap-candidate with 'biggest' duplication.
        return max(
            potential_candidates,
            key=_duplication_size_from_candidate_breakpoints)

    def _add_event(self, candidate):
        # Create an event from a Barnacle candidate and add it to the list
        # of candidates for this record.
        new_event = EventCls(
            self.gene, self.type, candidate, self._genome, self.gene2)
        for field_name in self._output_fields:
            try:
                load_data_function = new_event.load_output_data[field_name]
            except KeyError:
                LogMsg(
                    self, 'Warning: unrecognized output field type in '
                    'report template: {0}'.format(field_name))
                continue
            load_data_function(field_name, candidate)
        # Determine whether the candidate represents the same event as any of
        # the already existing events for this record. If so, choose the one
        # with more read support.
        for i, old_event in enumerate(self._events):
            if old_event.is_duplicate_of(new_event):
                if new_event.has_more_support_than(old_event):
                    self._events[i] = new_event
                return
        # If the candidate does not represent the same event as any existing
        # events, add it to the list of events.
        self._events.append(new_event)

class EventCls(object):

    """Details of a specific event that can be written into the report."""

    def __init__(self, gene, type, candidate, genome_file=None, gene2=None):
        self._gene = gene
        # Note: _gene2 is only used for fusions.
        self._gene2 = gene2
        self._sorted_genes = sorted((self._gene, self._gene2))
        self._type = type
        self._genome = genome_file
        self._output_data = dict()
        # Functions for loading data for output fields.
        self.load_output_data = {
            OF_GENE: self._load_gene_output,
            OF_GENES: self._load_genes_output,
            OF_TYPE: self._load_type_output,
            OF_PRESENCE: self._load_presence_output,
            OF_EVENT_READS: self._load_event_reads_output,
            OF_BREAKPOINTS: self._load_breakpoints_output,
            OF_SIZE: self._load_size_output,
            OF_SEQUENCE: self._load_sequence_output,
            OF_ENDS: self._load_ends_output,
        }
        # Store the sorted candidate breakpoints and read-support for
        # comparing potential duplicates.
        self._candidate_breakpoints = candidate.breakpoints
        self._candidate_support = candidate.read_to_ctg_unique

    def is_duplicate_of(self, other):
        return BreakpointPairsAreEqual(
            self._candidate_breakpoints, other._candidate_breakpoints,
            buffer=BREAKPOINT_EQUALITY_BUFFER, check_dirs=True)

    def has_more_support_than(self, other):
        return self._candidate_support > other._candidate_support

    def output_string(self, output_fields):
        return StrJoin(
            ',', (self._output_data.get(field_name, UNKNOWN) for
            field_name in output_fields))

    def _load_gene_output(self, field_name, candidate):
        # Load the gene output value from the provided Barnacle prediction.
        # Also, check that the candidate gene(s) match the gene from the record.
        candidate_genes = set(candidate.genes_A.keys())
        if not candidate.gap:
            candidate_genes = candidate_genes.union(candidate.genes_B.keys())
        if self._gene not in candidate_genes:
            raise ReportGeneratorError(
                'Member does not overlap target gene: {0} ({1}).'.format(
                    self._gene, candidate.MainGenesString()))
        self._output_data[field_name] = self._gene

    def _load_genes_output(self, field_name, candidate):
        # Load the two genes output value from the provided Barnacle
        # fusion prediction.
        # Also, check that the candidate gene(s) match the gene from the record.
        candidate_genes = candidate.genes_A.keys()
        candidate_genes2 = candidate.genes_B.keys()
        if (self._gene not in candidate_genes or
            self._gene2 not in candidate_genes2):
            candidate_genes, candidate_genes2 = (
                candidate_genes2, candidate_genes)
        if (self._gene not in candidate_genes or
            self._gene2 not in candidate_genes2):
            raise ReportGeneratorError(
                'Member does not overlap target genes: {0} ({1}).'.format(
                    '/'.join(self._sorted_genes),
                    candidate.MainGenesString()))
        self._output_data[field_name] = '/'.join(self._sorted_genes)

    def _load_type_output(self, field_name, candidate):
        # Load the gene output value from the provided Barnacle prediction.
        # Actually, just use the type from the record.
        self._output_data[field_name] = self._type

    def _load_presence_output(self, field_name, candidate):
        # Load the presence output value from the provided Barnacle
        # prediction.
        self._output_data[field_name] = PRESENT

    def _load_event_reads_output(self, field_name, candidate):
        # Load the event reads output value from the provided Barnacle
        # prediction.
        self._output_data[field_name] = candidate.read_to_ctg_unique

    def _load_breakpoints_output(self, field_name, candidate):
        # Load the breakpoints output value from the provided Barnacle
        # prediction. Use one format for single gene event types (presumably
        # they will also involve a range on a single chromosome), and a
        # different format two gene events (presumably they will involve
        # independent breakpoints, possibly on distinct chromosomes).
        # TODO: This might need to be different for different event types.
        if self._gene2 is None:
            breakpoints = SortBreakpointTuple(candidate.breakpoints)
            self._output_data[field_name] = 'chr{0}:{1}'.format(
                breakpoints[0].chr, StrJoin(
                    '-', (bp.coord for bp in breakpoints)))
        else:
            breakpoints = candidate.breakpoints
            # Ensure the breakpoints are in the same order as the
            # alphabetically sorted genes.
            if (self._sorted_genes[0] not in candidate.genes_A or
                self._sorted_genes[1] not in candidate.genes_B):
                breakpoints = reversed(breakpoints)
            self._output_data[field_name] = '/'.join('chr{0}:{1}'.format(
                bp.chr, bp.coord) for bp in breakpoints)

    def _load_size_output(self, field_name, candidate):
        # Load the size output value from the provided Barnacle prediction.
        # TODO: This might need to be different for different event types.
        self._output_data[field_name] = candidate.meta_fields.get(
            'breakpoint_dist', _duplication_size_from_candidate_breakpoints(
                candidate))

    def _load_sequence_output(self, field_name, candidate):
        # Load the event sequence output value from the provided Barnacle
        # prediction and stored genome sequence file.
        if self._genome is None:
            LogMsg(
                self, 'WARNING: cannot get sequence data for {0} {1} '
                'without 2bit genome sequence file.'.format(
                    self.gene, self.type))
            return
        breakpoints = SortBreakpointTuple(candidate.breakpoints)
        self._output_data[field_name] = self._genome.GetSequence(
            breakpoints[0].chr, breakpoints[0].coord, breakpoints[1].coord)

    def _load_ends_output(self, field_name, candidate):
        # Load the ends (5' or 3') for each gene involved in the fusion.
        ends = candidate.ends
        # Ensure the ends are in the same order as the alphabetically sorted
        # genes.
        if (self._sorted_genes[0] not in candidate.genes_A or
            self._sorted_genes[1] not in candidate.genes_B):
            ends = reversed(ends)
        self._output_data[field_name] = '/'.join(ends)


def _duplication_size_from_candidate_breakpoints(candidate):
    breakpoints = SortBreakpointTuple(candidate.breakpoints)
    if breakpoints[0].chr != breakpoints[1].chr:
        raise ReportGeneratorError(
            'Cannot get duplication size from candidate breakpoints when '
            'breakpoints are on different chromosomes:\n{0}'.format(
                candidate.DataString()))
    return (breakpoints[1].coord - breakpoints[0].coord) + 1


#### EXCEPTION CLASSES ####
class ReportGeneratorError(MyError):
    """For errors raised while generating a report."""
    pass


def create_argument_parser():
    """Define the positional and optional arguments that this script can
    accept.

    """
    unique_parser = ArgumentParser(add_help=False)
    # Positional arguments:
    unique_parser.add_argument(
        'lib', help='The ID of the library to process.')
    # TODO: Consider adding AbsPathArg function that converts exceptions to
    #       argparse.ArgumentTypeError exceptions?
    unique_parser.add_argument(
        'barnacle_dir', type=EnsureAbsPath,
        help='The path to the directory containing the Barnacle prediction '
             'data files to process.')
    unique_parser.add_argument(
        'template_path', type=EnsureAbsPath,
        help='The path to the report template file to use.')
    # Optional arguments:
    unique_parser.add_argument(
        '--twobit-genome', type=EnsureAbsPath,
        help='The path to a 2bit formatted reference genome sequence file, '
             'used to extract event sequences.')
    description_string = (
        'Parse Barnacle results to create a report file based on the '
        'provided report template file. The report template file should '
        'have one record per line, with these space-delimited template '
        'fields: type, gene, and output_fields. "type" is the type of '
        'event: ITD, PTD, or FUSION. "gene" is the target gene. '
        '"output_fields" is a comma-delimited list that defines the '
        'contents and order of the output produced for each record. Valid '
        'values for items in the "output_fields" template field are: '
        '{0}.'.format(', '.join(ALL_OUTPUT_FIELDS)))
    return ArgumentParser(
        description=description_string,
        parents=[unique_parser, common_parser()])


def check_paths(options):
    """Construct paths strings used by this script, and check that all
    provided and constructed path strings are valid paths.

    """
    opts_good = True
    if not opts_good:
        ErrMsg('bad option') #TODO
        opts_good = False
    path_errors = list()
    # Check the input paths.
    CheckDirPath(options.barnacle_dir, 'Barnacle predictions', path_errors)
    CheckFilePath(options.template_path, 'report template', path_errors)
    if options.twobit_genome is not None:
        CheckFilePath(options.twobit_genome, 'genome sequence', path_errors)
    # Get and check the output and log paths.
    if opts_good and not path_errors:
        report_dir = GetOutDir(options.barnacle_dir, 'report')
        CheckDirPath(
            report_dir, 'report', path_errors, create=True, replace=True)
        options.report_path = GetFilePath(
            report_dir, options.lib, 'report.csv')
        if not options.force:
            CheckNewFilePath(options.report_path, 'report', path_errors)
        options.log_file_name = GetLogPath(
            options.template_path, 'create_report', report_dir)
    if 0 < len(path_errors):
        ErrMsg('Errors in input arguments:\n  {0}'.format(
            '\n  '.join(path_errors)))
    # The paths are good if there are no path errors and
    # no conflicting options.
    return (opts_good and not path_errors)


def main():
    """Create a report generator object and have it generate a report."""
    parser = create_argument_parser()
    # For backwards compatibility, call the return value of parse_args
    # 'options', rather than following convention and calling it 'args'.
    options = parser.parse_args()
    if check_paths(options):
        try:
            report_generator = ReportGeneratorCls(options)
            WriteCommand(report_generator, sys.argv)
            report_generator.create_report()
        except (MyError), e:
            ErrMsg('ERROR while generating report:\n  {0}'.format(e))
            return ES_RUN_ERR
    else:
        return ES_PATH_ERR
    return ES_SUCCESS


if __name__ == '__main__':
    try:
        exit_status = main()
    except Exception, e:
        traceback.print_exc()
        exit_status = ES_EXCEPTION
    if ES_SUCCESS == exit_status:
        print MSG_SUCCESS
    else:
        print MSG_FAIL
    sys.exit(exit_status)
