#! /usr/bin/env python

# import standard modules
from optparse import OptionParser, OptionGroup
import os, re, sys, time, traceback

# ensure that the sys.path is set up to allow importing Barnacle modules
barnacle_dir = sys.path[0]
if ("" == barnacle_dir): #{
  barnacle_dir = os.getcwd()
#} end if
while (not os.path.isfile(os.path.join(barnacle_dir, "barnacle.pl"))): #{
  barnacle_dir = os.path.dirname(barnacle_dir)
  if ("" == barnacle_dir or "/" == barnacle_dir): #{
    print "cannot find BARNACLE directory"
    raise Exception
  #} end if
#} end while
barnacle_dir = os.path.expanduser(barnacle_dir)
barnacle_dir = os.path.abspath(barnacle_dir)
if (barnacle_dir not in sys.path): #{
  sys.path.insert(0, barnacle_dir)
#} end if

# import custom modules
from version import VERSION
from utils.log import GetLogPath, CloseLogFile
from utils.error import MyError
from utils.general import (SetupMainClass, TimeSpent, WriteCommand,
  OppositeStrand, StrJoin)
from utils.messages import ErrMsg, LogMsg, DebugMsg, ExtremeDebugMsg
from utils.files_paths import (CheckFilePath, CheckDirPath, EnsureAbsPath,
  GetOutDir, FileBoxCls)
from parsers.candidate_group_parser import CandidateGroupParserCls
from common.coord_pair import CutOrExtendBlocks
from common.breakpoint import BreakpointTupleIsSorted
from common.grouped_candidate import IsLeftBreakpoint
from simcheck import WriteSimCheckRecord

# CONSTANTS
ES_SUCCESS = 0
ES_RUN_ERR = 1
ES_OPT_ERR = 2
ES_PATH_ERR = 3
ES_EXCEPTION = 4
MSG_SUCCESS = "SUCCESS"
MSG_FAIL = "FAIL"
END_L = 0
END_R = 1

class MainClassCls: #{
  def __init__(self, options): #{
    SetupMainClass(self, options)
  #} end def

  def __del__(self): #{
    CloseLogFile(self)
  #} end def

  def Run(self): #{
    LogMsg(self, "Converting Barnacle file...")
    start = time.time()
    outfile = FileBoxCls(self.options.output_path, "w")
    LogMsg(self, "Writing output to %s" % self.options.output_path)
    for group in CandidateGroupParserCls(self.options.barnacle_path): #{
      for candidate in group.candidates: #{
        simcheck_ends = self.ConvertCandidate(candidate)
        ExtremeDebugMsg(self, "Created %i records" % len(simcheck_ends))
        for simcheck_record in simcheck_ends: #{
          ExtremeDebugMsg(self, "Writing record: %s" % simcheck_record)
          WriteSimCheckRecord(simcheck_record, outfile, self.log_info)
        #} end for
      #} end for
    #} end for
    outfile.Close()
    LogMsg(self, "Time spent converting Barnacle file: %s" %
      TimeSpent(start))
  #} end def

  def ConvertCandidate(self, candidate): #{
    fus_id = "barnacle%i" % candidate.group_id
    simcheck_ends = [dict(id=fus_id, end=ctg_end) for
        ctg_end in (END_L, END_R)]
    for (sc_record, align, blocks) in zip(simcheck_ends,
        candidate.alignments, candidate.blocks): #{
      sc_record["chrom"] = align.chrom
      if (END_L == sc_record["end"]): #{
        sc_record["strand"] = align.Strand()
      else:
        sc_record["strand"] = OppositeStrand(align.Strand())
      #} end if
      is_left_breakpoint = IsLeftBreakpoint(END_L == sc_record["end"],
          align.IsPosStrand())
      sc_record["blocks"] = CutOrExtendBlocks(blocks, self.options.read_length,
          is_left_breakpoint)
      ExtremeDebugMsg(self, "Created record: %s" % sc_record)
    #} end for
    ExtremeDebugMsg(self, "Breakpoints: %s" % StrJoin(" ",
      candidate.breakpoints))
    if (not BreakpointTupleIsSorted(candidate.breakpoints)): #{
      ExtremeDebugMsg(self, "Swapping ends!")
      simcheck_ends[END_L]["end"] = END_R
      simcheck_ends[END_R]["end"] = END_L
    #} end if
    return simcheck_ends
  #} end def
#} end class

#### EXCEPTION CLASSES ####
class MainClassError(MyError): #{
  pass
#} end class

def SetupOptionsParser(): #{
  description_string = ("Converts a barnacle data file to the format "
      "expected by the \"simcheck\" tool")
  args = [ "LIB", "BARNACLE_FILE", "READ_LENGTH" ]
  usage_string = "%prog " + " ".join(args) + " [ OPTIONS ]"
  parser = OptionParser(description=description_string,
                        version="%prog " + VERSION,
                        usage=usage_string)
  parser.num_args = len(args)
  misc_group = OptionGroup(parser, "Miscellaneous Options")
  misc_group.add_option("-f", "--force",
                    action="store_true",
                    help="Force filtering to take place, even if the output "
                         "directory already exists.")
  misc_group.add_option("-d", "--debug",
                    action="store_true",
                    help="Print debug information while the program runs.")
  misc_group.add_option("--extreme-debug",
                    action="store_true", dest="extreme_debug",
                    help="Print extremely in-depth debug information while "
                      "the program runs. Not recommended for large jobs.")
  parser.add_option_group(misc_group)
  parser.set_defaults(force=False,
                      debug=False,
                      extreme_debug=False)
  return parser
#} end def

def CheckPaths(options): #{
  opts_good = True
  try: #{
    options.read_length = int(options.read_length)
    if (0 >= options.read_length): #{
      ErrMsg("Read length must be a positive integer, not %s" %
          options.read_length)
      opts_good = False
    #} end if
  except ValueError:
    ErrMsg("Read length must be a positive integer, not %s" %
        options.read_length)
    opts_good = False
  #} end if
  path_errors = list()
  CheckFilePath(options.barnacle_path, "Barnacle predictions", path_errors)
  # get and check the output path
  (options.output_dir, infilename) = os.path.split(options.barnacle_path)
  options.output_path = os.path.join(options.output_dir,
      "%s.simcheck" % infilename)
  if (opts_good and 0 == len(path_errors)): #{
    CheckDirPath(options.output_dir, "output", path_errors,
      create=False, replace=True)
    # get the log file name
    options.log_file_name = GetLogPath(options.barnacle_path,
      "bdf2simcheck", options.output_dir)
  #} end if
  if (0 < len(path_errors)): #{
    ErrMsg("Errors in input arguments:\n  %s" % "\n  ".join(path_errors))
  #} end if
  # the paths are good if there are no path errors and no conflicting options
  return (opts_good and 0 == len(path_errors))
#} end def

def Main(): #{
  # get options and arguments
  parser = SetupOptionsParser()
  (options, args) = parser.parse_args()
  # if the right number of args was used
  if (parser.num_args == len(args)): #{
    options.lib           = args[0]
    options.barnacle_path = EnsureAbsPath(args[1])
    options.read_length   = args[2]
    if (CheckPaths(options)): #{
      try: #{
        main_class_object = MainClassCls(options)
        WriteCommand(main_class_object, sys.argv)
        main_class_object.Run()
      except (MyError), e:
        ErrMsg("ERROR while converting Barnacle file:\n  %s" % e) #TODO
        return ES_RUN_ERR
      #} end try
    else:
      return ES_PATH_ERR
    #} end if
  else:
    parser.error("you must specify a library name (LIB); the path to a "
      "Barnacle data file for that library (BARNACLE_FILE); and the "
      "length of the reads that were assembled (READ LENGTH).")
    return ES_OPT_ERR
  #} end if
  return ES_SUCCESS
#} end def

if __name__ == '__main__': #{
  try: #{
    exit_status = Main()
  except Exception, e:
    traceback.print_exc()
    exit_status = ES_EXCEPTION
  #} end try
  if (ES_SUCCESS == exit_status): #{
    print MSG_SUCCESS
  else:
    print MSG_FAIL
  #} end if
  sys.exit(exit_status)
#} end if
