#!/usr/bin/env perl

# This is a program that takes a file with a list of stable IDs (not
# exon stable IDs), and outputs a comma-separated list of the history of
# each of these stable IDs.  The history ends with either the current
# release or at the point when the stable ID was retired.

use strict;
use warnings;

use Bio::EnsEMBL::Registry;
use IO::File;
use Getopt::Long;
use DBI qw(:sql_types);

my ( $filename, $species );
my $help = '';

if ( !GetOptions( 'file|f=s'    => \$filename,
                  'species|s=s' => \$species,
                  'help|h!'     => \$help )
     || !defined($species)
     || $help )
{
  print <<END_USAGE;

Usage:
  $0 --species=species [ --file=filename ]

  $0 --help


    --species / -s  Name of species.

    --file    / -f  (Optional) Name of file containing a list of stable
                    IDs.  The default is to read the list from standard
                    input.

    --help    / -h  To see this text.

Example usage:

  $0 -s mouse -f idlist.txt

  # Same as the above:
  $0 -s mouse < idlist.txt

  sort -u longlist.txt | $0 -s human

END_USAGE

  exit(1);
} ## end if ( !GetOptions( 'file|f=s'...))

$filename ||= '-';

my $registry = 'Bio::EnsEMBL::Registry';

$registry->load_registry_from_db( -host => 'ensembldb.ensembl.org',
                                  -user => 'anonymous' );
#Bio::EnsEMBL::Registry->load_registry_from_db(
#  -HOST => 'ensembldb.ensembl.org',
#  -PORT => 5306,
#  -USER => 'anonymous',
#  -VERBOSE => 0,
#  -DB_VERSION => 67
#);

my $adaptor = $registry->get_DBAdaptor( $species, 'Core' );
#my $dba = Bio::EnsEMBL::Registry->get_DBAdaptor('human', 'core');

my $mc = $adaptor->get_MetaContainer();
#my $mc = $dba->get_MetaContainer();
my $release = $mc->get_schema_version();

my $in = IO::File->new($filename);

if ( !defined($in) ) {
  die( sprintf( "Could not open file '%s' for reading", $filename ) );
}

my $asia = $adaptor->get_ArchiveStableIdAdaptor();
my $ta = $adaptor->get_TranscriptAdaptor();
#my $asia = $dba->get_ArchiveStableIdAdaptor();
#my $ta = $dba->get_TranscriptAdaptor();

# We could do what we want to do with the API, but this is simpler and
# quicker, at the moment.  As always, when using plain SQL against our
# databases, the user should not be surprised to see the code break when
# we update the schema...
#my $statement = q(
#SELECT  old_version,
#        old_release,
#        new_stable_id, new_version,
#        new_release,
#        score
#FROM    stable_id_event
#  JOIN  mapping_session USING (mapping_session_id)
#WHERE   old_stable_id = ?
#ORDER BY old_version ASC, CAST(new_release AS UNSIGNED)
#);

#my $sth = $adaptor->dbc()->db_handle()->prepare($statement);

while ( my $old_id = $in->getline() ) {
  chomp($old_id);

  # Strip off any comment (from '#' to the end of the line).
  $old_id =~ s/\s*#.*$//;

  # Skip lines containing only whitespace.
  if ( $old_id =~ /^\s*$/ ) { next }

  print("Old stable ID, New stable ID, Release, Mapping score\n");

  # Get our original ID
  my $archive_id = $asia->fetch_by_stable_id($old_id);
  # Then get the associated archive info which tells us what Transcripts it had at retirement
  my $archived_info = $archive_id->get_all_associated_archived();
  my %possible_new_genes;
  foreach my $associated (@{$archived_info}) {
    my ($arch_gene, $arch_tr, $arch_tl, $pep_seq) = @{$associated};
    #printf("TRANSCRIPT: %s\n", $arch_tr->stable_id());
    my $successor = $arch_tr->get_latest_incarnation();
    if($successor->release() == $release) {
      my $live_transcript = $ta->fetch_by_stable_id($successor->stable_id());
      my $gene = $live_transcript->get_Gene();
      #printf("  GENE: %s\n", $gene->stable_id());
      $possible_new_genes{$gene->stable_id()} = $gene;
    }
  }

  foreach my $new_id (keys %possible_new_genes) {
    #printf("%d | %s -> %s\n", $release, $old_id, $new_id);
    printf("%s, %s, %d, 0\n", $old_id, $new_id, $release);
  }

  #$sth->bind_param( 1, $old_id, SQL_VARCHAR );
  #
  #$sth->execute();
  #
  #my ( $version, $release, $new_stable_id, $new_version, $new_release,
  #     $score );
  #
  #$sth->bind_columns( \( $version,     $release,     $new_stable_id,
  #                       $new_version, $new_release, $score ) );
  #
  #while ( $sth->fetch() ) {
  #  if ( defined($new_stable_id) ) {
  #    printf( "%s.%s, %s.%s, %s, %s\n",
  #            $old_id, $version, $new_stable_id, $new_version,
  #            $new_release, $score );
  #  } elsif ( !defined($new_stable_id) ) {
  #    printf( "%s.%s, <retired>, %s, %s\n",
  #            $old_id, $version, $new_release, $score );
  #  }
  #}
  #
  print("\n");
} ## end while ( my $old_id = $in...)

#NEW
