#! /usr/bin/env python

# import standard modules
from optparse import OptionParser, OptionGroup
import os, re, sys, time, traceback

# import custom modules
from version import VERSION
from utils.log import GetLogPath, CloseLogFile
from utils.error import MyError
from utils.general import (SetupMainClass, TimeSpent, WriteCommand, StrJoin,
  NormalizeChrID, RunOverlapCode, RemoveIndexingFlag)
from utils.messages import ErrMsg, LogMsg, DebugMsg, ExtremeDebugMsg
from utils.files_paths import (CheckFilePath, CheckDirPath, EnsureAbsPath,
  GetOutDir, FileBoxCls, CleanLine)
from utils.subprocesses import STREAM_OUT
from parsers.tokenizer import TokenizerCls

# CONSTANTS
ES_SUCCESS = 0
ES_RUN_ERR = 1
ES_OPT_ERR = 2
ES_PATH_ERR = 3
ES_EXCEPTION = 4
MSG_SUCCESS = "SUCCESS"
MSG_FAIL = "FAIL"
END_L = 0
END_R = 1
ID_FIELD = 3

class MainClassCls: #{
  def __init__(self, options): #{
    SetupMainClass(self, options)
  #} end def

  def __del__(self): #{
    CloseLogFile(self)
  #} end def

  def Run(self): #{
    LogMsg(self, "Checking predictions against groundtruth...")
    start = time.time()
    events = dict(gtruth=dict(), pred=dict())
    for etype in events: #{
      self.CreateCoordsFile(etype, events[etype])
    #} end for
    overlap_stream = RunOverlapCode(self.options.coords_paths["gtruth"],
      self.options.coords_paths["pred"], STREAM_OUT, log_info=self.log_info)
    self.ProcessOverlaps(overlap_stream, events)
    self.CountTrueFalse(events)
    LogMsg(self, "Time spent checking predictions against groundtruth: %s" %
      TimeSpent(start))
  #} end def

  def CreateCoordsFile(self, etype, edict): #{
    ExtremeDebugMsg(self, "Creating coordinates file for: %s" %
        getattr(self.options, "%s_path" % etype))
    infile = FileBoxCls(getattr(self.options, "%s_path" % etype), "r")
    coords_file = FileBoxCls(self.options.coords_paths[etype], "w")
    for simcheck_record in infile: #{
      ExtremeDebugMsg(self, "  Record: %s" % simcheck_record)
      (id, end, chrom, strand, left, right) = simcheck_record.split("\t")
      # only use the portion of the ID preceding any spaces
      id = id.split(" ")[0]
      if (id not in edict): #{
        edict[id] = dict(overlaps_0=set(), overlaps_1=set())
      #} end if
      chrom = NormalizeChrID(chrom)
      strand_key = "strand_%s" % end
      if (strand_key in edict[id] and strand != edict[id][strand_key]): #{
        raise MainClassError("inconsistent strand for event %s" % id)
      #} end if
      edict[id][strand_key] = strand
      coords_str = " ".join((chrom, left, right, ".".join((id, end))))
      coords_file.WriteLine(coords_str)
    #} end for
    coords_file.Close()
    infile.Close()
    for event_key in edict: #{
      for end in (END_L, END_R): #{
        strand_key = "strand_%i" % end
        if (strand_key not in edict[event_key]): #{
          raise MainClassError("End %i strand missing for event %s" %
              (end, event_key))
        #} end if
      #} end for
    #} end for
  #} end def

  def ProcessOverlaps(self, overlap_stream, events): #{
    ExtremeDebugMsg(self, "Processing overlaps...")
    for overlap_line in overlap_stream: #{
      overlap_line = CleanLine(overlap_line)
      ExtremeDebugMsg(self, "OVERLAP: \"%s\"" % overlap_line)
      overlap_fields = overlap_line.split(" ")
      (pred_id, gtruth_ids) = overlap_fields[ID_FIELD:]
      ExtremeDebugMsg(self, "  Pred: %s, Truths: %s" % (pred_id, gtruth_ids))
      (pred_id, pred_end) = pred_id.split(".")
      p_strand_key = "strand_%s" % pred_end
      for gtruth_id in TokenizerCls(gtruth_ids, delimiter="^"): #{
        ExtremeDebugMsg(self, "    Truth: %s" % (gtruth_id))
        gtruth_id = RemoveIndexingFlag(gtruth_id)
        (gtruth_id, gtruth_end) = gtruth_id.split(".")
        g_strand_key = "strand_%s" % gtruth_end
        ExtremeDebugMsg(self, " "*6 + "PStrand: %s, GStrand: %s" %
          (events["pred"][pred_id][p_strand_key],
            events["gtruth"][gtruth_id][g_strand_key]))
        if (events["pred"][pred_id][p_strand_key] ==
            events["gtruth"][gtruth_id][g_strand_key]): #{
          p_overlaps_key = "overlaps_%s" % pred_end
          events["pred"][pred_id][p_overlaps_key].add(gtruth_id)
          g_overlaps_key = "overlaps_%s" % gtruth_end
          events["gtruth"][gtruth_id][g_overlaps_key].add(pred_id)
        #} end if
      #} end for
    #} end for
  #} end def

  def CountTrueFalse(self, events): #{
    counts = dict(TP=0,FN=0,FP=0)
    tp_file = FileBoxCls(self.options.tp_path, "w")
    fn_file = FileBoxCls(self.options.fn_path, "w")
    for gkey in events["gtruth"]: #{
      if (self.CheckEvent(gkey, events["gtruth"][gkey])): #{
        counts["TP"] += 1
        tp_file.WriteLine("%s: %s" % (gkey,
          ",".join(events["gtruth"][gkey]["shared"])))
      else:
        counts["FN"] += 1
        fn_file.WriteLine(gkey)
      #} end if
    #} end for
    tp_file.Close()
    fn_file.Close()
    fp_file = FileBoxCls(self.options.fp_path, "w")
    for pkey in events["pred"]: #{
      if (not self.CheckEvent(pkey, events["pred"][pkey])): #{
        counts["FP"] += 1
        fp_file.WriteLine(pkey)
      #} end if
    #} end for
    fp_file.Close()
    counts["TPR"] = counts["TP"] / float(counts["TP"] + counts["FN"])
    counts["FDR"] = counts["FP"] / float(counts["TP"] + counts["FP"])
    for key in ["TP", "FN", "FP"]: #{
      LogMsg(self, "%s: %i" % (key, counts[key]))
    #} end for
    for key in ["TPR", "FDR"]: #{
      LogMsg(self, "%s: %.4f" % (key, counts[key]))
    #} end for
  #} end def

  def CheckEvent(self, event_id, event): #{
    ExtremeDebugMsg(self, "Checking event: %s\n  O0: %s\n  O1: %s" %
      (event_id, ",".join(sorted(event["overlaps_0"])),
        ",".join(sorted(event["overlaps_1"]))))
    event["shared"] = event["overlaps_0"].intersection(event["overlaps_1"])
    return 0 < len(event["shared"])
  #} end def
#} end class

def WriteSimCheckRecord(simcheck_record, outfile, log_info=None): #{
  ExtremeDebugMsg(log_info, "Writing record %s.%i" % (simcheck_record["id"],
    simcheck_record["end"]))
  for block in simcheck_record["blocks"]: #{
    output = StrJoin("\t", (simcheck_record["id"], simcheck_record["end"],
      simcheck_record["chrom"], simcheck_record["strand"],
      block.min, block.max))
    outfile.WriteLine(output)
  #} end for
#} end def

#### EXCEPTION CLASSES ####
class MainClassError(MyError): #{
  pass
#} end class

def SetupOptionsParser(): #{
  description_string = ("") #TODO
  args = [ "GROUNDTRUTH_FILE", "PREDICTION_FILE", ] #TODO
  usage_string = "%prog " + " ".join(args) + " [ OPTIONS ]"
  parser = OptionParser(description=description_string,
                        version="%prog " + VERSION,
                        usage=usage_string)
  parser.num_args = len(args)
  misc_group = OptionGroup(parser, "Miscellaneous Options")
  misc_group.add_option("-f", "--force",
                    action="store_true",
                    help="Force filtering to take place, even if the output "
                         "directory already exists.")
  misc_group.add_option("-d", "--debug",
                    action="store_true",
                    help="Print debug information while the program runs.")
  misc_group.add_option("--extreme-debug",
                    action="store_true", dest="extreme_debug",
                    help="Print extremely in-depth debug information while "
                      "the program runs. Not recommended for large jobs.")
  parser.add_option_group(misc_group)
  parser.set_defaults(force=False,
                      debug=False,
                      extreme_debug=False)
  return parser
#} end def

def CheckPaths(options): #{
  opts_good = True
  if (not opts_good): #{
    ErrMsg("bad option") #TODO
    opts_good = False
  #} end if
  path_errors = list()
  CheckFilePath(options.gtruth_path, "ground truth fusions", path_errors)
  gtruth_filename = os.path.basename(options.gtruth_path)
  CheckFilePath(options.pred_path, "predicted fusions", path_errors)
  # get and check the output path
  (options.output_dir, pred_filename) = os.path.split(options.pred_path)
  if (opts_good and 0 == len(path_errors)): #{
    options.tp_path = os.path.join(options.output_dir,
        "%s.tp" % pred_filename)
    options.fn_path = os.path.join(options.output_dir,
        "%s.fn" % pred_filename)
    options.fp_path = os.path.join(options.output_dir,
        "%s.fp" % pred_filename)
    options.coords_paths = dict(
        gtruth=os.path.join(options.output_dir,
          gtruth_filename.replace("simcheck", "bed")),
        pred=os.path.join(options.output_dir,
          pred_filename.replace("simcheck", "bed")))
    CheckDirPath(options.output_dir, "output", path_errors,
      create=False, replace=True)
    # get the log file name
    options.log_file_name = GetLogPath(options.pred_path,
      "simcheck", options.output_dir) #TODO
  #} end if
  if (0 < len(path_errors)): #{
    ErrMsg("Errors in input arguments:\n  %s" % "\n  ".join(path_errors))
  #} end if
  # the paths are good if there are no path errors and no conflicting options
  return (opts_good and 0 == len(path_errors))
#} end def

def Main(): #{
  # get options and arguments
  parser = SetupOptionsParser()
  (options, args) = parser.parse_args()
  # if the right number of args was used
  if (parser.num_args == len(args)): #{
    options.gtruth_path = EnsureAbsPath(args[0])
    options.pred_path = EnsureAbsPath(args[1])
    if (CheckPaths(options)): #{
      try: #{
        main_class_object = MainClassCls(options)
        WriteCommand(main_class_object, sys.argv)
        main_class_object.Run()
      except (MyError), e:
        ErrMsg("ERROR while checking predictions against groundtruth:\n  %s" % e) #TODO
        return ES_RUN_ERR
      #} end try
    else:
      return ES_PATH_ERR
    #} end if
  else:
    parser.error("you must specify the paths to groundtruth and prediction "
      "simcheck files (GROUNDTRUTH_FILE, PREDICTION_FILE).")
    return ES_OPT_ERR
  #} end if
  return ES_SUCCESS
#} end def

if __name__ == '__main__': #{
  try: #{
    exit_status = Main()
  except Exception, e:
    traceback.print_exc()
    exit_status = ES_EXCEPTION
  #} end try
  if (ES_SUCCESS == exit_status): #{
    print MSG_SUCCESS
  else:
    print MSG_FAIL
  #} end if
  sys.exit(exit_status)
#} end if
